(* $Id: da.mli,v 1.26 2011-02-08 11:02:09 deraugla Exp $ *)

open Dim_def;

type da_data =
  { postscript : mutable option out_channel;
    text_buf : Buffer.t;
    ps_A3 : mutable bool;
    ps_landscape : mutable bool;
    rt_dim : mutable damdim;
    ps_dim : mutable damdim;
    sched_width : mutable int;
    sched_height : mutable int;
    inter_task_dy : mutable int;
    da_co : common }
and common =
  { xd : Rt.xdata;
    xa : Rt.xargs;
    c_white_col : (Rt.color * (int * int * int));
    c_black_col : (Rt.color * (int * int * int));
    c_red_col : (Rt.color * (int * int * int));
    c_blue_col : (Rt.color * (int * int * int));
    c_backg_rep_col : (Rt.color * (int * int * int));
    c_gray_col : (Rt.color * (int * int * int));
    c_color : array (Rt.color * (int * int * int));
    c_raw_font : Rt.font;
    pixmap_sched : Rt.pixmap;
    pixmap_dameid : Rt.pixmap;
    pixmap_caption : Rt.pixmap;
    watch : Rt.mouse;
    hand : Rt.mouse }
;

type da_color =
  { da_color : Rt.color; da_rgb : (int * int * int); da_c_dd : da_data }
and da_pixmap =
  { da_pixmap : Rt.pixmap; da_p_name : string; da_p_dd : da_data }
and da_font =
  { da_font : Rt.font;
    da_f_size : int;
    da_f_name : string;
    da_f_dd : da_data }
;

type da_paint =
  [ DaColorPn of da_color
  | DaPixmapPn of da_pixmap
  | DaNonePn ]
;

type drawable = (da_data * string * Rt.widget);

value xcoord_of_coord : drawable -> (int * int) -> (int * int);

value da_drawable_width : drawable -> int;
value da_drawable_height : drawable -> int;
value da_drawable_dim : drawable -> damdim;
value da_set_line_width : da_data -> int -> unit;
value da_set_dashes : da_data -> int -> list int -> unit;
value da_create_color : da_data -> (Rt.color * (int * int * int)) -> da_color;
value da_black_color : da_data -> da_color;
value da_white_color : da_data -> da_color;
value da_select_color : da_color -> unit;
value da_select_pixmap : da_pixmap -> unit;
value da_select_paint : da_paint -> unit;
value da_select_font : da_font -> unit;

value da_clear_widget : drawable -> unit;
value da_clear_area : drawable -> (int * int * int * int) -> unit;
value da_draw_arc : drawable -> (int * int * int * int * int * int) -> unit;
value da_fill_arc : drawable -> (int * int * int * int * int * int) -> unit;
value da_fill_rectangle : drawable -> (int * int * int * int) -> unit;
value da_draw_rectangle : drawable -> (int * int * int * int) -> unit;
value da_draw_line : drawable -> (int * int) -> (int * int) -> unit;
value da_draw_lines : drawable -> list (int * int) -> unit;
value da_erase_draw_string : drawable -> (int * int) -> string -> unit;
value da_draw_string : drawable -> (int * int) -> string -> unit;
value da_term_send : da_data -> Rt.widget -> string -> unit;
value da_text_width : da_font -> string -> int;
value da_font_size : da_font -> int;
value da_commit_display : drawable -> unit;

value rev_implode : list char -> string;
value postscript_save : (unit -> 'a) -> da_data -> out_channel -> unit;

value print_def_font_with_accents : out_channel -> unit;
