(* $Id: dameidfile.ml,v 1.8 2011-02-03 13:08:54 deraugla Exp $ *)

open Da;
open Data_rt;
open Glob_glob;
open Printf;
open Rt;

type dir_files_info =
  { selected : mutable option string;
    time_last_click : mutable int;
    scroll_ini : mutable int;
    scroll_pos : mutable int;
    old_x : mutable int;
    nrows : mutable int;
    max_rows : mutable int }
;

value vmax = 1000;
value double_click_time = 500;

value kill_widget gd wid = do {
  let xd = xdata_of_widget wid in
  rt_destroy_widget wid;
  Dutil.redirect_to_main gd;
  rt_unfreeze_widget (widget_named xd "main");
};

value set_dir_shift gd dfi files_wid incr =
  let v = term_shift_value files_wid + incr in
  if v >= 0 && v <= term_history_size files_wid then do {
    term_shift files_wid v;
    let xd = xdata_of_widget files_wid in
    let wid = widget_named xd "dir_files_scr" in
    let glop = vmax / dfi.max_rows in
    scroll_set wid (scroll_val wid - incr * glop);
  }
  else ()
;

value draw_dir_files gd wid ncols = do {
  let (a, _) = Openfile.dir_files gd.open_dir ncols in
  term_send wid "\027[3J";
  let (nrow, ncol) = term_size wid in
  let last_row =
    loop 1 0 0 where rec loop crow ccol i =
      if i = Array.length a then crow
      else do {
        let (row, col, name) = a.(i) in
        loop crow ccol where rec loop crow ccol =
          if crow < row then do {
            term_send wid "\r\n";
            loop (crow + 1) 0
          }
          else if ccol < col then do {
            term_send wid " ";
            loop crow (ccol + 1)
          }
          else ();
        term_send wid name;
        loop row (col + String.length name) (i + 1)
      }
  in
  term_send wid Std_term.vt_erase_from_cursor;
  if last_row > nrow then term_shift wid (last_row - nrow) else ();
  let max_row = Openfile.dir_files_nb_rows a in
  let scr_wid = widget_named (xdata_of_widget wid) "dir_files_scr" in
  scroll_set scr_wid (if max_row <= nrow then vmax else vmax * nrow / max_row)
};

value add_scenario_windows xa xd ti tg def_keyp len = do {
  rt_raise_widget (widget_named xd "main_tab");
  let wid = widget_named xd "open" in
  kill_widget tg wid;
  for i = len to Array.length tg.scenario - 1 do {
    let sc = tg.scenario.(i) in
    let wdesc = Scenario_rt.scenario_desc xa tg ti def_keyp i in
    Tabwid.tab_add xd ti sc.fname wdesc;
    Scenario_rt.make_moving_schedule_widget tg i;
    Expose.draw_text sc
      (widget_named xd (sprintf "text_%d" tg.ind_of_isc.(i)));
  };
  if Array.length tg.scenario > 1 then Tabwid.tab_show xd ti
  else Tabwid.tab_hide xd ti;
  Schedule_step_rt.close_progr_bar xd;
  Scenario_rt.raise_scenario tg ti (Array.length tg.ind_of_isc - 1);
  let main_wid = widget_named xd "main" in
  let (_, _) = rt_adjust_widget main_wid in
  ()
};

value chdir_open_file xa xd tg ti dfi def_keyp fname = do {
  Sys.chdir tg.open_dir;
  if try Sys.is_directory fname with [ Sys_error _ -> False ] then
    let dir =
      try do { Sys.chdir fname; Some (Sys.getcwd ()) } with
      [ Sys_error _ -> None ]
    in
    match dir with
    [ Some dir -> do {
        let wid = widget_named xd "opened_dir" in
        term_send wid Std_term.vt_erase_in_display;
        term_send wid Std_term.vt_home;
        term_send wid Std_term.vt_bold;
        term_send wid dir;
        tg.open_dir := dir;
        let wid = widget_named xd "dir_files" in
        let (nrows, ncols) = term_size wid in
        term_send wid Std_term.vt_erase_in_display;
        term_send wid Std_term.vt_home;
        term_send wid Std_term.vt_bold;
        let (a, nb_files) = Openfile.dir_files tg.open_dir ncols in
        dfi.max_rows := Openfile.dir_files_nb_rows a;
        term_set_max_history_size wid (max 50 nb_files);
        let sz =
          if dfi.max_rows <= nrows then vmax
          else nrows * vmax / dfi.max_rows
        in
        let scr_wid = widget_named xd "dir_files_scr" in
        scroll_set_bar_size scr_wid sz;
        draw_dir_files tg wid ncols;
      }
    | None -> () ]
  else do {
    let make_scenario = Scenario_rt.make_scenario xd tg.common in
    match Openfile.input_dameid_scenarios tg fname make_scenario with
    [ Some len -> add_scenario_windows xa xd ti tg def_keyp len
    | None -> () ]
  }
};

value action_open_file xa xd gd ti dfi def_keyp () =
  match dfi.selected with
  [ Some fname -> do {
      dfi.selected := None;
      let files_wid = widget_named xd "dir_files" in
      term_emphasize_from files_wid 0 0;
      term_emphasize_to files_wid 0 0;
      let fname =
        let len = String.length fname in
        if len > 0 && fname.[len-1] = '/' then String.sub fname 0 (len - 1)
        else fname
      in
      Openfile.save_dir_excursion
        (fun () -> chdir_open_file xa xd gd ti dfi def_keyp fname);
    }
  | None -> () ]
;

value action_scroll_dir_files gd dfi wid ev =
  let xd = xdata_of_widget wid in
  match ev with
  [ ScrollEvButtonPress but v -> do {
      dfi.scroll_ini := scroll_val wid;
      dfi.scroll_pos := v - dfi.scroll_ini;
    }
  | ScrollEvButtonMotion v ->
      if dfi.max_rows <= dfi.nrows then ()
      else do {
        let scroll_bsize = min vmax (vmax * dfi.nrows / dfi.max_rows) in
        let glop = vmax / dfi.max_rows in
        let x = (v - dfi.scroll_pos) / glop * glop in
        let x = min vmax (max scroll_bsize x) in
        if dfi.old_x <> 0 && dfi.old_x = x then ()
        else do {
          dfi.old_x := x;
          scroll_set wid x;
          let wid = widget_named xd "dir_files" in
          term_shift wid ((vmax - x) / glop);
        }
      }
  | ScrollEvArrowPress 1 -> do {
      let files_wid = widget_named xd "dir_files" in
      set_dir_shift gd dfi files_wid 1
    }
  | ScrollEvArrowPress 2 -> do {
      let files_wid = widget_named xd "dir_files" in
      set_dir_shift gd dfi files_wid (-1)
    }
  | _ -> () ]
;

value action_dir_files xa gd ti dfi def_keyp files_wid =
  fun
  [ TermEvSizeChange nrows ncols -> do {
      term_send files_wid Std_term.vt_home;
      term_send files_wid Std_term.vt_erase_in_display;
      term_send files_wid Std_term.vt_bold;
      draw_dir_files gd files_wid ncols;
      let xd = xdata_of_widget files_wid in
      let (a, _) = Openfile.dir_files gd.open_dir ncols in
      dfi.selected := None;
      dfi.nrows := nrows;
      dfi.max_rows := Openfile.dir_files_nb_rows a;
      let sz =
        if dfi.max_rows <= nrows then vmax
        else nrows * vmax / dfi.max_rows
      in
      let wid = widget_named xd "dir_files_scr" in
      scroll_set_bar_size wid sz;
    }
  | TermEvButtonPress row char_col curs_col {item = 1 | 2 | 3} ->
      let (nrow, ncol) = term_size files_wid in
      if row >= nrow || char_col >= ncol then ()
      else do {
        let (a, _) =
          Openfile.dir_files gd.open_dir (snd (term_size files_wid))
        in
        let shift =
          max 0 (dfi.max_rows - nrow) - term_shift_value files_wid
        in
        let row = row + shift in
        let row_col_fname_opt =
          loop 0 where rec loop i =
            if i = Array.length a then
              if i > 0 then Some a.(i-1) else None
            else
              let (r, c, f) = a.(i) in
              if row + 1 < r then
                if i > 0 then Some a.(i-1) else None
              else if row + 1 > r then loop (i + 1)
              else if char_col < c then Some a.(i-1)
              else loop (i + 1)
        in
        match row_col_fname_opt with
        [ Some (r, c, fn) -> do {
            term_emphasize_from files_wid (r - shift - 1) c;
            term_emphasize_to files_wid (r - shift - 1)
              (c + String.length fn);
            let tm = rt_current_time xa in
            if tm - dfi.time_last_click <= double_click_time &&
               dfi.selected = Some fn
            then do {
              dfi.time_last_click := tm;
              dfi.selected := Some fn;
              let xd = xdata_of_widget files_wid in
              action_open_file xa xd gd ti dfi def_keyp ();
            }
            else do {
              dfi.time_last_click := tm;
              dfi.selected := Some fn;
            }
          }
        | None -> () ];
      }
  | TermEvButtonPress row char_col curs_col {item = 4} ->
      set_dir_shift gd dfi files_wid 1
  | TermEvButtonPress row char_col curs_col {item = 5} ->
      set_dir_shift gd dfi files_wid (-1)
  | _ -> () ]
;

value open_file xa gd ti def_keyp ul_xy (x, y) = do {
  let xd = gd.common.xd in
  let main_wid = widget_named xd "main" in
  let nrows = 20 in
  let ncols = 90 in
  let (files, nb_files) = Openfile.dir_files gd.open_dir ncols in
  let max_rows = Openfile.dir_files_nb_rows files in
  let wid =
    let dfi =
      {selected = None; time_last_click = 0; scroll_ini = 0; scroll_pos = 0;
       old_x = 0; nrows = nrows; max_rows = max_rows}
    in
    let filler = pack_desc [FillerAtt] (Vertical, []) (fun _ _ -> ()) in
    rt_create_transient_widget main_wid (I18n.transl "Open")
      (Some (kill_widget gd))
      (pack_desc [NameAtt "open"]
         (Vertical,
          [pack_desc []
             (Horizontal,
              [title_desc [] (I18n.transl "Directory") (fun _ _ -> ());
               filler])
             (fun _ _ -> ());
           term_desc [NameAtt "opened_dir"] (1, 60, 0) (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [filler;
               title_desc [] (I18n.transl "Files") (fun _ _ -> ());
               filler])
             (fun _ _ -> ());
           pack_desc [FillerAtt]
             (Horizontal,
              [scroll_desc [NameAtt "dir_files_scr"]
                 (Vertical, 0, vmax, min vmax (vmax * nrows / max_rows))
                 (action_scroll_dir_files gd dfi);
               term_desc [NameAtt "dir_files"; FillerAtt]
                 (nrows, ncols, max 50 nb_files)
                 (action_dir_files xa gd ti dfi def_keyp)])
             (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [filler;
               button_desc [] (I18n.transl "Open", None)
                 (Dutil.button_press_action
                    (action_open_file xa xd gd ti dfi def_keyp));
               filler;
               button_desc [] (I18n.transl "Cancel", None)
                 (Dutil.button_press_action
                    (fun () ->
                       let wid = widget_named xd "open" in
                       kill_widget gd wid));
               filler])
             (fun _ _ -> ())])
         (Dutil.action_pack_esc_ret
            (fun wid -> kill_widget gd (widget_named xd "open"))
            (fun wid -> ())))
  in
  let dir = Sys.getcwd () in
  term_send (widget_named xd "opened_dir") (Std_term.vt_bold ^ dir);
  let files_wid = widget_named xd "dir_files" in
  term_send files_wid Std_term.vt_bold;
  draw_dir_files gd files_wid ncols;
  let (x, y) =
    match ul_xy with
    [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
    | None -> (x, y) ]
  in
  rt_freeze_widget (widget_named xd "main");
  rt_redirect_key_press_to wid;
  rt_map_transient_widget wid x y
};
