(* $Id: draw_dameid.ml,v 1.94 2011-02-08 10:52:42 deraugla Exp $ *)

open Ast;
open Dim_def;
open Glob_glob;
open Dtasks;
open Printf;

value pi = 3.1415926535897933238462643383279502884197189399;
value round x = int_of_float (x +. 0.5);

value grand_tour = 360 * 64;
value quarter_tour = 90 * 64;

value disp_lim = 500000;

value xy_of_rad_grad lcm (radius, graduation, delta_ang) =
  let angle = float graduation *. 2. *. pi /. float lcm +. delta_ang in
  (round (float radius *. cos angle), round (float radius *. sin angle))
;

value visible_angles xc yc w h =
  let c = 180. *. 64. /. pi in
  if yc < 0 then
    if xc < 0 then
      (270 * 64 + truncate (atan2 (float (- xc)) (float (h - yc)) *. c),
       270 * 64 + truncate (atan2 (float (w - xc)) (float (- yc)) *. c))
    else if xc < w then
      (270 * 64 - truncate (atan2 (float xc) (float (- yc)) *. c),
       270 * 64 + truncate (atan2 (float (w - xc)) (float (- yc)) *. c))
    else
      (180 * 64 + truncate (atan2 (float (- yc)) (float xc) *. c),
       180 * 64 + truncate (atan2 (float (h - yc)) (float (xc - w)) *. c))
  else if yc < h then
    if xc < 0 then
      (360 * 64 - truncate (atan2 (float (h - yc)) (float (- xc)) *. c),
       truncate (atan2 (float yc) (float (- xc)) *. c))
    else if xc < w then (0, 360 * 64)
    else
      (180 * 64 - truncate (atan2 (float yc) (float (xc - w)) *. c),
       180 * 64 + truncate (atan2 (float (h - yc)) (float (xc - w)) *. c))
  else
    if xc < 0 then
      (truncate (atan2 (float (yc - h)) (float (w - xc)) *. c),
       truncate (atan2 (float yc) (float (- xc)) *. c))
    else if xc < w then
      (90 * 64 - truncate (atan2 (float (w - xc)) (float (yc - h)) *. c),
       90 * 64 + truncate (atan2 (float xc) (float (yc - h)) *. c))
    else
      (90 * 64 + truncate (atan2 (float (xc - w)) (float yc) *. c),
       90 * 64 + truncate (atan2 (float xc) (float (yc - h)) *. c))
;

value draw_zoomed_sector gfun sc drw n x y dx dy gc isect = do {
  let a = Sutil.mul_div isect grand_tour sc.ts.lcm in
  let da = grand_tour / sc.ts.lcm + 1 in
  let dim = gfun.g_dam_dim drw in
  let w = gfun.g_drawable_width drw in
  let h = gfun.g_drawable_height drw in
  let bbx = n * (dim.dam_bb_x - x) - dx + w / 2 in
  let bby = n * (dim.dam_bb_y - y) - dy + h / 2 in
  let r = n * dim.dam_bb_width / 2 in
  let xc = bbx + r in
  let yc = bby + r in
  let r =
    min r
      (3 *
       max (if xc < w / 2 then w - xc else xc)
         (if yc < h / 2 then h - yc else yc)
      / 2)
  in
  let (a1_vis, a2_vis) = visible_angles xc yc w h in
  let a1 =
    let a1 = a mod grand_tour in
    if a1 < 0 then a1 + grand_tour else a1
  in
  let a2 =
    let a2 = (a + da) mod grand_tour in
    if a2 < 0 then a2 + grand_tour else a2
  in
  if a1_vis < a2_vis &&
       (a1 >= a1_vis && a1 <= a2_vis || a2 >= a1_vis && a2 <= a2_vis ||
        a1 <= a1_vis && a2 >= a2_vis) ||
     a1_vis > a2_vis &&
       (a1 >= a1_vis || a1 <= a2_vis || a2 >= a1_vis || a2 <= a2_vis)
  then do {
    let bbx = xc - r in
    let bby = yc - r in
    let bbw = 2 * r in
    let bbh = 2 * r in
    let m = 16000 in
    if abs bbx > m || abs bby > m || abs bbw > m || abs bbh > m then do {
      let a0 = atan2 (float (yc - h / 2)) (float (xc - w / 2)) in
      let rr = float r in
      let x0 = truncate (float xc -. rr *. cos a0) in
      let y0 = truncate (float yc -. rr *. sin a0) in

      let a1 = float isect *. 2. *. pi /. float sc.ts.lcm in
      let k1 =
        (float (x0 - xc) *. sin a1 +. float (y0 - yc) *. cos a1) /.
        (sin a0 *. sin a1 -. cos a0 *. cos a1)
      in
      let x1 = x0 - truncate (k1 *. sin a0) in
      let y1 = y0 + truncate (k1 *. cos a0) in

      let a2 = a1 +. 2. *. pi /. float sc.ts.lcm in
      let k2 =
        (float (x0 - xc) *. sin a2 +. float (y0 - yc) *. cos a2) /.
        (sin a0 *. sin a2 -. cos a0 *. cos a2)
      in
      let x2 = x0 - truncate (k2 *. sin a0) in
      let y2 = y0 + truncate (k2 *. cos a0) in

      let kk = 16000.0 in
      let x3 = x2 + truncate (kk *. cos a0) in
      let y3 = y2 + truncate (kk *. sin a0) in
      let x4 = x1 + truncate (kk *. cos a0) in
      let y4 = y1 + truncate (kk *. sin a0) in

      let zmin = -32000 in
      let zmax = -zmin in
      let x1 = max zmin (min zmax x1) in
      let x2 = max zmin (min zmax x2) in
      let x3 = max zmin (min zmax x3) in
      let x4 = max zmin (min zmax x4) in
      let y1 = max zmin (min zmax y1) in
      let y2 = max zmin (min zmax y2) in
      let y3 = max zmin (min zmax y3) in
      gfun.g_fill_polygon drw gc
        [(x1, y1); (x2, y2); (x3, y3); (x4, y4)];
    }
    else do {
      gfun.g_fill_arc drw gc (bbx, bby, bbw, bbh) (a, da)
    };
  }
  else ()
};

value draw_normal_sector gfun sc drw gc isect = do {
  let a = Sutil.mul_div isect grand_tour sc.ts.lcm in
  let da = grand_tour / sc.ts.lcm + 1 in
  let dim = gfun.g_dam_dim drw in
  let (bbx, bby, bbw, bbh) =
    (dim.dam_bb_x, dim.dam_bb_y, dim.dam_bb_width, dim.dam_bb_height)
  in
  gfun.g_fill_arc drw gc (bbx, bby, bbw, bbh) (a, da)
};

value draw_sector tg sc drw gc zoom isect =
  if sc.ts.schedule_len < disp_lim then do {
    match zoom with
    [ Zoom zi wid dx dy ->
        draw_zoomed_sector tg sc drw zi.zi_n zi.zi_x zi.zi_y dx dy gc isect
    | NoZoom ->
        draw_normal_sector tg sc drw gc isect ]
  }
  else ()
;

value task_color sc itask =
  loop 0 where rec loop i =
    if sc.initial_order.(i) = itask then i
    else loop (i + 1)
;

value draw_sectors gfun sc graph drw gc zoom = do {
  let ts = sc.ts in
  if ts.schedule_len < disp_lim then do {
    let sh_dam = ts.schedule_len - ts.lcm in
    for i = 0 to ts.lcm - 1 do {
      let j = sh_dam + i in
      match Sarray.get ts.disp_schedule j with
      [ Some u -> do {
          if u.u_preempt then
            gfun.g_set_color graph gc C_black
          else if sc.coloured then
            gfun.g_set_color graph gc (C_number (task_color sc u.u_task))
          else
            gfun.g_set_pixmap graph gc u.u_task;
          let isect = (j + ts.min_rel_time) mod ts.lcm in
          let isect = if isect < 0 then isect + ts.lcm else isect in
          draw_sector gfun sc drw gc zoom isect
        }
      | None -> () ];
    }
  }
  else ()
};

value max_total_axes = 30;

value draw_task_axis gfun sc drw gc i =
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let (xc, yc) = gfun.g_xcoord_of_coord drw (0, 0) in
  let td = ts.tasks_def.(i) in
  let dim = gfun.g_dam_dim drw in
  let step =
    let max_axes = max_total_axes / Array.length ts.tasks_def in
    let n_axes = ts2.lcm / td.period in
    (n_axes + max_axes - 1) / max_axes
  in
  for j = 0 to ts2.lcm / td.period - 1 do {
    if j mod step = 0 then do {
      let k = td.release + j * td.period in
      let (x, y) = xy_of_rad_grad ts2.lcm (dim.radius + 50, k, 0.) in
      let (x, y) = gfun.g_xcoord_of_coord drw (x, y) in
      gfun.g_draw_line drw gc (xc, yc) (x, y);
      let (x1, y1) = xy_of_rad_grad ts2.lcm (dim.radius + 45, k, 0.02) in
      let (x1, y1) = gfun.g_xcoord_of_coord drw (x1, y1) in
      let (x2, y2) = xy_of_rad_grad ts2.lcm (dim.radius + 45, k, -0.02) in
      let (x2, y2) = gfun.g_xcoord_of_coord drw (x2, y2) in
      gfun.g_draw_lines drw gc [(x1, y1); (x, y); (x2, y2)];
    }
    else ();
  }
;

value draw_tasks_axes gfun sc drw gc = do {
  gfun.g_set_line_width sc.graph gc 0;
  gfun.g_set_color sc.graph gc C_black;
  for i = 0 to Array.length sc.ts.ast.tasks_def - 1 do {
    draw_task_axis gfun sc drw gc i
  };
};

value in_angle_range a a1_vis a2_vis =
  let a = a mod grand_tour in
  let a = if a < 0 then a + grand_tour else a in
  if a1_vis < a2_vis then a >= a1_vis && a <= a2_vis
  else a >= a1_vis - 1 || a <= a2_vis
;

value draw_outline gfun sc drw gc zoom = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  gfun.g_set_line_width sc.graph gc 2;
  gfun.g_set_color sc.graph gc C_black;
  gfun.g_set_font sc.graph gc F_raw;
  let dim = gfun.g_dam_dim drw in
  let (a1_vis, a2_vis) =
    match zoom with
    [ Zoom zi _ dx dy ->
        let n = zi.zi_n in
        let x = zi.zi_x in
        let y = zi.zi_y in
        let w = gfun.g_drawable_width drw in
        let h = gfun.g_drawable_height drw in
        let r = n * dim.dam_bb_width / 2 in
        let bbx = n * (dim.dam_bb_x - x) - dx + w / 2 in
        let bby = n * (dim.dam_bb_y - y) - dy + h / 2 in
        visible_angles (bbx + r) (bby + r) w h
    | NoZoom ->
        (0, grand_tour) ]
  in
  let da =
    if a1_vis < a2_vis then a2_vis - a1_vis
    else a2_vis - a1_vis + grand_tour
  in
  match zoom with
  [ Zoom zi _ dx dy ->
      let n = zi.zi_n in
      let x = zi.zi_x in
      let y = zi.zi_y in
      let w = gfun.g_drawable_width drw in
      let h = gfun.g_drawable_height drw in
      let r = n * dim.dam_bb_width / 2 in
      let bbx = n * (dim.dam_bb_x - x) - dx + w / 2 in
      let bby = n * (dim.dam_bb_y - y) - dy + h / 2 in
      let bbw = 2 * r in
      let bbh = 2 * r in
      let m = 16000 in
      if abs bbx > m || abs bby > m || abs bbw > m || abs bbh > m then
        let xc = bbx + r in
        let yc = bby + r in
        let a0 = atan2 (float (yc - h / 2)) (float (xc - w / 2)) in
        let rr = float r in
        let x0 = truncate (float xc -. rr *. cos a0) in
        let y0 = truncate (float yc -. rr *. sin a0) in
        let len = float (max w h) in
        let x1 = x0 - truncate (len *. sin a0) in
        let y1 = y0 + truncate (len *. cos a0) in
        let x2 = x0 + truncate (len *. sin a0) in
        let y2 = y0 - truncate (len *. cos a0) in
        gfun.g_draw_line drw gc (x1, y1) (x2, y2)
      else
        gfun.g_draw_arc drw gc (bbx, bby, bbw, bbh) (a1_vis, da)
  | NoZoom ->
      let (bbx, bby, bbw, bbh) =
        (dim.dam_bb_x, dim.dam_bb_y, dim.dam_bb_width, dim.dam_bb_height)
      in
      gfun.g_draw_arc drw gc (bbx, bby, bbw, bbh) (a1_vis, da) ];
  let show_graduations =
    match zoom with
    [ Zoom zi wid dx dy -> ts2.lcm <= 400 * zi.zi_n
    | NoZoom -> ts2.lcm <= 100 ]
  in
  if not show_graduations then ()
  else
    let smallest_period =
      Array.fold_left (fun sp td -> min sp td.period) max_int ts.tasks_def
    in
    let glop = ref True in
    for i = 0 to ts2.lcm - 1 do {
      match zoom with
      [ Zoom zi _ dx dy ->
          let n = zi.zi_n in
          let x = zi.zi_x in
          let y = zi.zi_y in
          let w = gfun.g_drawable_width drw in
          let h = gfun.g_drawable_height drw in
          if (ts2.lcm <= 100 * n || i mod smallest_period = 0) &&
             in_angle_range (Sutil.mul_div i grand_tour ts2.lcm) a1_vis
               a2_vis
          then do {
            let a = -. float i *. 2. *. pi /. float ts2.lcm in
            let r = n * dim.radius in
            let xc = n * (dim.dam_bb_x - x) - dx + w / 2 + r in
            let yc = n * (dim.dam_bb_y - y) - dy + h / 2 + r in
            let x1 = xc + truncate (float (r - 5) *. cos a) in
            let y1 = yc + truncate (float (r - 5) *. sin a) in
            let len =
              if i mod smallest_period = 0 && ts2.lcm <= 200 * n then 10
              else 5
            in
            let x2 = xc + truncate (float (r + len) *. cos a) in
            let y2 = yc + truncate (float (r + len) *. sin a) in
            gfun.g_draw_line drw gc (x1, y1) (x2, y2)
          }
          else ()
      | NoZoom ->
          let (x1, y1) = xy_of_rad_grad ts2.lcm (dim.radius - 5, i, 0.) in
          let (x2, y2) = xy_of_rad_grad ts2.lcm (dim.radius + 5, i, 0.) in
          let (x1, y1) = gfun.g_xcoord_of_coord drw (x1, y1) in
          let (x2, y2) = gfun.g_xcoord_of_coord drw (x2, y2) in
          gfun.g_draw_line drw gc (x1, y1) (x2, y2) ];
      let ii = i in
      let text = string_of_int ii in
      let text = " " ^ text ^ " " in
      let x0 = gfun.g_text_width sc.graph F_raw text / 2 in
      let y0 = gfun.g_font_size sc.graph F_raw * 9 / 20 (* gauging *) in
      match zoom with
      [ Zoom zi _ dx dy ->
          let n = zi.zi_n in
          let x = zi.zi_x in
          let y = zi.zi_y in
          let w = gfun.g_drawable_width drw in
          let h = gfun.g_drawable_height drw in
          if in_angle_range (Sutil.mul_div i grand_tour ts2.lcm) a1_vis
               a2_vis
          then
            if ts2.lcm <= 20 * n ||
               ts2.lcm <= 40 * n && i mod smallest_period = 0
            then
              let a = -. float i *. 2. *. pi /. float ts2.lcm in
              let r = n * dim.radius in
              let xc = n * (dim.dam_bb_x - x) - dx + w / 2 + r in
              let yc = n * (dim.dam_bb_y - y) - dy + h / 2 + r in
              let x = xc + truncate (float (r + 25) *. cos a) in
              let y = yc + truncate (float (r + 25) *. sin a) in
              let (x, y) = (x - x0, y + y0) in
              gfun.g_erase_draw_string drw gc (x, y) text
            else if
               ts2.lcm > 40 * n &&
               (ts2.lcm <= 100 * n || i mod smallest_period = 0)
            then do {
              let xy i =
                let a = -. float i *. 2. *. pi /. float ts2.lcm in
                let r = n * dim.radius in
                let xc = n * (dim.dam_bb_x - x) - dx + w / 2 + r in
                let yc = n * (dim.dam_bb_y - y) - dy + h / 2 + r in
                let x = xc + truncate (float (r + 25) *. cos a) in
                let y = yc + truncate (float (r + 25) *. sin a) in
                (x, y)
              in
              let (x, y) = xy i in
              let dist_center (x, y) =
                let dx = x - w / 2 in
                let dy = y - h / 2 in
                dx * dx + dy * dy
              in
              let dc = dist_center (x, y) in
              if dc < dist_center (xy (i + 1)) && glop.val then do {
                glop.val := False;
                let (x, y) = (x - x0, y + y0) in
                gfun.g_set_color sc.graph gc C_red;
                gfun.g_erase_draw_string drw gc (x, y) text;
                gfun.g_set_color sc.graph gc C_black;
              }
              else ();
            }
            else ()
          else ()
      | NoZoom ->
          if i mod smallest_period = 0 then
            let (x, y) = xy_of_rad_grad ts2.lcm (dim.radius + 25, i, 0.) in
            let (x, y) = gfun.g_xcoord_of_coord drw (x - x0, y - y0) in
            gfun.g_erase_draw_string drw gc (x, y) text
          else () ];
    };
};

value draw_axis_text gfun sc drw gc = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let axis_tasks = Array.create ts2.lcm [] in
  for i = 0 to Array.length ts.tasks_def - 1 do {
    let td = ts.tasks_def.(i) in
    let step =
      let max_axes = max_total_axes / Array.length ts.tasks_def in
      let n_axes = ts2.lcm / td.period in
      (n_axes + max_axes - 1) / max_axes
    in
    for j = 0 to ts2.lcm / td.period - 1 do {
      if j mod step = 0 then do {
        let k = (td.release + j * td.period) mod ts2.lcm in
        let k = if k < 0 then k + ts2.lcm else k in
        axis_tasks.(k) := axis_tasks.(k) @ [i];
      }
      else ();
    };
  };
  gfun.g_set_font sc.graph gc F_raw;
  let dim = gfun.g_dam_dim drw in
  let y0 = gfun.g_font_size sc.graph F_raw * 9 / 20 in
  for i = 0 to ts2.lcm - 1 do {
    let tl = axis_tasks.(i) in
    if tl = [] then ()
    else
(**)
      let text =
        loop "" tl where rec loop s =
          fun
          [ [t :: tl] ->
              let (range, tl) =
                loop t tl where rec loop t1 =
                  fun
                  [ [t2 :: tl] ->
                      if t2 = t1 + 1 then loop t2 tl
                      else (t1 - t, [t2 :: tl])
                  | [] ->
                      (t1 - t, []) ]
              in
              let s =
                let s1 = sprintf "t%d" (t + 1) in
                let s1 =
                  if range > 1 then sprintf "%s-t%d" s1 (t + 1 + range)
                  else if range = 1 then sprintf "%s,t%d" s1 (t + 2)
                  else s1
                in
                if s = "" then s1 else s ^ "," ^ s1
              in
              loop s tl
          | [] -> s ]
      in
(*
      let x0 =
        let text =
          String.concat "," (List.map (fun i -> sprintf "t%d" (i + 1)) tl)
        in
        da_text_width font text / 2
      in
*)
      let x0 = gfun.g_text_width sc.graph F_raw text / 2 in
(**)
      let (x, y) = xy_of_rad_grad ts2.lcm (dim.radius + 70, i, 0.) in
      let (x, y) = gfun.g_xcoord_of_coord drw (x - x0, y - y0) in
(*
      let _ =
        List.fold_left
          (fun (sep, x) i -> do {
             let x =
               if sep = "" then x
               else do {
                 da_select_color sc.black_col;
                 da_erase_draw_string drw (x, y) sep;
                 x + da_text_width font sep
               }
             in
             let text = sprintf "t%d" (i + 1) in
             if sc.coloured then da_select_color (task_color sc i) else ();
             da_erase_draw_string drw (x, y) text;
             (",", x + da_text_width font text)
           })
          ("", x) tl
      in
      ();
*)
      gfun.g_erase_draw_string drw gc (x, y) text
(**)
  };
};

value jammed_by_axis ts2 a da =
  let ts = ts2.ast in
  loop 0 where rec loop itask =
    if itask = Array.length ts.tasks_def then False
    else
      let td = ts.tasks_def.(itask) in
      loop2 0 where rec loop2 j =
        if j = ts2.lcm / td.period then loop (itask + 1)
        else
          let a1 =
            ((td.release + j * td.period) mod ts2.lcm) * grand_tour /
            ts2.lcm
          in
          let a1 = if a1 < 0 then a1 + grand_tour else a1 in
          if a1 >= a - 3 * 64 && a1 <= a + da + 3 * 64 then True
          else loop2 (j + 1)
;

value jammed_by_latencies gd a da =
  let ok_lat =
    List.fold_right
      (fun l ok_lat ->
         match l.l_status with
         [ LS_values be1 be2 lat -> [(l, be1, be2, lat) :: ok_lat]
         | _ -> ok_lat ])
      gd.ts.latency_constr []
  in
  let len = List.length ok_lat in
  if len > 0 then
    let ts2 = gd.ts in
    let (l, be1, be2, lat) = List.nth ok_lat (gd.latency_disp_cnt mod len) in
    let lim = if ts2.rmax = 0 then 0 else ts2.rmax + ts2.lcm in
    let beg1 = fst be1 in
    let end2 = snd be2 in
    if beg1 >= lim && end2 >= lim then
      if lat <= ts2.lcm then False else True
    else False
  else False
;

value jammed gd a da =
  jammed_by_axis gd.ts a da || jammed_by_latencies gd a da
;

value draw_orient gfun sc drw gc erase =
  let d = 40 in
  let a = 20 * 64 in
  let da = 15 * 64 in
  if not (jammed sc a da) then do {
    gfun.g_set_line_width sc.graph gc 2;
    if erase then gfun.g_set_color sc.graph gc C_white
    else gfun.g_set_color sc.graph gc C_black;
    let dim = gfun.g_dam_dim drw in
    gfun.g_draw_arc drw gc
      (dim.dam_bb_x - d, dim.dam_bb_y - d, dim.dam_bb_width + 2 * d,
       dim.dam_bb_height + 2 * d) (a, da);
    let (x1, y1) =
      xy_of_rad_grad grand_tour (dim.radius + d - 4, a + da - 3 * 64, 0.)
    in
    let (x1, y1) = gfun.g_xcoord_of_coord drw (x1, y1) in
    let (x2, y2) = xy_of_rad_grad grand_tour (dim.radius + d, a + da, 0.) in
    let (x2, y2) = gfun.g_xcoord_of_coord drw (x2, y2) in
    let (x3, y3) =
      xy_of_rad_grad grand_tour (dim.radius + d + 4, a + da - 3 * 64, 0.)
    in
    let (x3, y3) = gfun.g_xcoord_of_coord drw (x3, y3) in
    gfun.g_draw_lines drw gc [(x1, y1); (x2, y2); (x3, y3)];
    let (x, y) =
      xy_of_rad_grad grand_tour (dim.radius + d + 20, a + da / 2, 0.)
    in
    let (x, y) = gfun.g_xcoord_of_coord drw (x, y) in
    gfun.g_draw_line drw gc (x - 5, y) (x + 5, y);
    gfun.g_draw_line drw gc (x, y - 5) (x, y + 5);
  }
  else ()
;

value draw_skeleton gfun sc drw gc zoom =
  if sc.ts.schedule_len < disp_lim then do {
    if zoom = NoZoom then draw_tasks_axes gfun sc drw gc else ();
    draw_outline gfun sc drw gc zoom;
    if zoom = NoZoom then draw_axis_text gfun sc drw gc else ();
    if zoom = NoZoom then draw_orient gfun sc drw gc False
    else do {
      let x = gfun.g_drawable_width drw / 2 in
      let y = gfun.g_drawable_height drw / 2 in
      let r = 5 in
      gfun.g_set_line_width sc.graph gc 1;
      gfun.g_set_color sc.graph gc C_black;
      gfun.g_draw_arc drw gc (x - r, y - r, 2 * r, 2 * r) (0, 360 * 64);
    }
  }
  else ()
;

value draw_spiral gfun sc drw gc (xc, yc) radius dr a da = do {
  let incr_x = [| 1; 0; 1; 2 |] in
  let incr_y = [| 0; 1; 2; 1 |] in
  let a =
    let a = a mod grand_tour in
    if a < 0 then a + grand_tour else a
  in
  let (da, need_dash) =
    if da > 3 * grand_tour then
      (2 * grand_tour + da mod grand_tour, True)
    else
      (da, False)
  in
  let da_lim1 = da / 2 + 45 * 64 in
  let da_lim2 = da / 2 - 45 * 64 in
  let bb_x = xc - radius in
  let bb_y = yc - radius in
  let bb_w = 2 * radius in
  let bb_h = 2 * radius in
  let (prev_start_a, i) =
    if a < quarter_tour then (0, 3)
    else if a < 180 * 64 then (quarter_tour, 0)
    else if a < 270 * 64 then (180 * 64, 1)
    else (270 * 64, 2)
  in
  let loop_start_a = prev_start_a + quarter_tour in
  gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, loop_start_a - a);
  loop i bb_x bb_y bb_w bb_h prev_start_a (da - (loop_start_a - a))
  where rec loop i bb_x bb_y bb_w bb_h a da =
    let i = (i + 1) mod 4 in
    let bb_x = bb_x - incr_x.(i) * dr in
    let bb_y = bb_y - incr_y.(i) * dr in
    let bb_w = bb_w + 2 * dr in
    let bb_h = bb_h + 2 * dr in
    let a = (a + quarter_tour) mod grand_tour in
    if da >= quarter_tour then do {
      if need_dash && da >= da_lim1 && da - quarter_tour < da_lim1
      then do {
        let da1 = da - da_lim1 in
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, da1);
        gfun.g_set_dashes sc.graph gc 0 [5; 5];
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h)
          (a + da1, quarter_tour - da1);
        gfun.g_set_line_width sc.graph gc 2;
      }
      else if
        need_dash && da >= da_lim2 && da - quarter_tour < da_lim2
      then do {
        let da1 = da - da_lim2 in
        gfun.g_set_dashes sc.graph gc 0 [5; 5];
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, da1);
        gfun.g_set_line_width sc.graph gc 2;
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h)
          (a + da1, quarter_tour - da1);
      }
      else do {
        gfun.g_set_line_width sc.graph gc 2;
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, quarter_tour);
      };
      loop i bb_x bb_y bb_w bb_h a (da - quarter_tour)
    }
    else do {
      let new_radius = bb_w / 2 in
      let dda =
        (* angle error due to the shift of the center *)
        round
          (2. *. asin (float dr /. 2. /. float new_radius) /. pi *. 180. *.
           64.)
      in
      let da = da - dda in
      gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, da);
      (bb_x + new_radius, bb_y + new_radius, a + da, new_radius)
    }
};

value draw_latency gfun sc drw gc zoom l (beg1, _) (_, end2) lat
    erase = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let lc = l.l_constr in
  let lim =
    if ts2.rmax = 0 then 0
    else
      match ts2.repetition with
      [ Some (i, _) -> i
      | None -> ts2.rmax + ts2.lcm ]
  in
  let dim = gfun.g_dam_dim drw in
  if beg1 >= lim && end2 >= lim then do {
    let isect1 =
      let isect = (beg1 + ts2.min_rel_time) mod ts2.lcm in
      if isect < 0 then isect + ts2.lcm else isect
    in
    let isect1_0 = beg1 + ts2.min_rel_time in
    let isect2_0 = end2 + ts2.min_rel_time in
    let isect2 = isect2_0 - (isect1_0 - isect1) in
    gfun.g_set_line_width sc.graph gc 2;
    if erase then gfun.g_set_color sc.graph gc C_white
    else if lat > lc.l_max then
      if sc.coloured then gfun.g_set_color sc.graph gc C_red
      else gfun.g_set_color sc.graph gc C_gray
    else
      gfun.g_set_color sc.graph gc C_black;
    let a = isect1 * grand_tour / ts2.lcm in
    let da =
      if isect2 > isect1 + max_int /grand_tour then
        (isect2 - isect1) mod ts2.lcm * grand_tour / ts2.lcm +
        (isect2 - isect1) / ts2.lcm * grand_tour
      else if isect2 > isect1 then
        (isect2 - isect1) * grand_tour / ts2.lcm
      else
        (isect2 - isect1 + ts2.lcm) * grand_tour / ts2.lcm
    in
    let many_tasks = Dim.many_tasks ts.tasks_def in
    let dr = if many_tasks then 10 else 20 in
    let dr_i = if many_tasks then 3 else 5 in
    let (ending_xc, ending_yc, ending_angle, ending_radius) =
      if lat <= ts2.lcm then do {
        let bb_x = dim.dam_bb_x - dr in
        let bb_y = dim.dam_bb_y - dr in
        let bb_w = dim.dam_bb_width + 2 * dr in
        let bb_h = dim.dam_bb_height + 2 * dr in
        gfun.g_draw_arc drw gc (bb_x, bb_y, bb_w, bb_h) (a, da);
        (bb_x + bb_w / 2, bb_y + bb_h / 2, a + da, bb_w / 2);
      }
      else
        draw_spiral gfun sc drw gc (dim.x_center, dim.y_center)
          (dim.radius + dr) dr_i a da
    in
    let draw_arrow i d =
      let xy_of_ang_rad a da r =
        let a = float a *. pi /. 180. /. 64. in
        let x = round (float r *. cos (a +. da)) in
        let y = round (float r *. sin (a +. da)) in
        (x + ending_xc, - y + ending_yc)
      in
      let xy1 = xy_of_ang_rad (ending_angle + 20) d (ending_radius + 7) in
      let xy2 = xy_of_ang_rad (ending_angle + 20) 0. (ending_radius + 1) in
      let xy3 = xy_of_ang_rad (ending_angle + 20) d (ending_radius - 5) in
      gfun.g_draw_lines drw gc [xy1; xy2; xy3]
    in
    let dot_rd = 4 in
    let (x, y) =
      let xy = xy_of_rad_grad ts2.lcm (dim.radius + dr, isect1, 0.) in
      gfun.g_xcoord_of_coord drw xy
    in
    let x = x - 1 in
    gfun.g_fill_arc drw gc
      (x - dot_rd, y - dot_rd, 2 * dot_rd, 2 * dot_rd) (0, grand_tour);
    draw_arrow isect2 (-0.1);
  }
  else
    ()
};

value draw_latencies gfun sc drw gc zoom erase =
  let ok_lat =
    List.fold_right
      (fun l ok_lat ->
         match l.l_status with
         [ LS_values be1 be2 lat -> [(l, be1, be2, lat) :: ok_lat]
         | _ -> ok_lat ])
      sc.ts.latency_constr []
  in
  let len = List.length ok_lat in
  if len > 0 then
    let (l, be1, be2, lat) = List.nth ok_lat (sc.latency_disp_cnt mod len) in
    draw_latency gfun sc drw gc zoom l be1 be2 lat erase
  else ()
;

value draw_caption gfun sc drw gc = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  gfun.g_clear_drawable drw;
  gfun.g_set_line_width sc.graph gc 2;
  let x = 0 in
  let w = 25 in
  let h = 15 in
  let font = F_raw in
  gfun.g_set_font sc.graph gc font;
  let sz = gfun.g_font_size sc.graph font in
  let i_start = ts2.schedule_len - ts2.lcm in
  for itask = 0 to Array.length ts.tasks_def - 1 do {
    let to_display =
      if sc.ts.schedule_len < disp_lim then do {
        loop i_start where rec loop i =
          if i = ts2.schedule_len then False
          else
            match Sarray.get ts2.disp_schedule i with
            [ Some {u_task = j} -> if j = itask then True else loop (i + 1)
            | None -> loop (i + 1) ]
      }
      else False
    in
    if to_display then
      if sc.coloured then
        gfun.g_set_color sc.graph gc (C_number (task_color sc itask))
      else
        gfun.g_set_pixmap sc.graph gc itask
    else gfun.g_set_color sc.graph gc C_style_bg;
    let y = 20 + (h + 10) * itask in
    gfun.g_fill_rectangle drw gc (x, y, w, h);
    if to_display then gfun.g_set_color sc.graph gc C_black else ();
    if not sc.coloured then do {
      gfun.g_set_line_width sc.graph gc 1;
      gfun.g_draw_rectangle drw gc (x, y, w, h);
    }
    else ();
    gfun.g_erase_draw_string drw gc (x + w + 10, y + h / 2 + sz * 9 / 20)
      (sprintf "t%d" (itask + 1));
  };
  let preemp_to_display =
    loop i_start where rec loop i =
      if i = ts2.schedule_len then False
      else
        match Sarray.get ts2.disp_schedule i with
        [ Some {u_preempt = True} -> True
        | Some _ | None -> loop (i + 1) ]
  in
  if preemp_to_display then gfun.g_set_color sc.graph gc C_black
  else gfun.g_set_color sc.graph gc C_style_bg;
  let y = 20 + (h + 10) * (Array.length ts.tasks_def) in
  gfun.g_fill_rectangle drw gc (x, y, w, h);
  if preemp_to_display then gfun.g_set_color sc.graph gc C_black else ();
  if not sc.coloured then do {
    gfun.g_set_line_width sc.graph gc 1;
    gfun.g_draw_rectangle drw gc (x, y, w, h);
  }
  else ();
  let font = F_raw in
  gfun.g_set_font sc.graph gc font;
  let sz = gfun.g_font_size sc.graph font in
  gfun.g_erase_draw_string drw gc (x + w + 10, y + h / 2 + sz / 2)
    (I18n.transl "preemption");
  gfun.g_commit_display drw;
};
