(* $Id: draw_resp_f_time.ml,v 1.47 2010-12-04 07:06:11 deraugla Exp $ *)

open Printf;
open Rt;

open Ast;
open Da;
open Data_rt;
open Dutil;
open Glob_glob;
open Dtasks;

value close_resp_f_time sc wid = do {
  sc.graph.resp_f_time :=
    List.filter (fun w -> compare w wid <> 0) sc.graph.resp_f_time;
  rt_destroy_widget wid;
};

value expose_resp_f_time xd gd info wid = do {
  let (response_values, n, min_rel_time, rmax, lcm, autorep, caption) =
    info
  in
  let drw = (gd.graph.dd, "resp_f_time", wid) in
  let width = da_drawable_width drw in
  let height = da_drawable_height drw in
  da_clear_widget drw;
  da_select_font gd.graph.raw_font;
  let wrt =
    List.fold_left
      (fun wrt (itask, list) ->
         List.fold_left (fun wrt (x, y) -> max wrt y) wrt list)
      1 response_values
  in
  (* adding two extra lcm *)
  let response_values =
    List.map
      (fun (itask, list) ->
         let ext =
           List.fold_right
             (fun (x, y) ext ->
                if x >= n - lcm then [(x, y) :: ext] else ext)
             list []
         in
         let ext =
           if ext = [] then
             if rmax = 0 then list
             else if autorep then
               match List.rev list with
               [ [(x, y) :: _] -> [(x + lcm, y)]
               | [] -> [] ]
             else []
           else ext
         in
         let list =
           list @
             List.map (fun (x, y) -> (x + lcm, y)) ext @
             List.map (fun (x, y) -> (x + 2 * lcm, y)) ext
         in
         (itask, list))
      response_values
  in
  (* but displaying only one extra one *)
  let min_x = 60 in
  let max_x = width - 60 in
  let min_y = 40 + List.length caption * da_font_size gd.graph.raw_font in
  let max_y = height - 60 in
  let n = n + lcm in
  List.iter
    (fun (itask, list) -> do {
       let lines =
         List.map
           (fun (x, y) ->
              (x * (max_x - min_x) / n + min_x,
               y * (max_y - min_y - 20) / wrt + min_y))
           list
       in
       let lines =
         List.map (fun (x, y) -> (x, max_y + min_y - y)) lines
       in
       da_set_line_width gd.graph.dd 2;
       if gd.coloured then da_select_color (task_color gd itask) else ();
       da_draw_lines drw lines;
       List.iter
         (fun (x, y) ->
            da_fill_arc drw (x - 3, y - 3, 6, 6, 0, Draw_dameid.grand_tour))
         lines;
     })
    response_values;
  (* and clipping the last one *)
  let x = n * (max_x - min_x) / n + min_x in
  da_clear_area drw (x, 0, width, height);
  (* tasks names caption *)
  let lim = n in
  let tl =
    List.map
      (fun (itask, list) ->
         let v =
           loop list where rec loop =
             fun
             [ [(x1, y1) :: ([(x2, y2) :: _] as list)] ->
                 if x1 <= lim && x2 > lim then Some (x1, y1, x2, y2)
                 else loop list
             | [(x, y)] -> Some (x, y, x, y)
             | [] -> None ]
         in
         (itask, v))
      response_values
  in
  da_select_color gd.graph.black_col;
  let x = max_x + 15 in
  List.iter
    (fun (itask, v) ->
       match v with
       [ Some (x1, y1, x2, y2) ->
           let y =
             if x2 <> x1 then
               max_y -
                 y1 * (max_y - min_y - 20) / wrt -
                 (lim - x1) * (y2 - y1) * (max_y - min_y - 20) / (x2 - x1) /
                   wrt
             else
               max_y - (max_y - min_y - 20) * y1 / wrt
           in
           let txt = sprintf "t%d" (itask + 1) in
           let y = y + (da_font_size gd.graph.raw_font) * 5 / 20 in
           da_erase_draw_string drw (x, y) txt
       | None -> () ])
     tl;
  (* axis *)
  da_select_color gd.graph.black_col;
  da_set_line_width gd.graph.dd 0;
  let x = max_x + 20 in
  let y = max_y in
  da_draw_line drw (min_x - 20, y) (x, y);
  da_draw_lines drw [(x - 5, y - 3); (x, y); (x - 5, y + 3)];
  let x = min_x in
  let y = min_y - 20 in
  da_draw_line drw (x, max_y + 20) (x, y);
  da_draw_lines drw [(x - 3, y + 5); (x, y); (x + 3, y + 5)];
  (* vertical dashes and graduation on X axis *)
  List.iter
    (fun (itask, list) -> do {
       let (xw, yw) =
         List.fold_left
           (fun (xw, yw) (x, y) -> if y > yw then (x, y) else (xw, yw))
           (min_int, 0) list
       in
       let x = xw * (max_x - min_x) / n + min_x in
       let y = max_y - yw * (max_y - min_y - 20) / wrt in
       da_set_line_width gd.graph.dd 0;
       da_draw_line drw (x, max_y - 3) (x, max_y + 3);
       da_set_dashes gd.graph.dd 0 [5; 5];
       da_draw_line drw (x, max_y) (x, y);
       let txt = string_of_int (xw + min_rel_time) in
       let x = x - da_text_width gd.graph.raw_font txt / 2 in
       let y = max_y + 20 in
       da_erase_draw_string drw (x, y) txt;
     })
    response_values;
  da_set_line_width gd.graph.dd 0;
  da_select_font gd.graph.raw_font;
  for i = 0 to 3 do {
    let x = (rmax + i * lcm) * (max_x - min_x) / n + min_x in
    da_draw_line drw (x, max_y + 3) (x, max_y - 3);
    let txt = string_of_int (rmax + min_rel_time + i * lcm) in
    let x = x - da_text_width gd.graph.raw_font txt / 2 in
    da_erase_draw_string drw (x, max_y + 20) txt;
  };
  (* graduation on Y axis *)
  List.iter
    (fun (itask, list) -> do {
       let (xw, yw) =
         List.fold_left
           (fun (xw, yw) (x, y) -> if y > yw then (x, y) else (xw, yw))
           (min_int, 0) list
       in
       let x = xw * (max_x - min_x) / n + min_x in
       let y = max_y - yw * (max_y - min_y - 20) / wrt in
       da_set_dashes gd.graph.dd 0 [5; 5];
       da_draw_line drw (min_x, y) (x, y);
       da_set_line_width gd.graph.dd 0;
       da_draw_line drw (min_x - 3, y) (min_x + 3, y);
       let txt = string_of_int yw in
       let x = min_x - 10 - da_text_width gd.graph.raw_font txt in
       let y = y + da_font_size gd.graph.raw_font / 2 in
       da_erase_draw_string drw (x, y) txt;
     })
    response_values;
  da_set_line_width gd.graph.dd 0;
  (* caption *)
  let max_len =
    List.fold_left
      (fun m txt -> max m (da_text_width gd.graph.raw_font txt)) 0
      caption
  in
  let _ =
    let x = width - max_len - da_text_width gd.graph.raw_font "m" in
    let sz = da_font_size gd.graph.raw_font in
    List.fold_left
      (fun y txt -> do {
         da_draw_string drw (x, y) txt;
         y + sz
       })
      sz caption
  in
  ()
};

value print_ps_resp_f_time xd sc info oc = do {
  let wid = widget_named xd "dameid_0" in
  let dd = sc.graph.dd in
  dd.postscript := Some oc;
  fprintf oc "%%!\n";
  let drw = (dd, "resp_f_time", wid) in
  let v1 = da_drawable_width drw in
  let v2 = da_drawable_height drw in
(*
  let (llx, lly) =
    if dd.ps_landscape then (0, 0)
    else if dd.ps_A3 then (0, v1)
    else (0, v2 / 2)
  in
  let (urx, ury) = if dd.ps_A3 then (v2, 2 * v1) else (v1, v2) in
  fprintf oc "%%%%BoundingBox: %d %d %d %d\n" llx lly urx ury;
*)
  let (llx, lly) = (30, 130) in
  let (urx, ury) = (v1, v2) in
  fprintf oc "%%%%BoundingBox: %d %d %d %d\n" llx lly urx ury;
(**)
  let scale = 0.65 in
(*
fprintf oc "gsave\n";
*)
  fprintf oc "\
/cm { 2.54 div 72 mul } def
0 10 cm translate
%.2f %.2f scale\n" scale (-. scale);
  Da.print_def_font_with_accents oc;
  expose_resp_f_time xd sc info wid;
(*
fprintf oc "grestore\n";
fprintf oc "%d %d %d %d\n" llx lly urx ury;
fprintf oc "/ury exch 5 sub def /urx exch 5 sub def\n";
fprintf oc "/lly exch 5 add def /llx exch 5 add def\n";
fprintf oc "\
1 setlinewidth
llx lly moveto llx ury lineto urx ury lineto urx lly lineto closepath
stroke\n";
*)
  fprintf oc "showpage\n";
  dd.postscript := None;
  flush oc;
};

value action_resp_f_time xd gd info wid =
  fun
  [ RawEvExpose _ _ _ _ | RawEvConfigureNotify _ _ ->
      expose_resp_f_time xd gd info wid
  | _ -> () ]
;

value filter_escape_sequences s =
  loop [] 0 where rec loop rev_cl i =
    if i = String.length s then rev_implode rev_cl
    else
      match s.[i] with
      [ '\027' ->
          let i =
            loop (i + 1) where rec loop i =
              if i = String.length s then i
              else
                match s.[i] with
                [ 'a'..'z' | 'A'..'Z' -> i + 1
                | _ -> loop (i + 1) ]
          in
          loop rev_cl i
      | c -> loop [c :: rev_cl] (i + 1) ]
;

value response_f_time gd xd ul_xy (x, y) =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    gd.graph_nb := gd.graph_nb + 1;
    let name = sprintf "main_resp_f_time_%d" gd.graph_nb in
    let wid =
      let tit =
        I18n.transl "Response" ^ " = f (" ^ I18n.transl "time" ^ ")"
      in
      let response_values = Schedule.get_response_values ts2 in
      let n = ts2.schedule_len in
      let w = widget_width (Dutil.g_widget_named gd "schedule") / 2 in
      let h = widget_height (Dutil.g_widget_named gd "schedule") * 2 in
      let (x, y) =
        match ul_xy with
        [ Some (x, y) -> (x - w / 2, y - h / 2)
        | None -> (x, y) ]
      in
      let constr_caption =
        let (list1, list2, list3) =
          Schedule.constr_info_list Schedule.vt_term ts2 max_int
        in
        let list4 =
          let ended_and_schedulable =
            let ended =
              Array.length ts.tasks_def > 0 &&
              ts2.curr_ord_ind = Array.length ts.tasks_def ||
              ts2.schedule_status <> SS_running
            in
            ended &&
            match ts2.schedule_status with
            [ SS_running | SS_schedulable_but _ | SS_several_lcm _ -> True
            | SS_task_error _ _ | SS_incomplete_scheduling _ _ -> False ]
          in
          if ended_and_schedulable then
            let cuf = Schedule.compute_classical_utilization_factor ts2 in
            let exact_utilization = Schedule.compute_exact_utilization ts2 in
            let eppc = float (exact_utilization - cuf) /. float ts2.lcm in
            [sprintf "%s%s %1.4f"
               (I18n.transl "exact permanent preemption cost")
               (I18n.transl ":") eppc]
          else []
        in
        List.map filter_escape_sequences (list1 @ list2 @ list3 @ list4)
      in
      let info =
        (response_values, n, ts2.min_rel_time, ts2.rmax,
         ts2.lcm, False, constr_caption)
      in
      rt_create_widget xd tit tit (UserPosition x y)
        (Some (close_resp_f_time sc))
        (pack_desc [NameAtt name]
           (Vertical,
            [raw_desc [FillerAtt]
               (w, h, 0, [SelExposure; SelStructureNotify])
               (action_resp_f_time xd sc info)])
           (fun wid ->
              fun
              [ PackEvKeyPress {item = k} ->
                  match k with
                  [ K_Escape -> close_resp_f_time sc wid
                  | K_Ascii 'g' ->
                      let wid = widget_named xd name in
                      Save.print_ps xd wid gd "graph.ps"
                        (print_ps_resp_f_time xd sc info)
                        (fun () -> rt_dont_redirect_key_press_from wid)
                        (Some (x + w / 2, y + h / 2)) (0, 0)
                  | _ -> () ] ]))
    in
    sc.graph.resp_f_time := [wid :: sc.graph.resp_f_time];
    rt_dont_redirect_key_press_from wid;
    rt_map_widget wid;
  }
  else ()
;
