(* $Id: draw_schedule.ml,v 1.103 2011-02-08 10:52:42 deraugla Exp $ *)

open Ast;
open Dtasks;
open Glob_glob;
open Printf;

value left_sched = 65;
value right_sched width = width - 35;

value sched_x gfun sc drw zoom width j =
  match zoom with
  [ Zoom zi _ dx dy ->
      let n = zi.zi_n in
      let x = zi.zi_x in
      let (w, _) = gfun.g_schedule_zoom_size sc.graph zoom in
      let x =
        Sutil.mul_div (j * (right_sched width - left_sched)) n
          sc.ts.schedule_len -
        dx - n * x + n * left_sched + w / 2
      in
      max (-32000) (min 32000 x)
  | NoZoom ->
      left_sched + j * (right_sched width - left_sched) / sc.ts.schedule_len ]
;

value sched_y gfun sc drw zoom itask =
  match zoom with
  [ Zoom zi _ dx dy ->
      let y = zi.zi_y in
      let (_, h) = gfun.g_schedule_zoom_size sc.graph zoom in
      let sy = gfun.g_inter_task_dy drw * (itask + 1) in
      sy - (y + dy) + h / 2
  | NoZoom ->
      gfun.g_inter_task_dy drw * (itask + 1) ]
;

value draw_repetition gfun sc drw gc zoom = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  match ts2.repetition with
  [ Some (i, len) -> do {
      let width = gfun.g_drawable_width drw in
      let y1 = sched_y gfun sc drw zoom 0 - 40 in
      let y2 = sched_y gfun sc drw zoom (Array.length ts.tasks_def) + 30 in
      gfun.g_set_color sc.graph gc C_backg_rep_col;
      let x1 = sched_x gfun sc drw zoom width i in
      let x2 = sched_x gfun sc drw zoom width (i + len) in
      gfun.g_fill_rectangle drw gc (x1, y1, x2 - x1, y2 - y1);
    }
  | None -> () ]
};

value draw_colored_unit gfun sc drw gc zoom itask j = do {
  let width = gfun.g_drawable_width drw in
  let x1 = sched_x gfun sc drw zoom width j in
  let x2 = sched_x gfun sc drw zoom width (j + 1) in
  let y1 = sched_y gfun sc drw zoom itask - 15 in
  let y2 = sched_y gfun sc drw zoom itask in
  gfun.g_fill_rectangle drw gc (x1, y1, x2 - x1, y2 - y1)
};

value draw_unit gfun sc drw gc zoom color itask j =
  if sc.ts.schedule_len < Draw_dameid.disp_lim then do {
    gfun.g_set_color sc.graph gc color;
    draw_colored_unit gfun sc drw gc zoom itask j
  }
  else ()
;

value compute_min_j gfun sc drw zoom =
  match zoom with
  [ Zoom zi wid dx dy -> do {
      let n = zi.zi_n in
      let x = zi.zi_x in
      let width = gfun.g_drawable_width drw in
      let (wid_width, _) = gfun.g_schedule_zoom_size sc.graph zoom in
      int_of_float
        ((float x -. float left_sched +.
              (float dx -. float wid_width /. 2.0) /. float n) *.
          (float sc.ts.schedule_len) /.
              (float (right_sched width) -. float left_sched))
    }
  | NoZoom -> 0 ]
;

value compute_max_j gfun sc drw zoom =
  match zoom with
  [ Zoom zi wid dx dy ->
      let n = zi.zi_n in
      let x = zi.zi_x in
      let width = gfun.g_drawable_width drw in
      let (wid_width, _) = gfun.g_schedule_zoom_size sc.graph zoom in
      int_of_float
        ((float x -. float left_sched +.
              (float dx +. float wid_width /. 2.0) /. float n) *.
          (float sc.ts.schedule_len) /.
              (float (right_sched width) -. float left_sched)) + 1
  | NoZoom -> sc.ts.schedule_len ]
;

value draw_units gfun sc drw gc zoom = do {
  if sc.ts.schedule_len < Draw_dameid.disp_lim then do {
    let (wid_width, _) = gfun.g_schedule_zoom_size sc.graph zoom in
    let width = gfun.g_drawable_width drw in
    let min_j = compute_min_j gfun sc drw zoom in
    let max_j = compute_max_j gfun sc drw zoom in
    for i = 0 to Array.length sc.ts.ast.tasks_def - 1 do {
      let color =
        if not sc.coloured then C_gray
        else C_number (Draw_dameid.task_color sc i)
      in
      gfun.g_set_color sc.graph gc color;
      let min_j = max min_j 0 in
      let max_j = min max_j (sc.ts.schedule_len - 1) in
      for j = min_j to max_j do {
        match Sarray.get sc.ts.disp_schedule j with
        [ Some {u_task = itask; u_preempt = preempt} ->
            if itask = i && not preempt then
              let x1 = sched_x gfun sc drw zoom width j in
              let x2 = sched_x gfun sc drw zoom width (j + 1) in
              if zoom = NoZoom || x2 >= 0 && x1 <= wid_width then
                draw_colored_unit gfun sc drw gc zoom itask j
              else ()
            else ()
        | None -> () ];
      };
    };
    let min_j = max min_j 0 in
    let max_j = min max_j (sc.ts.schedule_len - 1) in
    gfun.g_set_color sc.graph gc C_black;
    for j = min_j to max_j do {
      match Sarray.get sc.ts.disp_schedule j with
      [ Some {u_task = itask; u_preempt = preempt} ->
          if preempt then
            let x1 = sched_x gfun sc drw zoom width j in
            let x2 = sched_x gfun sc drw zoom width (j + 1) in
            if zoom = NoZoom || x2 >= 0 && x1 <= wid_width then
              draw_colored_unit gfun sc drw gc zoom itask j
            else ()
          else ()
      | None -> () ];
    };
  }
  else ()
};

value draw_task_names gfun sc drw gc zoom = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let width = gfun.g_drawable_width drw in
  for i = 0 to Array.length ts.tasks_def - 1 do {
    let x = sched_x gfun sc drw zoom width 0 - 40 in
    let y = sched_y gfun sc drw zoom i in
    let color =
      if sc.coloured then C_number (Draw_dameid.task_color sc i)
      else C_black
    in
    gfun.g_set_color sc.graph gc color;
    gfun.g_erase_draw_string drw gc (x, y) (sprintf "t%d" (i + 1));
  };
  gfun.g_set_color sc.graph gc C_black;
};

value draw_skeleton gfun sc drw gc zoom = do {
  if sc.ts.schedule_len < Draw_dameid.disp_lim then do {
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let (wid_width, _) = gfun.g_schedule_zoom_size sc.graph zoom in
    let width = gfun.g_drawable_width drw in
    let n = ts2.schedule_len in
    gfun.g_set_font sc.graph gc F_raw;
    gfun.g_set_color sc.graph gc C_black;
    gfun.g_set_line_width sc.graph gc 0;
    (* black horizontal lines *)
    for i = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(i) in
      let x1 =
        sched_x gfun sc drw zoom width (td.release - ts2.min_rel_time)
      in
      let x2 = sched_x gfun sc drw zoom width n in
      let y = sched_y gfun sc drw zoom i in
      gfun.g_draw_line drw gc (x1, y) (x2, y);
    };
    gfun.g_set_line_width sc.graph gc 0;
    let zoom_n =
      match zoom with
      [ Zoom zi _ _ _ -> zi.zi_n
      | NoZoom -> 1 ]
    in
    let interv_size =
      zoom_n * (right_sched width - left_sched) / sc.ts.schedule_len
    in
    let min_j = compute_min_j gfun sc drw zoom in
    let max_j = compute_max_j gfun sc drw zoom in
    (* black vertical graduations and deadlines notches *)
    for i = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(i) in
      let pp =
        let step = (n - td.release + ts2.min_rel_time) / td.period in
        if step < 40 * zoom_n then td.period
        else td.period * (step / (40 * zoom_n) + 1)
      in
      let y = sched_y gfun sc drw zoom i in
      let y1 = y - 10 in
      let min_j = max min_j (td.release - ts2.min_rel_time) in
      let max_j = min max_j n in
      for j = min_j to max_j do {
        let x = sched_x gfun sc drw zoom width j in
        if zoom = NoZoom || x >= 0 && x <= wid_width then do {
          let k = (j - (td.release - ts2.min_rel_time)) mod pp in
          if k = 0 || interv_size <= 5 then ()
          else do {
            gfun.g_draw_line drw gc (x, y1) (x, y);
            if k = td.deadline then do {
              gfun.g_draw_line drw gc (x, y + 5) (x, y + 10);
              gfun.g_draw_line drw gc (x - 5, y + 10) (x, y + 10);
            }
            else ()
          }
        }
        else ();
      };
    };
    (* LCM starts vertical dashes *)
    if ts2.rmax = 0 then ()
    else do {
      let y1 = sched_y gfun sc drw zoom 0 - 40 in
      let y2 = sched_y gfun sc drw zoom (Array.length ts.tasks_def) + 10 in
      gfun.g_set_line_width sc.graph gc 0;
      gfun.g_set_dashes sc.graph gc 0 [5; 5];
      loop ts2.rmax where rec loop i =
        if i > n then ()
        else do {
          let x = sched_x gfun sc drw zoom width i in
          gfun.g_draw_line drw gc (x, y1) (x, y2);
          loop (i + ts2.lcm)
        };
      gfun.g_set_line_width sc.graph gc 0;
    };
    (* coloured vertical graduations at periods *)
    gfun.g_set_line_width sc.graph gc 2;
    for i = 0 to Array.length ts.tasks_def - 1 do {
      gfun.g_set_color sc.graph gc
        (if sc.coloured then C_number (Draw_dameid.task_color sc i)
         else C_black);
      let td = ts.tasks_def.(i) in
      let pp =
        let step = (n - td.release + ts2.min_rel_time) / td.period in
        if step < 40 * zoom_n then td.period
        else td.period * (step / (40 * zoom_n) + 1)
      in
      let y = sched_y gfun sc drw zoom i in
      let y1 = y - 10 in
      let min_j = max min_j (td.release - ts2.min_rel_time) in
      let max_j = min max_j n in
      for j = min_j to max_j do {
        let x = sched_x gfun sc drw zoom width j in
        if zoom = NoZoom || x >= 0 && x <= wid_width then
          let k = (j - (td.release - ts2.min_rel_time)) mod pp in
          if k = 0 then gfun.g_draw_line drw gc (x, y1 - 10) (x, y) else ()
        else ();
      };
    };
    (* coloured vertical graduations at gaps *)
    gfun.g_set_line_width sc.graph gc 2;
    for i = 0 to Array.length ts.tasks_def - 1 do {
      gfun.g_set_color sc.graph gc
        (if sc.coloured then C_number (Draw_dameid.task_color sc i)
         else C_black);
      let td = ts.tasks_def.(i) in
      let pp =
        let step = (n - td.release + ts2.min_rel_time) / td.period in
        if step < 40 * zoom_n then td.period
        else td.period * (step / (40 * zoom_n) + 1)
      in
      let y = sched_y gfun sc drw zoom i in
      let y1 = y - 10 in
      let min_j = max min_j (td.release - ts2.min_rel_time) in
      let max_j = min max_j n in
      for j = min_j to max_j do {
        let x = sched_x gfun sc drw zoom width j in
        if zoom = NoZoom || x >= 0 && x <= wid_width then
          let k = (j - (td.release - ts2.min_rel_time)) mod pp in
          List.iter
            (fun sd ->
               if k = sd.sub_rel then gfun.g_draw_line drw gc (x, y1) (x, y+5)
               else ())
            td.sub_duration
        else ();
      };
    };
    (* numberings *)
    gfun.g_set_color sc.graph gc C_black;
    for i = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(i) in
      let y = sched_y gfun sc drw zoom i in
      let y1 = y - 10 in
      let pp =
        let step = (n - td.release + ts2.min_rel_time) / td.period in
        if step < 30 * zoom_n then td.period
        else td.period * (step / (40 * zoom_n) + 1) * 2
      in
      let min_j = max min_j (td.release - ts2.min_rel_time) in
      let max_j = min max_j n in
      for j = min_j to max_j do {
        let x = sched_x gfun sc drw zoom width j in
        if zoom = NoZoom || x >= 0 && x <= wid_width then
          let k = (j - (td.release - ts2.min_rel_time)) mod pp in
          if k = 0 || zoom <> NoZoom && interv_size > 40 then do {
            let text = string_of_int (j + ts2.min_rel_time) in
            let x0 = gfun.g_text_width sc.graph F_raw text / 2 in
            gfun.g_draw_string drw gc (x - x0, y1 + 25) text;
          }
          else ()
        else ();
      };
    };
    if zoom = NoZoom then draw_task_names gfun sc drw gc zoom else ()
  }
  else ()
};

type multi_font_text α =
  [ Norm of α and string
  | UpDown of α and string and string ]
;

value draw_latency gfun sc drw gc zoom l (b1, e1) (b2, e2) lat erase = do {
  let ts = sc.ts in
  let lc = l.l_constr in
  let n =
    match ts.repetition with
    [ Some (n, _) -> n + ts.lcm
    | None -> ts.schedule_len ]
  in
  let width = gfun.g_drawable_width drw in
  if erase then gfun.g_set_color sc.graph gc C_style_bg
  else if lat > lc.l_max then
    if sc.coloured then gfun.g_set_color sc.graph gc C_red
    else gfun.g_set_color sc.graph gc C_gray
  else
    gfun.g_set_color sc.graph gc C_black;
  let b1_bef_n =
    if b1 < n then b1 else b1 - (b1 - n + ts.lcm) / ts.lcm * ts.lcm
  in
  let e2_bef_n =
    if e2 <= n then e2 else e2 - (e2 - n + ts.lcm - 1) / ts.lcm * ts.lcm
  in
  let nb_tasks = Array.length ts.ast.tasks_def in
  let draw_line itask j dy =
    let x = sched_x gfun sc drw zoom width j in
    let y1 = sched_y gfun sc drw zoom (nb_tasks - 1) + 50 - dy in
    let y2 = sched_y gfun sc drw zoom itask - 20 in
    gfun.g_draw_line drw gc (x, y1) (x, y2)
  in
  gfun.g_set_line_width sc.graph gc 0;
  gfun.g_set_dashes sc.graph gc 0 [5; 5];
  draw_line lc.l_task1 b1_bef_n 0;
  draw_line lc.l_task2 e2_bef_n 0;
  gfun.g_set_line_width sc.graph gc 2;
  let x = sched_x gfun sc drw zoom width b1_bef_n in
  let y = sched_y gfun sc drw zoom (nb_tasks - 1) + 25 in
  let rd = 4 in
  gfun.g_fill_arc drw gc (x - rd, y - rd, 2 * rd, 2 * rd)
    (0, Draw_dameid.grand_tour);

  let section1 = if b1 < n then 0 else (b1 - n + ts.lcm) / ts.lcm in
  let section2 = if e2 - 1 < n then 0 else (e2 - 1 - n + ts.lcm) / ts.lcm in

  let td1 = ts.ast.tasks_def.(lc.l_task1) in
  let td2 = ts.ast.tasks_def.(lc.l_task2) in
  let inst1_sh =
    (b1_bef_n + ts.min_rel_time - td1.release) / td1.period + 1
  in
  let inst2_sh =
    (e2 - (b1 - b1_bef_n) + ts.min_rel_time - td2.release - 1) / td2.period +
    1
  in

  let b1 = b1_bef_n in
  let e2 = e2_bef_n in

  if section2 = section1 then do {
    let x1 = sched_x gfun sc drw zoom width b1 in
    let x2 = sched_x gfun sc drw zoom width e2 in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    gfun.g_draw_lines drw gc [(x2 - 10, y - 5); (x2, y - 1); (x2 - 10, y + 5)]
  }
  else if section2 = section1 + 1 then do {
    let x1 = sched_x gfun sc drw zoom width b1 in
    let x2 = sched_x gfun sc drw zoom width n in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    let y = y + 10 in
    let x1 = sched_x gfun sc drw zoom width (n - ts.lcm) in
    let x2 = sched_x gfun sc drw zoom width e2 in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    gfun.g_draw_lines drw gc [(x2 - 10, y - 5); (x2, y - 1); (x2 - 10, y + 5)]
  }
  else if section2 = section1 + 2 then do {
    let x1 = sched_x gfun sc drw zoom width b1 in
    let x2 = sched_x gfun sc drw zoom width n in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    let y = y + 10 in
    let x1 = sched_x gfun sc drw zoom width (n - ts.lcm) in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    let y = y + 10 in
    let x2 = sched_x gfun sc drw zoom width e2 in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    gfun.g_draw_lines drw gc [(x2 - 10, y - 5); (x2, y - 1); (x2 - 10, y + 5)]
  }
  else if section2 > section1 + 2 then do {
    let x1 = sched_x gfun sc drw zoom width b1 in
    let x2 = sched_x gfun sc drw zoom width n in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    let y = y + 10 in
    let x1 = sched_x gfun sc drw zoom width (n - ts.lcm) in
    let dash_i = min 100 (max 30 ((x2 - x1) / 3)) in
    gfun.g_draw_line drw gc (x1, y) ((x1 + x2 - dash_i) / 2, y);
    gfun.g_draw_line drw gc ((x1 + x2 + dash_i) / 2, y) (x2, y);
    gfun.g_set_dashes sc.graph gc 0 [1; 3];
    gfun.g_draw_line drw gc ((x1 + x2 - dash_i) / 2, y)
      ((x1 + x2 + dash_i) / 2, y);
    gfun.g_set_line_width sc.graph gc 2;
    let y = y + 10 in
    let x2 = sched_x gfun sc drw zoom width e2 in
    gfun.g_draw_line drw gc (x1, y) (x2, y);
    gfun.g_draw_lines drw gc [(x2 - 10, y - 5); (x2, y - 1); (x2 - 10, y + 5)]
  }
  else ();

  let x1 = sched_x gfun sc drw zoom width b1 in
  let x2 = sched_x gfun sc drw zoom width e2 in

  let y = sched_y gfun sc drw zoom (Array.length ts.ast.tasks_def - 1) + 25 in
  let (xx, y) =
    let lim =
      if ts.rmax = 0 then 0
      else
        match ts.repetition with
        [ Some (i, _) -> i
        | None -> ts.rmax + ts.lcm ]
    in
    if section1 = section2 then (x1 + x2, y)
    else
      let d = if section2 = section1 + 1 then 0 else 10 in
      let len1 = min b1 e2 - lim in
      let len3 = n - max b1 e2 in
      let len2 =
        if b1 < e2 then ts.lcm - len1 - len3
        else if section2 >= section1 + 1 then b1 - e2
        else len3
      in
      if len1 > max len2 len3 then
        (sched_x gfun sc drw zoom width lim + min x1 x2, y + 10 + d)
      else if len2 >= max len1 len3 then
        if b1 > e2 then (x1 + x2, y + (*20 +*) d)
        else (x1 + x2, y + 10 + d)
      else
        (max x1 x2 + sched_x gfun sc drw zoom width n, y + d)
  in

  let dy = gfun.g_font_size sc.graph F_bigger + 2 in

  let multi_font_text =
    [Norm F_bigger "L";
     UpDown F_smaller (sprintf " %d,%d" lc.l_inst1 lc.l_inst2)
       (sprintf " %d,%d" (lc.l_task1 + 1) (lc.l_task2 + 1))]
  in
  let multi_font_text =
    if inst1_sh <> lc.l_inst1 || inst2_sh <> lc.l_inst2 then
      multi_font_text @
        [Norm F_raw " (= ";
         Norm F_bigger "L";
         UpDown F_smaller (sprintf " %d,%d" inst1_sh inst2_sh)
           (sprintf " %d,%d" (lc.l_task1 + 1) (lc.l_task2 + 1));
         Norm F_raw ")"]
    else multi_font_text
  in
  let multi_font_text =
    multi_font_text @ [Norm F_raw (sprintf " = %d" lat)]
  in
  let tot_len =
    List.fold_left
      (fun len ->
         fun
         [ Norm font txt -> len + gfun.g_text_width sc.graph font txt
         | UpDown font txt1 txt2 ->
             len + max (gfun.g_text_width sc.graph font txt1)
               (gfun.g_text_width sc.graph font txt2) ])
      0 multi_font_text
  in
  let x = (xx - tot_len) / 2 in
  let x =
    match zoom with
    [ Zoom _ _ _ _ -> x
    | NoZoom -> min x (sched_x gfun sc drw zoom width n - tot_len) ]
  in
  let _ =
    List.fold_left
      (fun x ->
         fun
         [ Norm font txt -> do {
             gfun.g_set_font sc.graph gc font;
             gfun.g_draw_string drw gc (x, y + dy) txt;
             x + gfun.g_text_width sc.graph font txt
           }
         | UpDown font txt1 txt2 -> do {
             gfun.g_set_font sc.graph gc font;
             gfun.g_draw_string drw gc (x, y + dy - 8) txt1;
             gfun.g_draw_string drw gc (x, y + dy + 4) txt2;
             x +
               max (gfun.g_text_width sc.graph font txt1)
                 (gfun.g_text_width sc.graph font txt2)
           } ])
       x multi_font_text
  in
  ()
};

value draw_latencies gfun sc drw gc zoom erase =
  let ok_lat =
    List.fold_right
      (fun l ok_lat ->
         match l.l_status with
         [ LS_values be1 be2 lat -> [(l, be1, be2, lat) :: ok_lat]
         | _ -> ok_lat ])
      sc.ts.latency_constr []
  in
  let len = List.length ok_lat in
  if len > 0 then
    let (l, be1, be2, lat) = List.nth ok_lat (sc.latency_disp_cnt mod len) in
    draw_latency gfun sc drw gc zoom l be1 be2 lat erase
  else ()
;

value draw_schedule gfun sc drw gc = do {
  let zoom = NoZoom in
  gfun.g_clear_drawable drw;
  if Array.length sc.ts.ast.tasks_def > 0 then do {
    draw_repetition gfun sc drw gc zoom;
    draw_units gfun sc drw gc zoom;
    draw_skeleton gfun sc drw gc zoom;
(*
    draw_schedule_instance_separators sc sched_wid False;
*)
    draw_latencies gfun sc drw gc zoom False;
 }
  else ();
  gfun.g_commit_display drw;
};

(* mesoid *)

value draw_mesoid_unit gfun sc drw gc color j = do {
  let width = gfun.g_drawable_width drw in
  let x1 = sched_x gfun sc drw NoZoom width j in
  let x2 = sched_x gfun sc drw NoZoom width (j + 1) in
  let y1 = 40 * (1) - 15 in
  let y2 = 40 * (1) in
  gfun.g_set_color sc.graph gc color;
  gfun.g_fill_rectangle drw gc (x1, y1, x2 - x1, y2 - y1)
};

value draw_mesoid_units gfun sc drw gc = do {
  if sc.ts.schedule_len < Draw_dameid.disp_lim then do {
    let ts2 = sc.ts in
    for j = 0 to ts2.schedule_len - 1 do {
      match Sarray.get ts2.disp_schedule j with
      [ Some {u_task = itask; u_preempt = preempt} ->
          let color =
            if preempt then C_black
            else if not sc.coloured then C_gray
            else C_number (Draw_dameid.task_color sc itask)
          in
          draw_mesoid_unit gfun sc drw gc color j
      | None -> () ];
    };
  }
  else ();
};

value draw_mesoid_skeleton gfun sc drw gc = do {
  let ts2 = sc.ts in
  let width = gfun.g_drawable_width drw in
  let n = ts2.schedule_len in
  let x1 = sched_x gfun sc drw NoZoom width 0 in
  let x2 = sched_x gfun sc drw NoZoom width n in
  let y = 40 * (1) in
  gfun.g_set_color sc.graph gc C_black;
  gfun.g_set_line_width sc.graph gc 0;
  gfun.g_draw_line drw gc (x1, y) (x2, y);
  let y1 = y - 10 in
  if n <= 200 then
    for j = 0 to n do {
      let x = sched_x gfun sc drw NoZoom width j in
      gfun.g_set_line_width sc.graph gc 0;
      gfun.g_set_color sc.graph gc C_black;
      gfun.g_draw_line drw gc (x, y1) (x, y);
    }
  else ();
  if ts2.rmax = 0 then ()
  else do {
    let y1 = 0 in
    let y2 = 100 in
    gfun.g_set_color sc.graph gc C_black;
    gfun.g_set_line_width sc.graph gc 0;
    gfun.g_set_dashes sc.graph gc 0 [5; 5];
    let x = sched_x gfun sc drw NoZoom width ts2.rmax in
    gfun.g_draw_line drw gc (x, y1) (x, y2);
    let x = sched_x gfun sc drw NoZoom width (ts2.rmax + ts2.lcm) in
    gfun.g_draw_line drw gc (x, y1) (x, y2);
    let x = sched_x gfun sc drw NoZoom width n in
    gfun.g_draw_line drw gc (x, y1) (x, y2);
    gfun.g_set_line_width sc.graph gc 0;
  };
  let draw_pos j = do {
    let x = sched_x gfun sc drw NoZoom width j in
    let text = string_of_int (j + ts2.min_rel_time) in
    let x0 = gfun.g_text_width sc.graph F_raw text / 2 in
    gfun.g_erase_draw_string drw gc (x - x0, y1 + 25) text;
  }
  in
  gfun.g_set_color sc.graph gc C_black;
  gfun.g_set_font sc.graph gc F_raw;
  draw_pos 0;
  if ts2.min_rel_time < 0 then draw_pos (- ts2.min_rel_time) else ();
  if ts2.rmax > 0 then do {
    draw_pos ts2.rmax;
    draw_pos (ts2.rmax + ts2.lcm);
  }
  else ();
  draw_pos n;
};

value draw_mesoid gfun sc gc drw = do {
  draw_mesoid_units gfun sc drw gc;
  draw_mesoid_skeleton gfun sc drw gc;
};

value gen_update_draw gfun sc gc dam_drw sched_drw mes_drw
    update_draw_zooms = do {
  let ts = sc.ts in
  let sh_dam = if ts.rmax = 0 then 0 else ts.rmax + ts.lcm in
  for j = 0 to ts.schedule_len - 1 do {
    if Sarray.get ts.comp_schedule j <> Sarray.get ts.disp_schedule j
    then do {
      let isect = (j + ts.min_rel_time) mod ts.lcm in
      let isect = if isect < 0 then isect + ts.lcm else isect in
      match Sarray.get ts.disp_schedule j with
      [ Some {u_task = itask} -> do {
          let color = C_style_bg in
          draw_unit gfun sc sched_drw gc NoZoom color itask j;
(*
          draw_mesoid_unit gfun sc mes_drw gc color j;
*)
          if j >= sh_dam then do {
            Draw_dameid.draw_sector gfun sc dam_drw gc NoZoom isect
          }
          else ();
          update_draw_zooms color itask j;
        }
      | None -> () ];
      match Sarray.get ts.comp_schedule j with
      [ Some {u_task = itask; u_preempt = preempt} -> do {
          let color =
            if preempt then C_black
            else if sc.coloured then
              C_number (Draw_dameid.task_color sc itask)
            else C_gray
          in
          draw_unit gfun sc sched_drw gc NoZoom color itask j;
(*
          draw_mesoid_unit gfun sc mes_drw gc color j;
*)
          if j >= sh_dam then
            Draw_dameid.draw_sector gfun sc dam_drw gc NoZoom isect
          else ();
          update_draw_zooms color itask j;
        }
      | None -> () ];
      Sarray.set ts.disp_schedule j (Sarray.get ts.comp_schedule j);
    }
    else ();
  };
  draw_latencies gfun sc sched_drw gc NoZoom False;
};
