(* $Id: expose.ml,v 1.100 2011-02-08 10:52:42 deraugla Exp $ *)

open Da;
open Data_rt;
open Glob_glob;
open Rt;
open Std_term;

(* Draw text *)

value redraw_text sc text_wid = do {
  let print = da_term_send sc.graph.dd text_wid in
  let ncols = snd (term_size text_wid) in
  print vt_no_auto_wrap;
  print vt_hide_cursor;
  print vt_bold;
  Schedule.gen_redraw_text print Schedule.vt_term ncols sc;
  print vt_erase_from_cursor;
};

value draw_text gd text_wid = do {
(*
  let (nb_row, _) = term_size text_wid in
  da_term_send gd.graph.dd text_wid
    (vt_goto nb_row 1 ^ "\n" ^ vt_erase_to_cursor);
*)
  da_term_send gd.graph.dd text_wid (vt_goto 1 1 ^ vt_erase_to_cursor);
(**)
  redraw_text gd text_wid
};

(* Expose zoom *)

value unfreeze_d wid zi =
  match zi.zi_d with
  [ Some d -> d
  | None -> do {
      let (dx, dy, _, _, _) = rt_query_pointer wid in
      let dx = widget_width wid / 2 - dx in
      let dy = widget_height wid / 2 - dy in
      zi.zi_d := Some (dx, dy);
      (dx, dy)
    } ]
;

value expose_zoom_schedule tg sc zi wid = do {
  let gfun = tg.g_fun in
  let (dx, dy) = unfreeze_d wid zi in
  let (dx2, dy2) = zi.zi_d2 in
  let zoom = Zoom zi wid (dx + dx2) (dy + dy2) in
  let drw = Draw_schedule_rt.schedule_drawable_of_zoom wid sc.graph.dd zoom in
  let gc = ref DaNonePn in
  Draw_schedule.draw_repetition gfun sc drw gc zoom;
  Draw_schedule.draw_units gfun sc drw gc zoom;
  Draw_schedule.draw_skeleton gfun sc drw gc zoom;
  Draw_schedule.draw_latencies gfun sc drw gc zoom False;

  let x = widget_width wid / 2 - dx in
  let y = widget_height wid / 2 - dy in
  let r = 5 in
  da_select_color sc.graph.black_col;
  da_set_line_width sc.graph.dd 1;
  rt_draw_arc (WidgetDr wid) (x - r, y - r, 2 * r, 2 * r, 0, 360 * 64);
};

value expose_zoom_dameid tg sc zi wid = do {
  let (dx, dy) = unfreeze_d wid zi in
  let (dx2, dy2) = zi.zi_d2 in
  let zoom = Zoom zi wid (dx + dx2) (dy + dy2) in
(*
printf "x %d dx %d nx+dx %d\n" zi.zi_x (dx + dx2) (n * zi.zi_x + dx + dx2); flush stdout;
*)
  let drw = Draw_dameid_rt.dameid_drawable_of_zoom wid sc.graph.dd zoom in
  let gc = ref DaNonePn in
  Draw_dameid.draw_sectors tg.g_fun sc sc.graph drw gc zoom;
  Draw_dameid.draw_skeleton tg.g_fun sc drw gc zoom;
(*
  Draw_dameid.draw_dameid_latencies zoom sc wid False;
*)
};

(* Expose *)

value expose tg = do {
  let gfun = tg.g_fun in
  let gc = ref DaNonePn in
  let sc = tg.scenario.(tg.curr_sc) in
  let dam_wid = Dutil.g_widget_named tg "dameid" in
  let cap_wid = Dutil.g_widget_named tg "caption" in
  let sched_wid = Dutil.g_widget_named tg "schedule" in
  let mes_wid = Dutil.g_widget_named tg "mesoid" in
  let dam_drw =
    Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
  in
  let cap_drw = Draw_dameid_rt.caption_drawable cap_wid sc.graph.dd in
  let sched_drw =
    Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd NoZoom
  in
  let mes_drw = Draw_schedule_rt.mesoid_drawable mes_wid sc.graph.dd in

  gfun.g_clear_drawable dam_drw;
  Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
  Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
  Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom False;
  gfun.g_commit_display dam_drw;
  Draw_dameid.draw_caption gfun sc cap_drw gc;
  Draw_schedule.draw_schedule gfun sc sched_drw gc;
  Draw_schedule.draw_mesoid gfun sc gc mes_drw;
  List.iter
    (fun (_, wid, zi) -> do {
       rt_clear_widget wid;
       expose_zoom_schedule tg sc zi wid;
     })
    sc.graph.zooms_sch;
  List.iter
    (fun (_, wid, zi) -> do {
       rt_clear_widget wid;
       expose_zoom_dameid tg sc zi wid;
     })
    sc.graph.zooms_dam;
};

value redisplay gd dam_wid = do {
  let cap_wid = Dutil.g_widget_named gd "caption" in
  let sched_wid = Dutil.g_widget_named gd "schedule" in
  let mes_wid = Dutil.g_widget_named gd "mesoid" in
  rt_clear_widget dam_wid;
  rt_clear_widget cap_wid;
  rt_clear_widget sched_wid;
  rt_clear_widget mes_wid;
  expose gd;
};

value redisplay_all gd =
  if gd.curr_sc < Array.length gd.scenario then do {
    let dam_wid = Dutil.g_widget_named gd "dameid" in
    let sc = gd.scenario.(gd.curr_sc) in
    redisplay gd dam_wid;
    let txt_wid = Dutil.g_widget_named gd "text" in
    da_term_send sc.graph.dd txt_wid vt_erase_in_display;
    draw_text sc txt_wid
  }
  else ()
;
