(* $Id: graph_gtk.ml,v 1.246 2011-11-03 13:17:07 deraugla Exp $ *)

(*
open Ast;
open Data_gtk;
open Dim_def;
open Dtasks;
open Glob_glob;
open Printf;

value color_list =
  [(0, 255, 0); (255, 0, 0); (0, 0, 255); (255, 0, 255);
   (0, 255, 255); (200, 200, 0); (255, 128, 0); (128, 0, 255);
   (128, 0, 128); (0, 128, 128);

   (128, 255, 0); (255, 128, 128); (128, 128, 255); (255, 128, 255);
   (128, 255, 255); (200, 200, 128); (128, 128, 128)]
;

value make_graph ts2 = do {
  let ts = ts2.ast in
  let many_tasks = Dim.many_tasks ts.tasks_def in
  let (dam_width, dam_height) =
    let h = Gdk.Screen.height () in
    if many_tasks then (h * 4 / 10, h / 3)
    else (h * 5 / 10, h / 2)
  in
  let damdim = Dim.dim_of_width_height ts2 dam_width dam_height in
  let inter_task_dy = if many_tasks then 35 else 40 in
  let raw_font =
    Gdk.Font.load "-*-courier-bold-r-normal-*-10-*-*-*-*-*-iso8859-1"
  in
  let bigger_font =
    Gdk.Font.load "-*-courier-bold-r-normal-*-18-*-*-*-*-*-iso8859-1"
  in
  let smaller_font =
    Gdk.Font.load "-*-courier-bold-r-normal-*-8-*-*-*-*-*-iso8859-1"
  in
  {zooms_sch = []; widget_ht = Hashtbl.create 1; dam_dim = damdim;
   inter_task_dy = inter_task_dy;
   dam_size = ref (damdim.dam_width, damdim.dam_height);
   cap_size = ref (120, 1); sched_size = ref (0, 0); mes_size = ref (0, 0);
   raw_font = raw_font; bigger_font = bigger_font;
   smaller_font = smaller_font}
};

value make_scenario (ts2, fname) = do {
  let graph = make_graph ts2 in
  Scenario.make_scenario_gen (ts2, fname) graph
};

value empty_scenario tg name =
  match Sparse.parse_and_check_dameid (Stream.of_string "0/") with
  [ [(_, df)] -> do {
      let ts = Scenario.make_tasks df in
      make_scenario (ts, name)
    }
  | [] | [_ :: _] -> failwith "internal error" ]
;

value current_scenario tg = do {
  if Array.length tg.scenario = 0 then do {
    let sc = empty_scenario tg "" in
    tg.scenario := [| sc |];
    tg.ind_of_isc := [| 0 |];
    tg.curr_sc := 0;
  }
  else ();
  tg.scenario.(tg.curr_sc)
};

value signal_connect ~{sgn; callback} item =
  ignore (GtkSignal.connect ~{sgn; callback} item : GtkSignal.id)
;

value open_file_on_dir tg gi initial_dir = do {
  let file_sel = GtkWindow.FileSelection.create (I18n.transl "Open") in
  Gobject.set GtkWindow.FileSelection.P.show_fileops file_sel False;
  GtkWindow.FileSelection.complete file_sel "*.dm";
  signal_connect ~{sgn = GtkBase.Object.S.destroy}
    ~{callback () =
        let fname = Gobject.get GtkWindow.FileSelection.P.filename file_sel in
        let dir =
          if Sys.file_exists fname && Sys.is_directory fname then fname
          else Filename.dirname fname
        in
        tg.open_dir := dir}
    file_sel;
  let ok = GtkWindow.FileSelection.get_ok_button file_sel in
  signal_connect ~{sgn = GtkButton.Button.S.clicked}
    ~{callback () = do {
        let fname = Gobject.get GtkWindow.FileSelection.P.filename file_sel in
        if Sys.file_exists fname && Sys.is_directory fname then do {
          Gobject.set GtkWindow.FileSelection.P.filename file_sel (fname ^ "/")
        }
        else do {
          tg.open_dir := Filename.dirname fname;
          match Openfile.input_dameid_scenarios tg fname make_scenario with
          [ Some len -> do {
              for isc = len to Array.length tg.scenario - 1 do {
                Scenario_gtk.build_scenario_windows tg gi isc;
              };
              Gobject.set GtkPack.Notebook.P.page gi.main_noteb 1;
              GtkBase.Object.destroy file_sel
            }
          | None -> () ]
        }
      }}
    ok;
  let cancel = GtkWindow.FileSelection.get_cancel_button file_sel in
  signal_connect ~{sgn = GtkButton.Button.S.clicked}
    ~{callback () = GtkBase.Object.destroy file_sel} cancel;
  GtkBase.Widget.show file_sel
};

value open_file tg gi = do {
  let initial_dir = Sys.getcwd () in
  Openfile.save_dir_excursion
    (fun () -> do {
       Sys.chdir tg.open_dir;
       open_file_on_dir tg gi initial_dir
     })
};

value goto_begin tg lazy_gc = do {
  let gc = (Lazy.force lazy_gc, (ref None, ref 0)) in
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    Schedule.goto_begin_schedule sc.ts;
    Scenario_gtk.redisplay tg gc
  }
  else ();
};

value forward_not_audsley tg sc gc = do {
  let ts = sc.ts in
  if ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
     ts.schedule_status <> SS_running
  then do {
    let ok_lat =
      List.fold_right
        (fun l ok_lat ->
           match l.l_status with
           [ LS_values be1 be2 lat -> [(l, be1, be2, lat) :: ok_lat]
           | _ -> ok_lat ])
        ts.latency_constr []
    in
    let len = List.length ok_lat in
    if len >= 2 && sc.latency_disp_cnt < len - 1 then do {
      sc.latency_disp_cnt := sc.latency_disp_cnt + 1;
      Scenario_gtk.redisplay tg gc;
    }
    else ()
  }
  else do {
    match
      try do { Schedule.schedule_next_instance ts; None } with e -> Some e
    with
    [ None -> ()
    | Some (Schedule.Schedule_interrupted sf) -> ts.schedule_status := sf
    | Some e -> raise e ];
    let comp_rep_part =
      match ts.schedule_status with
      [ SS_schedulable_but _ | SS_several_lcm _ -> True
      | SS_task_error _ _ | SS_incomplete_scheduling _ _ -> False
      | SS_running -> ts.curr_ord_ind = Array.length ts.ast.tasks_def ]
    in
    if comp_rep_part then
      match Schedule.compute_repetitive_part ts with
      [ Some r -> do {
          ts.repetition := Some r;
          ts.do_refresh := True
        }
      | None -> () ]
    else ();
    if ts.do_refresh then do {
(*
      recompute_dim sc sc.graph.dd.rt_dim.dam_width
        sc.graph.dd.rt_dim.dam_height;
*)
      for i = 0 to ts.schedule_len - 1 do {
        Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
      };
      ts.do_refresh := False;
      Scenario_gtk.redisplay tg gc
    }
    else do {
      Scenario_gtk.update_draw tg sc gc
    };
    if ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
       ts.schedule_status <> SS_running
    then do {
      let gfun = tg.g_fun in
      let dam_wid = Scenario_gtk.g_widget_named tg "dameid" in
      let sched_wid = Scenario_gtk.g_widget_named tg "sched" in
      let zoom = NoZoom in
      let dam_drw =
        Scenario_gtk.drawable_of_zoom dam_wid sc.graph.dam_size sc zoom
      in
      Draw_dameid.draw_latencies gfun sc dam_drw gc zoom False;
      let sched_drw =
        Scenario_gtk.drawable_of_zoom sched_wid sc.graph.sched_size sc zoom
      in
      Draw_schedule.draw_latencies gfun sc sched_drw gc zoom False;
      Scenario_gtk.expose_text tg gc
    }
    else ();
  }
};

value forward tg lazy_gc = do {
  let gc = (Lazy.force lazy_gc, (ref None, ref 0)) in
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    match sc.ts.ast.algorithm with
    [ AL_audsley a -> failwith "Audsdley algo not impl"
    | AL_deadline_monotonic | AL_rate_monotonic | AL_user _  | AL_edf ->
        forward_not_audsley tg sc gc ]
  }
  else ()
};

value recompute_dim sc dam_width dam_height = do {
  sc.graph.dam_dim := Dim.dim_of_width_height sc.ts dam_width dam_height;
  sc.latency_disp_cnt := 0;
};

value backward_schedule_not_edf tg sc gc = do {
  let ts = sc.ts in
  let must_goto_begin = Schedule_step.gen_backward_schedule ts in
(*
  let was_ended =
    ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
    ts.schedule_status <> SS_running
  in
*)
  if must_goto_begin then do {
    Schedule.goto_begin_schedule ts;
    Scenario_gtk.redisplay tg gc;
  }
  else if ts.do_refresh then do {
    recompute_dim sc sc.graph.dam_dim.dam_width sc.graph.dam_dim.dam_height;
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
    };
    Scenario_gtk.redisplay tg gc;
    ts.do_refresh := False;
  }
  else do {
    Scenario_gtk.update_draw tg sc gc;
  }
};

value backward_schedule_edf tg sc gc = do {
  let ts = sc.ts in
  let current_i =
    loop (ts.schedule_len - 1) where rec loop i =
      match Sarray.get ts.disp_schedule i with
      [ Some _ -> i + 1
      | None -> if i = 0 then 0 else loop (i - 1) ]
  in
  if current_i = 1 then do {
    Schedule.goto_begin_schedule ts;
    Scenario_gtk.redisplay tg gc;
  }
  else if current_i > 0 then do {
    ts.curr_step := ts.curr_step - 1;
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.comp_schedule i (Sarray.get ts.disp_schedule i);
    };
    if ts.curr_ord_ind = Array.length ts.ast.tasks_def then
      ts.curr_ord_ind := 0
    else
      Sarray.set ts.comp_schedule (current_i - 1) None;
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
    };
    Scenario_gtk.redisplay tg gc
  }
  else ()
};

value backward_schedule tg sc gc = do {
  Schedule_step.backward_schedule_init sc.ts;
  match sc.ts.ast.algorithm with
  [ AL_edf ->
     backward_schedule_edf tg sc gc
  | AL_audsley _ | AL_deadline_monotonic | AL_rate_monotonic | AL_user _ ->
     backward_schedule_not_edf tg sc gc ]
};

value backward tg lazy_gc = do {
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    let ts = sc.ts in
    let gc = (Lazy.force lazy_gc, (ref None, ref 0)) in
    let was_ended =
      ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
      ts.schedule_status <> SS_running
    in
    if was_ended && List.length ts.latency_constr >= 2 &&
       sc.latency_disp_cnt > 0
    then do {
      sc.latency_disp_cnt := sc.latency_disp_cnt - 1;
      Scenario_gtk.redisplay tg gc;
    }
    else do {
      backward_schedule tg sc gc
    }
  }
  else ()
};

value goto_normal_end tg ts gc = do {
  try Schedule.schedule_all ts with
  [ Schedule.Schedule_interrupted sf -> ts.schedule_status := sf ];
  for i = 0 to ts.schedule_len - 1 do {
    Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
  };
  if ts.do_refresh then Scenario_gtk.redisplay tg gc
  else Scenario_gtk.expose tg gc;
};

value goto_end tg lazy_gc = do {
  let gc = (Lazy.force lazy_gc, (ref None, ref 0)) in
  if tg.curr_sc < Array.length tg.scenario then do {
    let isc = tg.curr_sc in
    let sc = tg.scenario.(isc) in
    let ts = sc.ts in
    goto_normal_end tg ts gc
  }
  else ()
};

value display_change_parameters tg gi () = do {
  let no_sc = Array.length tg.scenario = 0 in
  let sc = current_scenario tg in
  let param_wid =
    GtkWindow.Dialog.create
      [Gobject.param GtkWindow.Window.P.title (I18n.transl "Parameters");
       Gobject.param GtkWindow.Window.P.transient_for (Some gi.root_wid)]
  in
  Parameters_gtk.parameter_wdesc tg gi no_sc sc param_wid;
  GtkBase.Widget.show param_wid;
};

value help_zoom tg gi = do {
  let help_wid =
    GtkWindow.Dialog.create
      [Gobject.param GtkWindow.Window.P.title (I18n.transl "How to zoom");
       Gobject.param GtkWindow.Window.P.transient_for (Some gi.root_wid)]
  in
  let vbox_wid = GtkWindow.Dialog.vbox help_wid in
  let label_wid =
    let text =
      match I18n.ptransl "~how to zoom" with
      [ Some s -> s
      | None -> Dmisc.default_zoom_help ]
    in
    GtkMisc.Label.create
      [Gobject.param GtkMisc.Label.P.label text;
       Gobject.param GtkMisc.Misc.P.xpad 42;
       Gobject.param GtkMisc.Misc.P.ypad 42]
  in
  GtkBase.Widget.show label_wid;
  GtkBase.Container.add vbox_wid label_wid;

  GtkWindow.Dialog.add_button help_wid "OK" 2;
  signal_connect help_wid
     ~{sgn = GtkWindow.Dialog.S.response;
       callback x = do {
         if x = 2 then GtkBase.Object.destroy help_wid
         else if x = -4 then ()
         else do {
           printf "help_wid %d\n" x;
           flush stdout
         }
       }};
  GtkBase.Widget.show help_wid;
};

value action_configure_root first_configure_root root ev = do {
  if first_configure_root.val then do {
    first_configure_root.val := False;
    let scr_w = Gdk.Screen.width () in
    let scr_h = Gdk.Screen.height () in
    let w = GdkEvent.Configure.width ev in
    let h = GdkEvent.Configure.height ev in
    let (x, y) = ((scr_w - w)/2, (scr_h - h)/2) in
    GtkWindow.Window.move root ~{x; y};
  }
  else ();
};

value set_algo tg lazy_gc algo = do {
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    if sc.ts.ast.algorithm <> algo then do {
      if tg.common.ignore_radialgo_callback then
        tg.common.ignore_radialgo_callback := False
      else do {
        let ts = {(sc.ts.ast) with algorithm = algo} in
        sc.ts.ast := ts;
        Scenario.set_task_order sc.ts;
        goto_begin tg lazy_gc;
      }
    }
    else ();
  }
  else ()
};

value set_audsley_algo tg lazy_gc = do {
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    let algo =
      let t = Array.init (Array.length sc.ts.ast.tasks_def) (fun i -> i) in
      let auds = {ai_pos = P_begin; ai_tab = t; ai_step = 0; ai_nsol = 0} in
      AL_audsley auds
    in
    set_algo tg lazy_gc algo
  }
  else ()
};

value make_gtk_info () = do {
  let root =
    GtkWindow.Window.create
      [Gobject.param GtkWindow.Window.P.title "dameid"]
  in
  let main_noteb = GtkPack.Notebook.create [] in
  let scen_noteb = GtkPack.Notebook.create [] in
  let lazy_gc =
    let root_win = lazy (GtkBase.Widget.window root) in
    lazy (Gdk.GC.create (Lazy.force root_win))
  in
  {root_wid = root; main_noteb = main_noteb; scen_noteb = scen_noteb;
   lazy_gc = lazy_gc}
};

value make_dameid_window tg gi = do {
  let main_text =
    match I18n.ptransl "~main text" with
    [ Some s -> s
    | None -> Dmisc.default_main_text ]
  in
  signal_connect GtkBase.Object.S.destroy GMain.Main.quit gi.root_wid;

  let vbox = GtkPack.Box.create `VERTICAL [] in
  GtkBase.Widget.show vbox;
  GtkBase.Container.add gi.root_wid vbox;
  let hbox = GtkPack.Box.create `HORIZONTAL [] in
  GtkBase.Widget.show hbox;
  GtkPack.Box.pack vbox hbox ~{from = `START};
  let menu_bar = GtkMenu.MenuBar.create [] in
  GtkBase.Widget.show menu_bar;
  GtkBase.Container.add hbox menu_bar;

  let accel_group = GtkData.AccelGroup.create () in
  GtkWindow.Window.add_accel_group gi.root_wid accel_group;

  GtkBase.Widget.show gi.main_noteb;
  GtkBase.Container.add vbox gi.main_noteb;
  Gobject.set GtkPack.Notebook.P.show_tabs gi.main_noteb False;

  let main =
    GtkMisc.Label.create
      [Gobject.param GtkMisc.Label.P.label main_text;
       Gobject.param GtkMisc.Misc.P.xpad 50;
       Gobject.param GtkMisc.Misc.P.ypad 50]
  in
  GtkBase.Widget.show main;
  GtkBase.Container.add gi.main_noteb main;

  GtkBase.Widget.show gi.scen_noteb;
  GtkBase.Container.add gi.main_noteb gi.scen_noteb;

  signal_connect ~{sgn = GtkPack.Notebook.S.switch_page}
    ~{callback ev n = tg.curr_sc := n} gi.scen_noteb;

  let is_prec = False in
  let is_strict_per = False in
  let is_fully_non_pree = False in
  let button_table = Hashtbl.create 1 in
  let check_table = Hashtbl.create 1 in
  let radio_table = Hashtbl.create 1 in
  let menu_buttons =
    List.map
      (Gtkn.make_one_menu button_table check_table radio_table accel_group)
      [(I18n.transl "File",
        [Gtkn.M_button "" (I18n.transl "Open") (Some ([], Char.code 'o'))
           (fun () -> open_file tg gi);
         Gtkn.M_button "" (I18n.transl "Save parameters")
           (Some ([], Char.code 's')) (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Save resulting schedule")
           (Some ([], Char.code 't')) (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Print in PostScript")
           (Some ([], Char.code 'g')) (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Quit")
           (Some ([`CONTROL], Char.code 'q')) GMain.Main.quit]);
       (I18n.transl "Edit",
        [Gtkn.M_button "" (I18n.transl "Change parameters")
           (Some ([], Char.code 'p')) (display_change_parameters tg gi)]);
       (I18n.transl "Display",
        [Gtkn.M_button "" (I18n.transl "Sort by decreasing priority") None
           (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Sort by increasing priority") None
           (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Sort by release date") None
           (fun () -> ());
         Gtkn.M_button "" (I18n.transl "Sort according to initial order")
           None (fun () -> ());
         Gtkn.M_separator;
         Gtkn.M_button "" (I18n.transl "Mesoid") (Some ([], Char.code 'm'))
           (fun () -> ());
         Gtkn.M_separator;
         Gtkn.M_button "" (I18n.transl "Zoom") (Some ([], Char.code 'z'))
           (fun () -> help_zoom tg gi)]);
       (I18n.transl "Constraints",
        [Gtkn.M_check "prec" is_prec (I18n.transl "Precedence") None
           (fun () -> do { printf "precedence\n"; flush stdout });
         Gtkn.M_check "strict_per" is_strict_per
           (I18n.transl "Strict Periodicity")
           None (fun () -> do { printf "strict per\n"; flush stdout });
         Gtkn.M_check "tot_non_pree" is_fully_non_pree
           (I18n.transl "Fully Non Preemptive") None
           (fun () -> do { printf "fully non pree\n"; flush stdout })]);
       (I18n.transl "Algorithm",
        [Gtkn.M_radiobuttons "radialgo"
           [(I18n.transl "Deadline-Monotonic",
             fun () -> set_algo tg gi.lazy_gc AL_deadline_monotonic);
            (I18n.transl "Rate-Monotonic",
             fun () -> set_algo tg gi.lazy_gc AL_rate_monotonic);
            (I18n.transl "Earliest Deadline First",
             fun () -> set_algo tg gi.lazy_gc AL_edf);
            (I18n.transl "Audsley++",
             fun () -> set_audsley_algo tg gi.lazy_gc);
            (I18n.transl "User Priorities",
             fun () -> do { printf "User Priorities\n"; flush stdout })]]);
       (I18n.transl "Run",
        [Gtkn.M_button "" (I18n.transl "Clear")
          (Some ([], GdkKeysyms._Home)) (fun () -> goto_begin tg gi.lazy_gc);
         Gtkn.M_button "" (I18n.transl "One step back")
          (Some ([], GdkKeysyms._BackSpace))
          (fun () -> backward tg gi.lazy_gc);
         Gtkn.M_button "" (I18n.transl "One step")
           (Some ([], Char.code ' ')) (fun () -> forward tg gi.lazy_gc);
         Gtkn.M_button "" (I18n.transl "Schedule")
           (Some ([], GdkKeysyms._End)) (fun () -> goto_end tg gi.lazy_gc)]);
       (I18n.transl "About", [])]
  in
  List.iter (Gtkn.menu_shell_append menu_bar) menu_buttons;

  let algowid = Hashtbl.find radio_table "radialgo" in
  List.iteri
    (fun i wid -> do {
       GtkBase.Widget.add_events wid [`STRUCTURE];
       signal_connect ~{sgn = GtkBase.Widget.Signals.Event.map}
         ~{callback _ = do {
             match tg.common.radialgo with
             [ Some _ -> ()
             | None -> tg.common.radialgo := Some algowid ];
             if tg.curr_sc < Array.length tg.scenario then do {
               let sc = tg.scenario.(tg.curr_sc) in
               let algo = sc.ts.ast.algorithm in
               let nalgo =
                 match algo with
                 [ AL_deadline_monotonic -> 0
                 | AL_rate_monotonic -> 1
                 | AL_edf -> 2
                 | AL_audsley _ -> 3
                 | AL_user _ -> 4 ]
               in
               if i = nalgo then
                 GtkMenu.CheckMenuItem.set ~{active = True} wid
               else ()
             }
             else ();
             False
           }}
          wid
     })
     algowid;

  for isc = 0 to Array.length tg.scenario - 1 do {
    Scenario_gtk.build_scenario_windows tg gi isc;
  };

  if Array.length tg.scenario > 0 then
    Gobject.set GtkPack.Notebook.P.page gi.main_noteb 1
  else ();

  let first_configure_root = ref True in
  signal_connect GtkBase.Widget.Signals.Event.configure
    (fun ev -> do {
       action_configure_root first_configure_root gi.root_wid ev;
       False
     })
    gi.root_wid;
  GtkBase.Widget.show gi.root_wid;
};

value go tg gi = do {
  make_dameid_window tg gi;
  Sys.catch_break True;
  let curr_dir = Sys.getcwd () in
  GMain.Main.main ();
  Sys.chdir curr_dir
};

value make_common () =
  let color =
    Array.of_list
      (List.map (fun (r, g, b) -> (256 * r, 256 * g, 256 * b)) color_list)
  in
  {color = color; radialgo = None; ignore_radialgo_callback = False}
;

value font_of_graph graph =
  fun
  [ F_raw -> graph.raw_font
  | F_bigger -> graph.bigger_font
  | F_smaller -> graph.smaller_font ]
;

value display_dialog gi txt = do {
  let wid =
    GtkWindow.Dialog.create
      [Gobject.param GtkWindow.Window.P.title (I18n.transl "Error");
       Gobject.param GtkWindow.Window.P.transient_for (Some gi.root_wid)]
  in
  let vbox_wid = GtkWindow.Dialog.vbox wid in
  let label_wid =
    GtkMisc.Label.create
      [Gobject.param GtkMisc.Label.P.label txt;
       Gobject.param GtkMisc.Misc.P.xpad 42;
       Gobject.param GtkMisc.Misc.P.ypad 42]
  in
  GtkBase.Widget.show label_wid;
  GtkBase.Container.add vbox_wid label_wid;

  GtkWindow.Dialog.add_button wid "OK" 2;
  signal_connect wid
     ~{sgn = GtkWindow.Dialog.S.response;
       callback x = do {
         if x = 2 then GtkBase.Object.destroy wid
         else if x = -4 then ()
         else do {
           printf "dialog wid %d\n" x;
           flush stdout
         }
       }};
  GtkBase.Widget.show wid
};

value g_fun co gi =
  let colormap = Gdk.Color.get_system_colormap () in
  let g_back_col = ref None in
  let g_black = Gdk.Color.alloc colormap `BLACK in
  let g_white = Gdk.Color.alloc colormap `WHITE in
  let g_gray = Gdk.Color.alloc colormap (`RGB (150*256) (150*256) (150*256)) in
  let g_red = Gdk.Color.alloc colormap (`RGB 65535 0 0) in
  let g_backg_rep_col =
    let (r, g, b) = (216* 256, 216*256, 255*256) in
    Gdk.Color.alloc colormap (`RGB r g b)
  in
  let g_color =
    Array.map (fun (r, g, b) -> Gdk.Color.alloc colormap (`RGB r g b))
      co.color
  in
  {g_record_style_bg_color (r16, g16, b16) = do {
      g_back_col.val := Some (Gdk.Color.alloc colormap (`RGB r16 g16 b16));
    };
   g_clear_drawable (drw, _) =
     match drw with
     [ Gtkn.WindowDr win -> Gdk.Window.clear win
     | Gtkn.PixmapDr _ -> () ];
   g_commit_display drw = ();
   g_draw_arc (drw, _) (gc, _) (x, y, w, h) (a, da) = do {
     Gtkn.draw_arc drw gc False x y w h
       (float a /. float Draw_dameid.grand_tour *. 360.)
       ((float da) /. float Draw_dameid.grand_tour *. 360.)
   };
   g_draw_line (drw, _) (gc, _) (x1, y1) (x2, y2) = do {
     Gtkn.draw_line drw gc (x1, y1) (x2, y2)
   };
   g_draw_lines (drw, _) (gc, _) list = do {
     Gtkn.draw_lines drw gc list
   };
   g_draw_rectangle (drw, _) (gc, _) (x, y, w, h) = do {
     failwith "not impl graph_gtk.g_draw_rectangle";
   };
   g_fill_arc (drw, _) (gc, _) (x, y, w, h) (a, da) = do {
     Gtkn.draw_arc drw gc True x y w h
       (float a /. float Draw_dameid.grand_tour *. 360.)
       ((float da) /. float Draw_dameid.grand_tour *. 360.)
   };
   g_fill_polygon drw (gc, _) list = match () with [];
   g_fill_rectangle (drw, _) (gc, _) (x, y, w, h) = do {
     Gtkn.draw_rectangle drw gc True x y w h;
   };
   g_draw_string (drw, _) (gc, (font, _)) (x, y) text = do {
     let text = Std_term.latin_1_of_utf_8 text in
     match font.val with
     [ Some font -> Gtkn.draw_string drw font gc x y text
     | None -> failwith "Graph_gtk.g_erase_draw_string: no font selected" ]
   };
   g_erase_draw_string (drw, _) (gc, (font, _)) (x, y) text = do {
(*
     Gtkn.draw_rectangle drw gc True x1 y1 (x2-x1) (y2-y1);
*)
     let text = Std_term.latin_1_of_utf_8 text in
     match font.val with
     [ Some font -> Gtkn.draw_string drw font gc x y text
     | None -> failwith "Graph_gtk.g_erase_draw_string: no font selected" ]
   };
   g_dam_dim (_, (_, graph)) = graph.dam_dim;
   g_drawable_width (_, (size, _)) = fst size.val;
   g_drawable_height (_, (size, _)) = snd size.val;
   g_schedule_zoom_size graph =
     fun
     [ Zoom _ wid _ _ -> Gdk.Drawable.get_size (Gtkn.widget_window wid)
     | NoZoom -> graph.sched_size.val ]
   ;
   g_inter_task_dy (_, (_, graph)) = graph.inter_task_dy;
   g_set_color _ (gc, _) col = do {
     let col =
       match col with
       [ C_black -> g_black
       | C_white -> g_white
       | C_gray -> g_gray
       | C_red -> g_red
       | C_backg_rep_col -> g_backg_rep_col
       | C_number i -> g_color.(i mod Array.length g_color)
       | C_style_bg -> do {
           match g_back_col.val with
           [ Some col -> col
           | None -> g_white ]
         } ]
     in
     Gdk.GC.set_foreground gc col;
   };
   g_set_dashes graph (gc, (font_ref, lw_ref)) off list = do {
     Gdk.GC.set_line_attributes gc
       ~{width = lw_ref.val; style = `ON_OFF_DASH; cap = `NOT_LAST;
         join = `MITER};
     Gdk.GC.set_dashes gc ~{offset = off} list;
   };
   g_set_font graph (gc, (font_ref, _)) font = do {
     let font = font_of_graph graph font in
     font_ref.val := Some font;
     Gdk.GC.set_font gc font
   };
   g_set_line_width _ (gc, (_, lw_ref)) w = do {
     Gdk.GC.set_line_attributes gc
       ~{width = w; style = `SOLID; cap = `NOT_LAST; join = `MITER};
     lw_ref.val := w;
   };
   g_set_pixmap _ (gc, _) i = do {
     failwith "Graph_gtk.g_set_pixmap not impl";
   };
   g_text_width graph font str = do {
     let font = font_of_graph graph font in
     Gdk.Font.string_width font str
   };
   g_font_size graph font = do {
     let font = font_of_graph graph font in
     Gdk.Font.ascent font + Gdk.Font.descent font
   };
   g_xcoord_of_coord (_, (_, graph)) (x, y) = do {
     (x + graph.dam_dim.x_center, - y + graph.dam_dim.y_center)
   };
   g_error exc = do {
     display_dialog gi (Printexc.to_string exc);
   };
   g_treat_pending_events () = failwith "not impl g_treat_pending_events"}
;

value start_with_dameid_window fnames algo_opt =
  let dfnl = Openfile.get_dameid_files fnames algo_opt in
  if dfnl = [] then exit 1
  else do {
    let dfnl =
      loop 7 dfnl where rec loop n =
        fun
        [ [dfn :: dfnl] -> if n = 0 then [] else [dfn :: loop (n - 1) dfnl]
        | [] -> [] ]
    in
    let tsnl =
      List.map (fun (df, fname) -> (Scenario.make_tasks df, fname)) dfnl
    in
    let _ : string = GtkMain.Main.init () in
    let co = make_common () in
    let gi = make_gtk_info () in
    let scs = List.map make_scenario tsnl in
    let scenario = Array.of_list scs in
    let ind_of_isc = Array.init (Array.length scenario) (fun i -> i) in
    let tg =
      {scenario = scenario; ind_of_isc = ind_of_isc; curr_sc = 0;
       open_dir = Sys.getcwd (); prefix_number = 0; last_zoom = 2;
       graph_nb = 0; common = co; g_fun = g_fun co gi}
    in
    go tg gi
  }
;

value start_empty () = do {
  let _ : string = GtkMain.Main.init () in
  let co = make_common () in
  let gi = make_gtk_info () in
  let gd =
    {scenario = [| |]; ind_of_isc = [| |]; curr_sc = 0;
     open_dir = Sys.getcwd (); prefix_number = 0; last_zoom = 2;
     graph_nb = 0; common = co; g_fun = g_fun co gi}
  in
  go gd gi
};

value start fnames algo_opt =
  if fnames = [] then start_empty ()
  else start_with_dameid_window fnames algo_opt
;
*)
