(* $Id: graph_rt.ml,v 1.51 2011-11-03 12:53:17 deraugla Exp $ *)

open Rt;
open Printf;

open Ast;
open Da;
open Data_rt;
open Dim_def;
open Dtasks;
open Glob_glob;

value raise_scenario gd ti sc = do {
  Schedule_step_rt.close_progr_bar gd.common.xd;
  Scenario_rt.raise_scenario gd ti sc
};

value erase_scenario gd ti isc = do {
  Schedule_step_rt.close_progr_bar gd.common.xd;
  Scenario_rt.erase_scenario gd ti isc
};

value action_main xa gd ti main_wid ev =
  match ev with
  [ PackEvKeyPress ksym ->
      match ksym with
      [ {item = K_Ascii ('1' .. '9' as c); control = True} ->
          let isc = Char.code c - Char.code '1' in
          if isc < Array.length gd.scenario then raise_scenario gd ti isc
          else ()
      | {item = K_KP_N n; control = True} ->
          let isc = n - 1 in
          if isc >= 0 && isc < Array.length gd.scenario then
            raise_scenario gd ti isc
          else ()
      | {item = K_Ascii ('0'..'9' as c)} ->
          let n = 10 * gd.prefix_number + Char.code c - Char.code '0' in
          gd.prefix_number := n
      | {item = K_KP_N i} ->
          gd.prefix_number := 10 * gd.prefix_number + i
      | {item = K_Ascii 'q'; control = True} ->
          rt_stop_main_loop xa
      | {item = k} -> do {
          gd.prefix_number := 0;
          Mainwid.action_main_key_press xa gd ti main_wid k
        } ] ]
;

(*
Put in the Olibrt Toolkit to allow interface with Camltk:

value menu_button attl args wid =
  let act xll yll = rt_map_popup_widget wid xll yll 0 in
  button_desc attl args
    (fun _ ->
       fun
       [ ButtonEvPress xll yll _ -> act xll yll
       | ButtonEvEnter xll yll -> act xll yll
       | _ -> () ])
;
*)

Schedule_step_rt.main_window_close_action.val := rt_stop_main_loop;

value multi_proc xa gd =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let nsol = ref 0 in
    let nproc = Array.length ts.tasks_def in
    let list_proc = Sutil.rho nproc in
    loop_combine [] (Sutil.rho (Array.length ts.tasks_def))
    where rec loop_combine rev_sol list_task =
      match list_task with
      [ [t :: tl] ->
          List.iter (fun p -> loop_combine [(p, t) :: rev_sol] tl)
            list_proc
      | [] ->
          if rev_sol <> [] then do {
            incr nsol;
            printf "  sol %d:" nsol.val;
            for i = 0 to nproc - 1 do {
              let pfx = ref (sprintf " p%d-" (i + 1)) in
              List.iter
                (fun (p, t) ->
                   if p = i then do {
                     printf "%s%d" pfx.val (t + 1);
                     pfx.val := "";
                   }
                   else ())
                (List.rev rev_sol)
            };
(*
            List.iter (fun (p, t) -> printf " p%d-t%d" (p + 1) (t + 1))
              (List.rev rev_sol);
*)
            printf "\n";
            flush stdout;
          }
          else () ]
  }
  else ()
;

value mainwid_make_multi_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "Multi", None)
           (Dutil.button_release_action (fun _ -> multi_proc xa gd))])
       (fun _ _ -> ()))
;

value make_dameid_widget xd xa gd ti = do {
  let backg = BackgroundAtt (ColorPn (fst gd.common.c_white_col)) in
  let foreg = ForegroundAtt (ColorPn (fst gd.common.c_red_col)) in
  let file_wid = Mainwid.make_file_popup xa xd gd ti backg foreg in
  let edit_wid = Mainwid.make_edit_popup xa xd gd ti backg foreg in
  let disp_wid = Mainwid.make_disp_popup xa xd gd ti backg foreg in
  let constr_wid = Mainwid.make_constr_popup xa xd gd ti backg foreg in
  let algo_wid = Mainwid.make_algo_popup xa xd gd ti backg foreg in
  let dam_wid = Mainwid.make_dam_popup xa xd gd ti backg foreg in
(**)
  let multi_wid = mainwid_make_multi_popup xa xd gd ti backg foreg in
(**)
  let help_wid = Mainwid.make_help_popup xa xd gd ti backg foreg in
  let about_wid = Mainwid.make_about_popup xa xd gd ti backg foreg in
  let def_keyp = Mainwid.action_main_key_press xa gd ti in
  let main_text =
    match I18n.ptransl "~main text" with
    [ Some s -> s
    | None -> Dmisc.default_main_text ]
  in
  let (main_text_rows, main_text_cols) = Dmisc.text_rows_cols main_text in
  let wid =
    let fname =
      if Array.length gd.scenario > 0 then gd.scenario.(0).fname else ""
    in
    let t = if fname = "-" then "" else " - " ^ fname in
    let wname = "Dameid" ^ t in
    let iname = "Dameid" in
    let filler x =
      pack_desc [FillerAtt; backg; BandAtt x] (Vertical, []) (fun _ _ -> ())
    in
    rt_create_widget xd wname iname AutoPosition
      (Some (fun _ -> Schedule_step_rt.main_window_close_action.val xa))
      (pack_desc [NameAtt "main"; BorderAtt 0; BandAtt 0; InterAtt 0; backg]
         (Vertical,
          [pack_desc [backg; BorderAtt 0]
             (Horizontal,
              [menu_button [backg] (I18n.transl "File", None) file_wid;
               menu_button [backg] (I18n.transl "Edit", None) edit_wid;
               menu_button [backg] (I18n.transl "Display", None) disp_wid;
               menu_button [backg] (I18n.transl "Constraints", None)
                 constr_wid;
               menu_button [backg] (I18n.transl "Algorithm", None) algo_wid;
               menu_button [backg] (I18n.transl "Run", None) dam_wid;
(**)
               menu_button [backg] (I18n.transl "Multi", None) multi_wid;
(**)
               menu_button [backg] (I18n.transl "Help", None) help_wid;
               menu_button [backg] (I18n.transl "About", None) about_wid])
             (fun _ _ -> ());
           pack_desc [backg; BorderAtt 0; BorderAtt 0; BandAtt 0; FillerAtt]
             (InDepth,
              [pack_desc [NameAtt "main_text_pack"; backg]
                (Vertical,
                 [filler 10;
                  pack_desc [backg]
                    (Horizontal,
                     [filler 30;
                      term_desc [NameAtt "main_text"]
                        (main_text_rows, main_text_cols, 0)
                        (fun _ _ -> ());
                      filler 30])
                    (fun _ _ -> ());
                  filler 10])
                 (fun _ _ -> ());
               Tabwid.tab_desc [NameAtt "main_tab"; backg; FillerAtt]
                 (ti,
                  List.map
                    (fun i -> Scenario_rt.scenario_desc xa gd ti def_keyp i)
                    (Sutil.rho (Array.length gd.scenario)))
                 (fun _ _ -> ())])
             (fun _ _ -> ())])
         (action_main xa gd ti))
  in
  for i = 0 to Array.length gd.scenario - 1 do {
    Scenario_rt.make_moving_schedule_widget gd i;
  };
(*
  rt_move_widget wid ((screen_width xd - widget_width wid) / 2)
    ((screen_height xd - widget_height wid) / 2);
*)
  let main_text_wid = widget_named xd "main_text" in
  term_send main_text_wid Std_term.vt_bold;
  term_send main_text_wid Std_term.vt_newline_mode;
  term_send main_text_wid main_text;
  if Array.length gd.scenario > 0 then
    rt_raise_widget (widget_named xd "main_tab")
  else ();
  if Array.length gd.scenario > 1 then do {
    rt_freeze_widget (widget_named xd "set_show_tabs");
    rt_unfreeze_widget (widget_named xd "set_hide_tabs");
  }
  else do {
    rt_freeze_widget (widget_named xd "set_hide_tabs");
    rt_unfreeze_widget (widget_named xd "set_show_tabs");
  };
  Array.iteri
    (fun i sc -> Expose.draw_text sc (widget_named xd (sprintf "text_%d" i)))
    gd.scenario;
  rt_freeze_widget (widget_named xd "initial_order");
  rt_freeze_widget (widget_named xd "set_schedule");
  rt_freeze_widget (widget_named xd "set_coloured");
  rt_redirect_key_press_to (widget_named xd "main");
  if Array.length gd.scenario > 0 then do {
    let sc = gd.scenario.(0) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let tc = ts.constraints in
    check_set (widget_named xd "prec_constr") tc.global_precedence;
    check_set (widget_named xd "strict_per_constr") ts2.strict_periodicity;
    check_set (widget_named xd "no_preempt_constr") ts2.fully_non_preemptive;
    check_set
      (widget_named xd
         (match ts.algorithm with
          [ AL_deadline_monotonic -> "dead_mono_algo"
          | AL_rate_monotonic -> "rate_mono_algo"
          | AL_edf -> "edf_algo"
          | AL_audsley _ -> "audsley_algo"
          | AL_user _ -> "user_prio_algo" ]))
      True;
  }
  else ();
  rt_map_widget wid;
};

(* Main *)

value make_common xa xd = do {
  let raw_font =
    rt_load_query_font xd
      "-*-courier-bold-r-normal-*-10-*-*-*-*-*-iso8859-1"
  in
  let watch = rt_create_font_mouse xd xC_watch in
  let hand = rt_create_font_mouse xd xC_hand1 in
  let c_create_color rgb = (rt_create_color xd rgb, rgb) in
  let color_list =
(*
    [(0, 256, 0); (256, 0, 0); (0, 0, 256); (256, 0, 256);
*)
    [(256, 0, 0); (0, 0, 256); (0, 256, 0); (256, 0, 256);
(**)
     (0, 256, 256); (200, 200, 0); (256, 128, 0); (128, 0, 256);
     (128, 0, 128); (0, 128, 128);

     (128, 256, 0); (256, 128, 128); (128, 128, 256); (256, 128, 256);
     (128, 256, 256); (200, 200, 128); (128, 128, 128)]
  in
  let c_color = Array.of_list (List.map c_create_color color_list) in
  let w = screen_width xd in
  let h = screen_width xd in
  let c_white = rt_white_color xd in
  let pixmap_sched = rt_create_pixmap xd w (Dim.schedule_height 80) in
  let pixmap_dameid = rt_create_pixmap xd w h in
  let pixmap_caption = rt_create_pixmap xd w h in
  rt_select_color c_white;
(*
  rt_fill_rectangle (PixmapDr pixmap_sched) (0, 0, w, h);
*)
  {xd = xd;
   xa = xa;
   c_white_col = (c_white, (256, 256, 256));
   c_black_col = (rt_black_color xd, (0, 0, 0));
   c_red_col = c_create_color (255, 0, 0);
   c_blue_col = c_create_color (0, 0, 255);
   c_gray_col = c_create_color (150, 150, 150);
   c_backg_rep_col = c_create_color (216, 216, 255);
   c_raw_font = raw_font; c_color = c_color;
   pixmap_sched = pixmap_sched; pixmap_dameid = pixmap_dameid;
   pixmap_caption = pixmap_caption;
   watch = watch; hand = hand}
};

value rt_init name char_set = do {
  let xd = rt_initialize name in
  rt_select_char_set xd char_set;
  xd
};

value make_tab_info gd =
  let tab_labels =
    List.map (fun i -> gd.scenario.(i).fname)
      (Sutil.rho (Array.length gd.scenario))
  in
  Tabwid.tab_info "tab" gd.common.c_raw_font tab_labels
    (raise_scenario gd) (erase_scenario gd) (Array.length gd.scenario <= 1)
;

value go xd gd fail_if_err = do {
  let xa = rt_args [xd] in
  let tab_info = make_tab_info gd in
  make_dameid_widget xd xa gd tab_info;
  Sys.catch_break True;
  let curr_dir = Sys.getcwd () in
  loop () where rec loop () =
    try rt_main_loop xa with e ->
      if fail_if_err then do {
        Sys.chdir curr_dir;
        if gd.curr_sc >= 0 && gd.curr_sc < Array.length gd.scenario then do {
          let sc = gd.scenario.(gd.curr_sc) in
          Schedule.output_task_data "failed.dm" sc.ts;
        }
        else ();
        raise e
      }
      else do {
        Printf.eprintf "%s\n" (Printexc.to_string e);
        flush stderr;
        loop ();
      };
  Sys.chdir curr_dir;
};

value font_of_graph graph =
  fun 
  [ F_raw -> graph.raw_font
  | F_bigger -> graph.bigger_font
  | F_smaller -> graph.smaller_font ]
;

value display_dialog xd txt = do {
  let main_wid = widget_named xd "main" in
  let (rows, cols) = Dmisc.text_rows_cols txt in
  let wid =
    rt_create_transient_widget main_wid "error"
      (Some rt_destroy_widget)
      (term_desc [] (rows + 2, cols + 1, 0) (fun _ _ -> ()))
  in
  term_send wid Std_term.vt_bold;
  term_send wid Std_term.vt_newline_mode;
  term_send wid "\n";
  term_send wid txt;
  let (x, y) = Dutil.center_of_subwidget main_wid wid in
  rt_map_transient_widget wid x y
};

value g_fun xa xd =
  {g_record_style_bg_color (r16, g16, b16) = ();
   g_clear_drawable = da_clear_widget;
   g_commit_display = da_commit_display;
   g_draw_arc drw gc (x, y, w, h) (a, da) = do {
     da_select_paint gc.val;
     da_draw_arc drw (x, y, w, h, a, da)
   };
   g_draw_line drw gc (x1, y1) (x2, y2) = do {
     da_select_paint gc.val;
     da_draw_line drw (x1, y1) (x2, y2);
   };
   g_draw_lines drw gc list = do {
     da_select_paint gc.val;
     da_draw_lines drw list;
   };
   g_draw_rectangle drw gc (x, y, w, h) = do {
     da_select_paint gc.val;
     da_draw_rectangle drw (x, y, w, h);
   };
   g_fill_arc drw gc (x, y, w, h) (a, da) = do {
     da_select_paint gc.val;
     da_fill_arc drw (x, y, w, h, a, da)
   };
   g_fill_polygon (_, _, wid) gc list = do {
     da_select_paint gc.val;
     rt_fill_polygon (WidgetDr wid) list
   };
   g_fill_rectangle drw gc (x, y, w, h) = do {
     da_select_paint gc.val;
     da_fill_rectangle drw (x, y, w, h);
   };
   g_draw_string drw gc (x, y) text = do {
     da_select_paint gc.val;
     da_draw_string drw (x, y) text
   };
   g_erase_draw_string drw gc (x, y) text = do {
     da_select_paint gc.val;
     da_erase_draw_string drw (x, y) text
   };
   g_dam_dim = da_drawable_dim;
   g_drawable_width = da_drawable_width;
   g_drawable_height = da_drawable_height;
   g_schedule_zoom_size graph =
     fun
     [ Zoom _ wid _ _ -> (widget_width wid, widget_height wid)
     | NoZoom -> (graph.dd.sched_width, graph.dd.sched_height) ]
   ;
   g_inter_task_dy (dd, _, _) = dd.inter_task_dy;
   g_set_color graph gc col = do {
     let col =
       match col with
       [ C_black -> graph.black_col
       | C_white -> graph.white_col
       | C_red -> graph.red_col
       | C_gray -> graph.gray
       | C_backg_rep_col -> graph.backg_rep_col
       | C_number i -> graph.color.(i mod Array.length graph.color)
       | C_style_bg -> graph.white_col ]
     in
     gc.val := DaColorPn col
   };
   g_set_dashes graph gc off list = do {
     da_set_dashes graph.dd off list
   };
   g_set_font graph gc font = do {
     let font = font_of_graph graph font in
     da_select_font font
   };
   g_set_line_width graph gc w = da_set_line_width graph.dd w;
   g_set_pixmap graph gc i = do {
     let pixm = graph.pixmap.(i mod Array.length graph.pixmap) in
     gc.val := DaPixmapPn pixm
   };
   g_text_width graph font str = do {
     let font = font_of_graph graph font in
     da_text_width font str
   };
   g_font_size graph font = do {
     let font = font_of_graph graph font in
     da_font_size font
   };
   g_xcoord_of_coord = Da.xcoord_of_coord;
   g_error exc = do {
     display_dialog xd (Printexc.to_string exc);
   };
   g_treat_pending_events () = do {
     while rt_pending_events xa do { rt_treat_one_event xa };
   }}
;

value start_empty fail_if_err = do {
  let xd = rt_init "" Utf_8 in
  let xa = rt_args [xd] in
  let co = make_common xa xd in
  let gd =
    {scenario = [| |]; ind_of_isc = [| |]; curr_sc = 0;
     open_dir = Sys.getcwd (); prefix_number = 0; last_zoom = 2;
     graph_nb = 0; common = co; g_fun = g_fun xa xd}
  in
  go xd gd fail_if_err;
  rt_end xd
};

value start_with_dameid_widget fnames algo_opt fail_if_err = do {
  let dfnl = Openfile.get_dameid_files fnames algo_opt in
  if dfnl = [] then exit 1
  else do {
    let dfnl =
      loop 7 dfnl where rec loop n =
        fun
        [ [dfn :: dfnl] -> if n = 0 then [] else [dfn :: loop (n - 1) dfnl]
        | [] -> [] ]
    in

    let tsnl =
      List.map (fun (df, fname) -> (Scenario.make_tasks df, fname)) dfnl
    in
    let xd = rt_init "" Utf_8 in
    let xa = rt_args [xd] in
    let co = make_common xa xd in
    let scs = List.map (Scenario_rt.make_scenario xd co) tsnl in
    let scenario = Array.of_list scs in
    let ind_of_isc = Array.init (Array.length scenario) (fun i -> i) in
    let gd =
      {scenario = scenario; ind_of_isc = ind_of_isc; curr_sc = 0;
       open_dir = Sys.getcwd (); prefix_number = 0; last_zoom = 2;
       graph_nb = 0; common = co; g_fun = g_fun xa xd}
    in
    go xd gd fail_if_err;
    rt_end xd
  };
};

value start fnames algo_opt fail_if_err = do {
  button_font.val := "-*-courier-*-r-normal-*-14-*-*-*-*-*-iso8859-*";
  title_font.val := "-*-times-medium-r-normal-*-18-*-*-*-*-*-iso8859-*";
  if fnames = [] then start_empty fail_if_err
  else start_with_dameid_widget fnames algo_opt fail_if_err
};
