(* $Id: mainwid.ml,v 1.76 2011-07-13 14:39:43 deraugla Exp $ *)

open Rt;

open Ast;
open Da;
open Data_rt;
open Glob_glob;
open Dtasks;

value button_release_action_after_events xa gd act wid =
  fun
  [ ButtonEvRelease x y _ -> do {
      while rt_pending_events xa do { rt_treat_one_event xa };
      if gd.curr_sc < Array.length gd.scenario then Expose.expose gd
      else ();
      act (x, y);
    }
  | _ -> () ]
;

value set_black_and_white xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    if sc.coloured then do {
      rt_freeze_widget (widget_named xd "set_black_and_white");
      rt_unfreeze_widget (widget_named xd "set_coloured");
      sc.coloured := False;
      Expose.expose gd
    }
    else ()
  else ()
;

value set_coloured xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    if not sc.coloured then do {
      rt_freeze_widget (widget_named xd "set_coloured");
      rt_unfreeze_widget (widget_named xd "set_black_and_white");
      sc.coloured := True;
      Expose.expose gd
    }
    else ()
  else  ()
;

value set_schedule xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    rt_freeze_widget (widget_named xd "set_schedule");
    rt_unfreeze_widget (widget_named xd "set_mesoid");
    rt_raise_widget (Dutil.g_widget_named gd "schedule");
  }
  else ()
;

value set_mesoid xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    rt_freeze_widget (widget_named xd "set_mesoid");
    rt_unfreeze_widget (widget_named xd "set_schedule");
    rt_raise_widget (Dutil.g_widget_named gd "mesoid");
  }
  else ()
;

value set_hide_tabs xd ti _ = do {
  Tabwid.tab_hide xd ti;
  let main_wid = widget_named xd "main" in
  ignore (rt_adjust_widget main_wid : (int * int));
  rt_freeze_widget (widget_named xd "set_hide_tabs");
  rt_unfreeze_widget (widget_named xd "set_show_tabs");
};

value set_show_tabs xd ti _ = do {
  Tabwid.tab_show xd ti;
  let main_wid = widget_named xd "main" in
  ignore (rt_adjust_widget main_wid : (int * int));
  rt_freeze_widget (widget_named xd "set_show_tabs");
  rt_unfreeze_widget (widget_named xd "set_hide_tabs");
};

value change_order ta xa xd gd = do {
  let sc = gd.scenario.(gd.curr_sc) in
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let tc = ts.constraints in
  let new_task =
    Array.init (Array.length ts.tasks_def) (fun i -> ts.tasks_def.(ta.(i)))
  in
  Array.blit new_task 0 ts.tasks_def 0 (Array.length ts.tasks_def);
  let inv_ta = Array.init (Array.length ts.tasks_def) (fun i -> i) in
  for i = 0 to Array.length ts.tasks_def - 1 do { inv_ta.(ta.(i)) := i };
  sc.initial_order :=
    Array.of_list
      (List.map (fun i -> inv_ta.(i)) (Array.to_list sc.initial_order));
  let new_user_graph_precedence =
    List.map (List.map (fun i -> inv_ta.(i))) ts2.user_graph_precedence
  in
  ts2.user_graph_precedence := new_user_graph_precedence;
  let new_graph_precedence =
    List.map (List.map (fun i -> inv_ta.(i))) tc.graph_precedence
  in
  let tc = {(tc) with graph_precedence = new_graph_precedence} in
  let ts = {(ts) with constraints = tc} in
  ts2.ast := ts;
  sc.new_graph_precedence := new_graph_precedence;
  let new_latency =
    List.map
      (fun l ->
         let lc = l.l_constr in
         let lc =
           {(lc) with
            l_task1 = inv_ta.(lc.l_task1); l_task2 = inv_ta.(lc.l_task2)}
         in
         {l_constr = lc; l_status = LS_unchecked})
      ts2.latency_constr
  in
  ts2.latency_constr := new_latency;
  sc.new_latency_constr := new_latency;
  Scenario.set_task_order ts2;
(*
  goto_begin gd (Dutil.g_widget_named gd "dameid")
    (Dutil.g_widget_named gd "schedule");
*)
  if ts2.curr_ord_ind = Array.length ts.tasks_def &&
     match ts2.schedule_status with
     [ SS_running | SS_schedulable_but _ -> True
     | _ -> False ]
  then do {
    for i = 0 to ts2.schedule_len - 1 do {
      match Sarray.get ts2.disp_schedule i with
      [ Some u ->
          Sarray.set ts2.disp_schedule i
            (Some {(u) with u_task = inv_ta.(u.u_task)})
      | None -> () ];
    };
    Expose.redisplay_all gd;
  }
  else
    Schedule_step_rt.goto_begin gd;
(**)
  rt_unfreeze_widget (widget_named xd "sort_priority");
  rt_unfreeze_widget (widget_named xd "sort_inc_priority");
  rt_unfreeze_widget (widget_named xd "sort_release");
  rt_unfreeze_widget (widget_named xd "initial_order");
};

value sort_by_priority xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let ta = Array.init (Array.length ts.tasks_def) (fun i -> i) in
    Array.sort (Sparse.compare_priorities ts) ta;
    change_order ta xa xd gd;
    rt_freeze_widget (widget_named xd "sort_priority");
  }
  else ()
;

value sort_by_inc_priority xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let ta = Array.init (Array.length ts.tasks_def) (fun i -> i) in
    Array.sort (fun i1 i2 -> Sparse.compare_priorities ts i2 i1) ta;
    change_order ta xa xd gd;
    rt_freeze_widget (widget_named xd "sort_inc_priority");
  }
  else ()
;

value sort_by_release xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let ta = Array.init (Array.length ts.tasks_def) (fun i -> i) in
    Array.sort
      (fun i1 i2 ->
         match
           compare ts.tasks_def.(i1).user_release
             ts.tasks_def.(i2).user_release
         with
         [ 0 -> i1 - i2
         | n -> n ])
      ta;
    change_order ta xa xd gd;
    rt_freeze_widget (widget_named xd "sort_release");
  }
  else ()
;

value sort_by_initial_order xa xd gd _ = do {
  let sc = gd.scenario.(gd.curr_sc) in
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let ta =
    Array.init (Array.length ts.tasks_def) (fun i -> sc.initial_order.(i))
  in
  change_order ta xa xd gd;
  rt_freeze_widget (widget_named xd "initial_order");
};

value switch_prec_constr xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let tc = ts.constraints in
    let wid = widget_named xd "prec_constr" in
    let v = not (check_val wid) in
    check_set wid v;
    let tc = {(tc) with global_precedence = v} in
    let ts = {(ts) with constraints = tc} in
    ts2.ast := ts;
    for itask = 0 to Array.length ts.tasks_def - 1 do {
      let td = ts.tasks_def.(itask) in
      let td = {(td) with release = if v then 0 else td.user_release} in
      ts.tasks_def.(itask) := td;
    };
    Scenario.set_task_order ts2;
    Schedule_step_rt.goto_begin gd;
  }
  else ()
;

value switch_strict_per_constr xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let wid = widget_named xd "strict_per_constr" in
    let v = not (check_val wid) in
    check_set wid v;
    sc.ts.strict_periodicity := v;
    Schedule_step_rt.goto_begin gd;
  }
  else ()
;

value switch_non_preemptive_algo_constr xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let wid = widget_named xd "no_preempt_constr" in
    let v = not (check_val wid) in
    check_set wid v;
    sc.ts.fully_non_preemptive := v;
    Schedule_step_rt.goto_begin gd;
  }
  else ()
;

value set_algo xa xd gd sc algo = do {
  let ts = {(sc.ts.ast) with algorithm = algo} in
  sc.ts.ast := ts;
  Scenario.set_task_order sc.ts;
  Schedule_step_rt.goto_begin gd;
  check_set (widget_named xd "dead_mono_algo") False;
  check_set (widget_named xd "rate_mono_algo") False;
  check_set (widget_named xd "edf_algo") False;
  check_set (widget_named xd "audsley_algo") False;
  check_set (widget_named xd "user_prio_algo") False;
};

value set_dead_mono_algo xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    set_algo xa xd gd sc AL_deadline_monotonic;
    check_set (widget_named xd "dead_mono_algo") True;
  }
  else ()
;

value set_rate_mono_algo xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    set_algo xa xd gd sc AL_rate_monotonic;
    check_set (widget_named xd "rate_mono_algo") True;
  }
  else ()
;

value set_edf_algo xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    set_algo xa xd gd sc AL_edf;
    check_set (widget_named xd "edf_algo") True;
  }
  else ()
;

value set_audsley_algo xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let algo =
      let t = Array.init (Array.length sc.ts.ast.tasks_def) (fun i -> i) in
      let auds = {ai_pos = P_begin; ai_tab = t; ai_step = 0; ai_nsol = 0} in
      AL_audsley auds
    in
    set_algo xa xd gd sc algo;
    check_set (widget_named xd "audsley_algo") True;
  }
  else ()
;

value set_user_prio_algo xa xd gd _ =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let algo = do {
      let a =
        Array.init (Array.length sc.ts.ast.tasks_def) (fun i -> max_int)
      in
      ignore
        (List.fold_left (fun p i -> do { a.(i) := p; p + 1 }) 1
           sc.new_user_priority);
      AL_user a
    }
    in
    set_algo xa xd gd sc algo;
    check_set (widget_named xd "user_prio_algo") True;
  }
  else ()
;

value default_about_dameid () =
  let lang = I18n.lang.val in
  if String.length lang >= 2 && String.sub lang 0 2 = "fr" then
    "Dameid\n\
     \n  \
       Conception: Patrick Meumeu Yomsi\n  \
       Implementation: Daniel de Rauglaudre\n\
     \n\
     Copyright (c) INRIA 2008-2009"
  else
    "Dameid\n\
     \n  \
       Concepts: Patrick Meumeu Yomsi\n  \
       Implementation: Daniel de Rauglaudre\n\
     \n\
     Copyright (c) INRIA 2008-2009"
;

value display_info title s xd gd = do {
  let main_wid = widget_named xd "main" in
  let (rows, cols) = Dmisc.text_rows_cols s in
  let wid =
    rt_create_transient_widget main_wid title
      (Some rt_destroy_widget)
      (term_desc [] (rows + 2, cols + 1, 0) (fun _ _ -> ()))
  in
  term_send wid Std_term.vt_bold;
  term_send wid Std_term.vt_newline_mode;
  term_send wid "\n";
  term_send wid s;
  let (x, y) = Dutil.center_of_subwidget main_wid wid in
  rt_map_transient_widget wid x y
};

value display_about_dameid xd gd _ =
  let text = default_about_dameid () in
  display_info (I18n.transl "About dameid") text xd gd
;

value display_how_to_zoom xd gd _ =
  let text =
    match I18n.ptransl "~how to zoom" with
    [ Some s -> s
    | None -> Dmisc.default_zoom_help ]
  in
  display_info (I18n.transl "How to zoom") text xd gd
;

value build_ps tg oc = do {
  let sc = tg.scenario.(tg.curr_sc) in
  let gc = ref DaNonePn in
  let expose () = do {
    let gfun = tg.g_fun in
    let dam_wid = Dutil.g_widget_named tg "dameid" in
    let cap_wid = Dutil.g_widget_named tg "caption" in
    let sched_wid = Dutil.g_widget_named tg "schedule" in
    let dam_drw =
      Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
    in
    Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
    Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
    Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom False;
    let cap_drw = Draw_dameid_rt.caption_drawable cap_wid sc.graph.dd in
    Draw_dameid.draw_caption gfun sc cap_drw gc;
    Draw_schedule_rt.expose_schedule gfun tg tg.curr_sc sched_wid;
    Expose.draw_text sc (Dutil.g_widget_named tg "text");
  }
  in
  postscript_save expose sc.graph.dd oc;
};

value sh_rft = ref 0;
value rec action_main_key_press xa tg ti main_wid k =
  let xd = xdata_of_widget main_wid in
  let def_keyp = action_main_key_press xa tg ti in
  match k with
  [ K_Ascii 'b' -> set_black_and_white xd tg ()
  | K_Ascii 'c' -> set_coloured xd tg ()
  | K_Ascii 'g' ->
      let wid = widget_named xd "main" in
      let x = widget_x wid in
      let y = widget_y wid in
      let w = widget_width wid in
      let h = widget_height wid in
      let wid = widget_named xd "main" in
      Save.print_ps xd wid tg "dameid.ps" (build_ps tg) (fun () -> ())
        (Some (x + w / 2, y + h / 2)) (0, 0)
  | K_Ascii 'h' -> set_schedule xd tg ()
  | K_Ascii 'm' -> set_mesoid xd tg ()
  | K_Ascii 'n' -> do {
      let len = Array.length tg.scenario in
      if len < 7 then do {
        let isc = if len = 0 then 0 else tg.ind_of_isc.(len-1) + 1 in
        let sc =
          Scenario_rt.empty_scenario xd tg.common (Printf.sprintf "%d.dm" isc)
        in
        tg.scenario := Array.append tg.scenario [| sc |];
        tg.ind_of_isc := Array.append tg.ind_of_isc [| isc |];
        tg.curr_sc := len;
        let wdesc = Scenario_rt.scenario_desc xa tg ti def_keyp len in
        Tabwid.tab_add xd ti sc.fname wdesc;
        if Array.length tg.scenario > 1 then Tabwid.tab_show xd ti
        else Tabwid.tab_hide xd ti;
        let (_, _) = rt_adjust_widget main_wid in
        rt_raise_widget (widget_named xd "main_tab");
      }
      else ()
    }
  | K_Ascii 'o' ->
      let wid = widget_named xd "main" in
      let x = widget_x wid in
      let y = widget_y wid in
      let w = widget_width wid in
      let h = widget_height wid in
      Dameidfile.open_file xa tg ti def_keyp (Some (x + w / 2, y + h / 2))
        (0, 0)
  | K_Ascii 'p' ->
      let wid = widget_named xd "main" in
      let x = widget_x wid in
      let y = widget_y wid in
      let w = widget_width wid in
      let h = widget_height wid in
      Parameters_rt.display_change_parameters xa xd tg ti def_keyp
        (Some (x + w / 2, y + h / 2)) (0, 0)
  | K_Ascii 'r' ->
      Expose.redisplay_all tg
  | K_Ascii 's' ->
      let wid = widget_named xd "main" in
      let x = widget_x wid in
      let y = widget_y wid in
      let w = widget_width wid in
      let h = widget_height wid in
      Save.save_as xd tg (Some (x + w / 2, y + h / 2)) (0, 0)
  | K_Ascii 't' ->
      let wid = widget_named xd "main" in
      let x = widget_x wid in
      let y = widget_y wid in
      let w = widget_width wid in
      let h = widget_height wid in
      Save.save_result xd tg (Some (x + w / 2, y + h / 2)) (0, 0)
  | K_Ascii 'z' ->
      display_how_to_zoom xd tg ()
  | K_Ascii '[' ->
      set_hide_tabs tg.common.xd ti ()
  | K_Ascii ']' ->
      set_show_tabs tg.common.xd ti ()
  | K_Ascii ' ' | K_Right ->
      Schedule_step_rt.forward xa tg
  | K_Ascii '/' | K_KP_Divide ->
      if tg.curr_sc < Array.length tg.scenario then do {
        sh_rft.val := (sh_rft.val + 10) mod 100;
        let wid = widget_named xd "main" in
        let x = widget_x wid + widget_width wid / 2 + sh_rft.val in
        let y = widget_y wid + widget_height wid / 2 + sh_rft.val in
        Draw_resp_f_time.response_f_time tg xd (Some (x, y)) (0, 0)
      }
      else ()
  | K_Prior -> Schedule_step_rt.backward_task tg
  | K_Tab | K_Next -> Schedule_step_rt.forward_task xa tg
  | K_BackSpace | K_Left -> Schedule_step_rt.backward tg
  | K_Home | K_Up -> Schedule_step_rt.goto_prev xa tg
  | K_End | K_Down -> Schedule_step_rt.goto_end xa tg
  | _ -> () ]
;

value action_save_ps xd gd _ =
  fun
  [ ButtonEvRelease x y _ ->
      let wid = widget_named xd "main" in
      Save.print_ps xd wid gd "dameid.ps" (build_ps gd) (fun () -> ()) None
        (x, y)
  | _ -> () ]
;

value make_file_popup xa xd tg ti backg foreg =
  let def_keyp = action_main_key_press xa tg ti in
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "Open" ^ "...\t[o]", None)
           (Dutil.button_release_action
              (Dameidfile.open_file xa tg ti def_keyp None));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Save parameters" ^ "\t[s]", None)
           (Dutil.button_release_action (Save.save_as xd tg None));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Save resulting schedule" ^ "\t[t]", None)
           (Dutil.button_release_action (Save.save_result xd tg None));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Print in PostScript" ^ "\t[g]", None)
           (action_save_ps xd tg);
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Refresh" ^ "\t[r]", None)
           (Dutil.button_release_action (fun _ -> Expose.redisplay_all tg));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Quit" ^ "\t[ctl-q]", None)
           (Dutil.button_release_action (fun _ -> rt_stop_main_loop xa))])
       (fun _ _ -> ()))
;

value make_edit_popup xa xd tg ti backg foreg =
  let def_keyp = action_main_key_press xa tg ti in
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "Change parameters" ^ "\t[p]", None)
           (Dutil.button_release_action
              (Parameters_rt.display_change_parameters xa xd tg ti def_keyp
                 None))])
       (fun _ _ -> ()))
;

value make_disp_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc
           [NameAtt "sort_priority";  LeftJustifAtt; backg]
           (I18n.transl "Sort by decreasing priority", None)
           (Dutil.button_release_action (sort_by_priority xa xd gd));
         button_desc
           [NameAtt "sort_inc_priority";  LeftJustifAtt; backg]
           (I18n.transl "Sort by increasing priority", None)
           (Dutil.button_release_action (sort_by_inc_priority xa xd gd));
         button_desc
           [NameAtt "sort_release";  LeftJustifAtt; backg]
           (I18n.transl "Sort by release date", None)
           (Dutil.button_release_action (sort_by_release xa xd gd));
         button_desc
           [NameAtt "initial_order";  LeftJustifAtt; backg]
           (I18n.transl "Sort according to initial order", None)
           (Dutil.button_release_action (sort_by_initial_order xa xd gd));
         line_desc [backg] () (fun _ _ -> ());
         button_desc
           [NameAtt "set_mesoid";  LeftJustifAtt; backg]
           (I18n.transl "Mesoid" ^ "\t[m]", None)
           (Dutil.button_release_action (set_mesoid xd gd));
         button_desc
           [NameAtt "set_schedule";  LeftJustifAtt; backg]
           (I18n.transl "Scheduling" ^ "\t[h]", None)
           (Dutil.button_release_action (set_schedule xd gd));
         line_desc [backg] () (fun _ _ -> ());
         button_desc
           [NameAtt "set_black_and_white";  LeftJustifAtt; backg]
           (I18n.transl "Black & White" ^ "\t[b]", None)
           (Dutil.button_release_action (set_black_and_white xd gd));
         button_desc
           [NameAtt "set_coloured"; LeftJustifAtt; backg]
           (I18n.transl "Color" ^ "\t[c]", None)
           (Dutil.button_release_action (set_coloured xd gd));
         line_desc [backg] () (fun _ _ -> ());
         button_desc
           [NameAtt "set_show_tabs";  LeftJustifAtt; backg]
           (I18n.transl "Show tabs" ^ "\t[]]", None)
           (Dutil.button_release_action (set_show_tabs xd ti));
         button_desc
           [NameAtt "set_hide_tabs"; LeftJustifAtt; backg]
           (I18n.transl "Hide tabs" ^ "\t[[]", None)
           (Dutil.button_release_action (set_hide_tabs xd ti))])
       (fun _ _ -> ()))
;

value make_constr_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "prec_constr"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Precedence", None)
               (Dutil.button_release_action (switch_prec_constr xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "strict_per_constr"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Strict Periodicity", None)
               (Dutil.button_release_action
                  (switch_strict_per_constr xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "no_preempt_constr"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Fully Non Preemptive", None)
               (Dutil.button_release_action
                  (switch_non_preemptive_algo_constr xa xd gd))])
           (fun _ _ -> ())])
       (fun _ _ -> ()))
;

value make_algo_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "dead_mono_algo"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Deadline-Monotonic", None)
               (Dutil.button_release_action (set_dead_mono_algo xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "rate_mono_algo"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Rate-Monotonic", None)
               (Dutil.button_release_action (set_rate_mono_algo xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "edf_algo"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Earliest Deadline First", None)
               (Dutil.button_release_action (set_edf_algo xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "audsley_algo"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "Audsley++", None)
               (Dutil.button_release_action (set_audsley_algo xa xd gd))])
           (fun _ _ -> ());
         pack_desc [InterAtt 10; backg; BandAtt 0]
           (Horizontal,
            [check_desc [NameAtt "user_prio_algo"; backg; foreg] ()
               (fun _ _ -> ());
             button_desc [LeftJustifAtt; FillerAtt; backg]
               (I18n.transl "User Priorities", None)
               (Dutil.button_release_action (set_user_prio_algo xa xd gd))])
           (fun _ _ -> ())])
       (fun _ _ -> ()))
;

value make_dam_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "Clear" ^ "\t[" ^ I18n.transl "Home" ^ "]", None)
           (Dutil.button_release_action
              (fun _ -> Schedule_step_rt.goto_begin gd));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "One step back" ^ "\t[" ^ I18n.transl "Backspace" ^
              "]",
            None)
           (Dutil.button_release_action
              (fun _ -> Schedule_step_rt.backward gd));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "One step" ^ "\t[" ^ I18n.transl "Space" ^ "]",
            None)
           (Dutil.button_release_action
              (fun _ -> Schedule_step_rt.forward xa gd));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "One task back" ^ "\t[" ^ I18n.transl "Page Up" ^
              "]",
            None)
           (Dutil.button_release_action
              (fun _ -> Schedule_step_rt.backward_task gd));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "One task" ^ "\t[" ^ I18n.transl "Page Down" ^ "]",
            None)
           (Dutil.button_release_action
              (fun _ -> Schedule_step_rt.forward_task xa gd));
         button_desc [LeftJustifAtt; backg]
           (I18n.transl "Schedule" ^ "\t[" ^ I18n.transl "End" ^ "]", None)
           (button_release_action_after_events xa gd
              (fun _ -> Schedule_step_rt.goto_end xa gd));
         button_desc
           [LeftJustifAtt; backg]
           (I18n.transl "Response" ^ "=f(" ^
            I18n.transl "time" ^ ")" ^ "\t[/]", None)
           (Dutil.button_release_action
              (Draw_resp_f_time.response_f_time gd xd None))])
       (fun _ _ -> ()))
;

value make_help_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "How to zoom" ^ "\t[z]", None)
           (Dutil.button_release_action (display_how_to_zoom xd gd))])
       (fun _ _ -> ()))
;

value make_about_popup xa xd gd ti backg foreg =
  rt_create_popup_widget xd
    (pack_desc [backg]
       (Vertical,
        [button_desc [LeftJustifAtt; backg]
           (I18n.transl "About Dameid", None)
           (Dutil.button_release_action (display_about_dameid xd gd))])
       (fun _ _ -> ()))
;
