(* $Id: parameters.ml,v 1.157 2010-12-07 11:08:22 deraugla Exp $ *)

open Ast;
open Dtasks;
open Glob_glob;

value ok_change_parameters sc = do {
  let ts = sc.ts in
  let tc = ts.ast.constraints in
  let new_ts_task =
    Array.map
      (fun td -> if tc.global_precedence then {(td) with release = 0} else td)
      sc.new_task
  in
  let tc = {(tc) with preemption_cost = sc.new_preemption_cost} in
  let ast = {(ts.ast) with tasks_def = new_ts_task; constraints = tc} in
  ts.ast := ast;
  ts.task_order := Array.init (Array.length sc.new_task) (fun i -> i);
  sc.initial_order := Array.init (Array.length ast.tasks_def) (fun i -> i);
  let lcm =
    Sutil.lcm_of_list
      (List.map (fun td -> td.period) (Array.to_list ts.ast.tasks_def))
  in
  ts.lcm := lcm;
  Schedule.recompute_init ts;
  sc.latency_disp_cnt := 0;
  ts.user_graph_precedence := sc.new_graph_precedence;
  let nb_task = Array.length ts.ast.tasks_def in
  let tc = ts.ast.constraints in
  let tc =
    {(tc) with
     graph_precedence =
       Sparse.genuine_graph_precedence nb_task sc.new_graph_precedence}
  in
  let ast = {(ts.ast) with constraints = tc} in
  ts.ast := ast;
  ts.latency_constr := sc.new_latency_constr;
  Scenario.set_task_order ts;
  match ts.ast.algorithm with
  [ AL_user _ -> do {
      let a = Array.init nb_task (fun i -> max_int) in
      ignore
        (List.fold_left (fun p i -> do { a.(i) := p; p + 1 }) 1
           sc.new_user_priority);
      let ast = {(ts.ast) with algorithm = AL_user a} in
      ts.ast := ast;
    }
  | AL_audsley _ -> do {
      let t = Array.init (Array.length sc.ts.ast.tasks_def) (fun i -> i) in
      let auds = {ai_pos = P_begin; ai_tab = t; ai_step = 0; ai_nsol = 0} in
      let ast = {(ts.ast) with algorithm = AL_audsley auds} in
      ts.ast := ast;
    }
  | AL_rate_monotonic | AL_deadline_monotonic | AL_edf -> () ];
};
