(* $Id: parameters_rt.ml,v 1.6 2011-07-13 14:39:43 deraugla Exp $ *)

open Printf;
open Rt;
open Scanf;

open Ast;
open Da;
open Data_rt;
open Dim_def;
open Dtasks;
open Glob_glob;
open Std_term;
open Text_rt;

value set_lcm_and_util_fact xd gd = do {
  let lcm =
    Sutil.lcm_of_list
      (List.map (fun td -> td.period) (Array.to_list gd.new_task))
  in
  let wid = widget_named xd "lcm" in
  term_send wid
    (sprintf "%s%s%s%12d" vt_home vt_erase_from_cursor vt_hide_cursor lcm);
  let util_fact_num =
    loop 0 0 where rec loop itask util_fact_num =
      if itask = Array.length gd.new_task then util_fact_num
      else
        let td = gd.new_task.(itask) in
        let util_fact_num =
          util_fact_num + td.duration * (lcm / td.period)
        in
        loop (itask + 1) util_fact_num
  in
  let wid = widget_named xd "util_fact" in
  let util_fact = float util_fact_num /. float lcm in
  term_send wid
    (sprintf "%s%s%s%s%.4f" vt_home vt_erase_from_cursor vt_hide_cursor
       (if util_fact > 1.0 then vt_color 256 0 0 ^ vt_bold
        else vt_no_color ^ vt_normal)
       util_fact);
};

value extract_text_value wid =
  let s = get_term_contents wid in
  loop (String.length s - 1) where rec loop i =
    if i < 0 then ""
    else if s.[i] = ' ' then loop (i - 1)
    else String.sub s 0 (i + 1)
;

value extract_value wid =
  let txt = extract_text_value wid in
  try sscanf txt " %d" (fun x -> x) with
  [ End_of_file -> 0 ]
;

value right_adjust_1 wid s =
(*
  let len = snd (term_size wid) - 1 in
  if String.length s >= len then s
  else String.make (len - String.length s) ' ' ^ s
*)s
;

value set_parameters xd sc = do {
  let pot = sc.iparam_of_itask in
  for i = 0 to Array.length sc.new_task - 1 do {
    let td = sc.new_task.(i) in
    let wid = widget_named xd (sprintf "rel_%d" pot.(i)) in
    let release = td.user_release in
    term_send wid (sprintf "%s%s" vt_home vt_erase_from_cursor);
    term_send wid (right_adjust_1 wid (string_of_int release));
    if td.sub_duration = [] && not sc.show_sub.(i) then ()
    else do {
      let wid = widget_named xd (sprintf "sub_rel_%d_0" pot.(i)) in
      term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
      term_send wid (right_adjust_1 wid (string_of_int 0));
      let wid = widget_named xd (sprintf "sub_dur_%d_0" pot.(i)) in
      term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
      let fst_dur = Sparse.first_duration td in
      term_send wid (right_adjust_1 wid (string_of_int fst_dur));
      ignore
        (List.fold_left
           (fun isub sd -> do {
              let wid =
                widget_named xd (sprintf "sub_rel_%d_%d" pot.(i) isub)
              in
              term_send wid
                (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
              term_send wid (right_adjust_1 wid (string_of_int sd.sub_rel));
              let wid =
                widget_named xd (sprintf "sub_dur_%d_%d" pot.(i) isub)
              in
              term_send wid
                (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
              term_send wid (right_adjust_1 wid (string_of_int sd.sub_dur));
              let wid =
                widget_named xd (sprintf "nonpre_%d_%d" pot.(i) isub)
              in
              check_set wid sd.sub_non_pree;
              let wid =
                widget_named xd (sprintf "strictper_%d_%d" pot.(i) isub)
              in
              check_set wid sd.sub_str_per;
              isub + 1
            })
           1 td.sub_duration : int)
    };
    let wid = widget_named xd (sprintf "dur_%d" pot.(i)) in
    term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
    term_send wid (right_adjust_1 wid (string_of_int td.duration));
    let wid = widget_named xd (sprintf "dea_%d" pot.(i)) in
    term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
    term_send wid (right_adjust_1 wid (string_of_int td.deadline));
    let wid = widget_named xd (sprintf "per_%d" pot.(i)) in
    term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
    term_send wid (right_adjust_1 wid (string_of_int td.period));
    if td.non_preemptive then
      check_set (widget_named xd (sprintf "nonpre_%d" pot.(i))) True
    else ();
    if td.strict_per then
      check_set (widget_named xd (sprintf "strictper_%d" pot.(i))) True
    else ();
    let wid = widget_named xd (sprintf "preecost_%d" pot.(i)) in
    term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
    term_send wid (right_adjust_1 wid (string_of_int td.pree_cost));
  };
  let wid = widget_named xd "file_name" in
  term_send wid (sprintf "%s%s%s" vt_home vt_erase_from_cursor vt_bold);
  term_send wid sc.fname;
  let wid = widget_named xd "pre_cost" in
  term_send wid
    (sprintf "%s%s%s%4d" vt_home vt_erase_from_cursor vt_bold
       sc.new_preemption_cost);
  let wid = widget_named xd "graph_prec_text" in
  term_send wid
    (sprintf "%s%s%s%s" vt_home vt_erase_from_cursor vt_bold vt_insert_mode);
  term_send wid (Schedule.string_of_graph_prec sc.new_graph_precedence);
  let wid = widget_named xd "latency_text" in
  term_send wid
    (sprintf "%s%s%s%s" vt_home vt_erase_from_cursor vt_bold vt_insert_mode);
  term_send wid (Schedule.string_of_latency sc.new_latency_constr);
  let wid = widget_named xd "priority_text" in
  term_send wid
    (sprintf "%s%s%s%s" vt_home vt_erase_from_cursor vt_bold vt_insert_mode);
  term_send wid (Schedule.string_of_user_priority sc.new_user_priority);
  set_lcm_and_util_fact xd sc;
};

value get_parameters xd sc adjust_deadline = do {
(*
let adjust_deadline = False in
*)
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let tc = ts.constraints in
  let pot = sc.iparam_of_itask in
  for itask = 0 to Array.length sc.new_task - 1 do {
    let ntd = sc.new_task.(itask) in
    let user_release =
      let wid = widget_named xd (sprintf "rel_%d" pot.(itask)) in
      extract_value wid
    in
    let release =
      if tc.global_precedence then ntd.release else user_release
    in
    let duration =
      let wid = widget_named xd (sprintf "dur_%d" pot.(itask)) in
      let v = extract_value wid in
      if v < 1 then ntd.duration else v
    in
    let sub_dur =
      loop [] 1 ntd.sub_duration where rec loop rev_sub_duration i =
        fun
        [ [] -> List.rev rev_sub_duration
        | [sd :: sdl] ->
            let wid = widget_named xd (sprintf "sub_rel_%d_%d" itask i) in
            let sub_rel = extract_value wid in
            let wid = widget_named xd (sprintf "sub_dur_%d_%d" itask i) in
            let sub_dur = extract_value wid in
            let wid = widget_named xd (sprintf "nonpre_%d_%d" itask i) in
            let sub_non_pree = check_val wid in
            let wid = widget_named xd (sprintf "strictper_%d_%d" itask i) in
            let sub_str_per = check_val wid in
            let sd =
              {sub_rel = sub_rel; sub_dur = sub_dur;
               sub_non_pree = sub_non_pree; sub_str_per = sub_str_per}
            in
            loop [sd :: rev_sub_duration] (i + 1) sdl ]
    in
    let deadline =
      let wid = widget_named xd (sprintf "dea_%d" pot.(itask)) in
      let v = extract_value wid in
      if v < 1 then ntd.period else v
    in
    let period =
      let wid = widget_named xd (sprintf "per_%d" pot.(itask)) in
      let v = extract_value wid in
      if v < 1 then ntd.period else v
    in
    let deadline =
      if adjust_deadline then min deadline period else deadline
    in
    let non_preemptive =
      let wid = widget_named xd (sprintf "nonpre_%d" pot.(itask)) in
      check_val wid
    in
    let strict_per =
      let wid = widget_named xd (sprintf "strictper_%d" pot.(itask)) in
      check_val wid
    in
    let pree_cost =
      let wid = widget_named xd (sprintf "preecost_%d" pot.(itask)) in
      let v = extract_value wid in
      if v < 1 then ntd.pree_cost else v
    in
    sc.new_task.(itask) :=
      {release = release; duration = duration; sub_duration = sub_dur;
       deadline = deadline; period = period; user_release = user_release;
       non_preemptive = non_preemptive; strict_per = strict_per;
       pree_cost = pree_cost};
  };
  set_lcm_and_util_fact xd sc;
  let wid = widget_named xd "pre_cost" in
  let v = extract_value wid in
  sc.new_preemption_cost := if v < 0 then sc.new_preemption_cost else v;
  let wid = widget_named xd "graph_prec_text" in
  let v = get_term_contents wid in
  match Sparse.parse_graph_prec_eoi (Stream.of_string v) with
  [ Some v -> sc.new_graph_precedence := v
  | None -> () ];
  let wid = widget_named xd "latency_text" in
  let v = get_term_contents wid in
  match Sparse.parse_latency_list_eoi (Stream.of_string v) with
  [ Some v ->
      sc.new_latency_constr :=
        List.map (fun l -> {l_constr = l; l_status = LS_unchecked}) v
  | None -> () ];
  let wid = widget_named xd "priority_text" in
  let v = get_term_contents wid in
  match Sparse.parse_user_priority_eoi (Stream.of_string v) with
  [ Some v -> sc.new_user_priority := v
  | None -> () ];
};

value set_parameter_widget xd sc = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let tc = ts.constraints in
  let pot = sc.iparam_of_itask in
  let freeze_unfreeze =
    if tc.global_precedence then rt_freeze_widget else rt_unfreeze_widget
  in
  for i = 0 to Array.length sc.new_task - 1 do {
    let wid = widget_named xd (sprintf "rel_%d" pot.(i)) in
    if tc.global_precedence then do {
      term_send wid vt_hide_cursor;
      term_send wid vt_normal_italic;
    }
    else do {
      term_send wid vt_show_cursor;
      term_send wid vt_bold;
    };
    freeze_unfreeze wid;
    let wid = widget_named xd (sprintf "rel_p_%d" pot.(i)) in
    freeze_unfreeze wid;
    let wid = widget_named xd (sprintf "rel_m_%d" pot.(i)) in
    freeze_unfreeze wid;
(*
    let wid = widget_named xd (sprintf "sub_rel_%d_0" pot.(i)) in
    term_send wid vt_hide_cursor;
    let wid = widget_named xd (sprintf "sub_rel_m_%d_0" pot.(i)) in
    rt_freeze_widget wid;
    let wid = widget_named xd (sprintf "sub_rel_p_%d_0" pot.(i)) in
    rt_freeze_widget wid;
*)
  };
  if tc.global_precedence then do {
    let wid = widget_named xd "param_mess" in
    term_send wid (vt_home ^ vt_erase_from_cursor ^ vt_hide_cursor);
    term_send wid
      ("(" ^
       I18n.transl "releases inhibited because of precedence constraint" ^
       ")");
    rt_raise_widget wid;
  }
  else rt_raise_widget (widget_named xd "no_param_mess");
  set_parameters xd sc;
};

value change_parameters xd gd ul_xy (x, y) = do {
  set_parameter_widget xd gd;
  let wid = widget_named xd "param" in
  let (x, y) =
    match ul_xy with
    [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
    | None -> (x, y) ]
  in
  rt_freeze_widget (widget_named xd "main");
  rt_dont_redirect_key_press xd;
  rt_map_transient_widget wid x y
};

value itask_of_iparam gd iparam =
  loop 0 where rec loop itask =
    if itask = Array.length gd.iparam_of_itask then itask
    else if gd.iparam_of_itask.(itask) = iparam then itask
    else loop (itask + 1)
;

value plus_or_minus sign sc typ iparam isub wid = do {
  let itask = itask_of_iparam sc iparam in
  let xd = xdata_of_widget wid in
  get_parameters xd sc True;
  match typ with
  [ "dea" -> do {
      let td = sc.new_task.(itask) in
      let v = max 1 (td.deadline + sign) in
      let p = td.period in
      let td = {(td) with deadline = v} in
      sc.new_task.(itask) := td;
      if p < v then sc.new_task.(itask) := {(td) with period = v}
      else ()
    }
  | "dur" ->
      let td = sc.new_task.(itask) in
      let v = max 1 (td.duration + sign) in
      sc.new_task.(itask) := {(td) with duration = v}
  | "per" -> do {
      let td = sc.new_task.(itask) in
      let v = max 1 (td.period + sign) in
      sc.new_task.(itask) := {(td) with deadline = v; period = v};
    }
  | "rel" ->
      let td = sc.new_task.(itask) in
      let v = td.user_release + sign in
      sc.new_task.(itask) := {(td) with release = v; user_release = v}
  | "preecost" ->
      let td = sc.new_task.(itask) in
      let v = max 0 (td.pree_cost + sign) in
      sc.new_task.(itask) := {(td) with pree_cost = v}
  | "pre_cost" ->
      let v = sc.new_preemption_cost + sign in
      sc.new_preemption_cost := max 0 v
  | "sub_rel" | "sub_dur" -> do {
      let td = sc.new_task.(itask) in
      let (rev_sub_dur, _) =
        List.fold_left
          (fun (rev_sdl, i) sd ->
             let sd =
               if i = isub then
                 if typ = "sub_rel" then
                   let d = sd.sub_rel + sign in
                   {(sd) with sub_rel = d}
                 else
                   let d = max 0 (sd.sub_dur + sign) in
                   {(sd) with sub_dur = d}
               else sd
             in
             ([sd :: rev_sdl], i + 1))
          ([], 1) td.sub_duration
      in
      let sub_dur_sum =
        List.fold_left (fun s sd -> s + sd.sub_dur) 0 rev_sub_dur
      in
      sc.new_task.(itask) :=
        {(td) with
         duration = max sc.new_task.(itask).duration sub_dur_sum;
         sub_duration = List.rev rev_sub_dur}
    }
  | s ->
      invalid_arg s ];
  set_parameters (xdata_of_widget wid) sc;
};

value unmap_change_parameters xd gd = do {
  let sc = gd.scenario.(gd.curr_sc) in
  get_parameters xd sc True;
  rt_destroy_widget (widget_named xd "param");
  Dutil.redirect_to_main gd;
  rt_unfreeze_widget (widget_named xd "main");
};

value gen_action_text is_sub_dur sc ok_action cancel_action wid ev =
  let xd = xdata_of_widget wid in
  match ev with
  [ TermEvButtonPress _ _ _ _ -> set_parameters xd sc
  | TermEvKeyPress {item = K_Escape} -> cancel_action ()
  | TermEvKeyPress {item = K_Return} -> ok_action ()
  | TermEvKeyPress {item = k} -> do {
      let txt = extract_text_value wid in
      set_parameters xd sc;
      let txt =
        match k with
        [ K_BackSpace ->
            if txt = "" then "0" else String.sub txt 0 (String.length txt - 1)
        | K_Ascii '-' | K_KP_Subtract ->
            if txt = "" then "-" else txt
        | K_Ascii ('0'..'9' as c) ->
            if txt = "0" then String.make 1 c
            else txt ^ String.make 1 c
        | K_KP_N i ->
            let c = Char.chr (Char.code '0' + i) in
            if txt = "0" then String.make 1 c
            else txt ^ String.make 1 c
        | s -> txt ]
      in
      term_send wid (sprintf "%s%s" vt_home vt_erase_from_cursor);
      term_send wid (right_adjust_1 wid txt);
      get_parameters xd sc False;
      if is_sub_dur then set_parameters xd sc else ();
    }
  | _ -> () ]
;

value action_text gd = gen_action_text False gd;
value action_sub_dur_text gd = gen_action_text True gd;

value action_prec_text ok_action cancel_action text_wid ev =
  match ev with
  [ TermEvKeyPress {item = k; control = False} ->
      match k with
      [ K_Ascii ('t' | '0'..'9' | '<' | ' ' as c) ->
          term_send text_wid (String.make 1 c)
      | K_KP_N i ->
          term_send text_wid (String.make 1 (Char.chr (Char.code '0' + i)))
      | K_Return -> ok_action ()
      | K_Escape -> cancel_action ()
      | _ -> std_term_ev text_wid ev ]
  | _ -> std_term_ev text_wid ev ]
;

value action_latency_text ok_action cancel_action text_wid ev =
  match ev with
  [ TermEvKeyPress {item = k; control = False} ->
      match k with
      [ K_Ascii ('t' | '0'..'9' | '(' | ')' | '-' | '>' | '<' | ' ' as c) ->
          std_term_insert text_wid c
      | K_KP_N i -> std_term_insert text_wid (Char.chr (Char.code '0' + i))
      | K_KP_Subtract -> std_term_insert text_wid '-'
      | K_Return -> ok_action ()
      | K_Escape -> cancel_action ()
      | _ -> std_term_ev text_wid ev ]
  | _ -> std_term_ev text_wid ev ]
;

value action_priority_text ok_action cancel_action text_wid ev =
  match ev with
  [ TermEvKeyPress {item = k; control = False} ->
      match k with
      [ K_Ascii ('t' | '0'..'9' | ',' | ' ' as c) ->
          std_term_insert text_wid c
      | K_KP_N i -> std_term_insert text_wid (Char.chr (Char.code '0' + i))
      | K_KP_Subtract -> std_term_insert text_wid '-'
      | K_Return -> ok_action ()
      | K_Escape -> cancel_action ()
      | _ -> std_term_ev text_wid ev ]
  | _ -> std_term_ev text_wid ev ]
;

value action_plus gd typ iparam isub wid =
  fun
  [ ButtonEvPress _ _ mbut ->
      if mbut.item <= 3 then plus_or_minus 1 gd typ iparam isub wid else ()
  | _ -> () ]
;

value action_minus gd typ iparam isub wid =
  fun
  [ ButtonEvPress _ _ mbut ->
      if mbut.item <= 3 then plus_or_minus (-1) gd typ iparam isub wid else ()
  | _ -> () ]
;

(* Parameters *)

value action_non_preemptive sc iparam wid =
  fun
  [ CheckEvPress -> do {
      let itask = itask_of_iparam sc iparam in
      let v = not (check_val wid) in
      check_set wid v;
      let td = sc.new_task.(itask) in
      sc.new_task.(itask) := {(td) with non_preemptive = v};
    }
  | _ -> () ]
;

value action_strict_periodicity sc iparam wid =
  fun
  [ CheckEvPress -> do {
      let itask = itask_of_iparam sc iparam in
      let v = not (check_val wid) in
      check_set wid v;
      let td = sc.new_task.(itask) in
      sc.new_task.(itask) := {(td) with strict_per = v};
    }
  | _ -> () ]
;

value has_sub_dur sc =
  loop 0 where rec loop itask =
    if itask = Array.length sc.new_task then False
    else if sc.new_task.(itask).sub_duration = [] && not sc.show_sub.(itask)
    then loop (itask + 1)
    else True
;

value no_action _ _ = ();
value filler =
  pack_desc [FillerAtt; BorderAtt 0; BandAtt 0] (Vertical, []) no_action
;

value task_line_dur_wdescl sc ok_action cancel_action itask isub =
  let iparam = sc.iparam_of_itask.(itask) in
  let backg = BackgroundAtt (ColorPn sc.graph.white_col.da_color) in
  let foreg = ForegroundAtt (ColorPn sc.graph.red_col.da_color) in
  let white_filler =
    pack_desc [FillerAtt; BorderAtt 0; BandAtt 0; backg] (Vertical, [])
      no_action
  in
  let list =
    let suff =
      if isub = 0 then sprintf "%d" iparam else sprintf "%d_%d" iparam isub
    in
    [TD []
       (pack_desc [backg; InterAtt 10]
          (Horizontal,
           [white_filler;
            check_desc
              [NameAtt (sprintf "nonpre_%s" suff); backg; foreg]
              () (action_non_preemptive sc iparam);
            white_filler])
          no_action);
     TD []
       (pack_desc [backg; InterAtt 10]
          (Horizontal,
           [white_filler;
            check_desc
              [NameAtt (sprintf "strictper_%s" suff); backg; foreg]
              () (action_strict_periodicity sc iparam);
            white_filler])
          no_action)]
  in
  [TD [Colspan 8] filler;
   TD []
     (term_desc
        [NameAtt (sprintf "sub_rel_%d_%d" iparam isub);
         FontAtt sc.graph.term_font; BandAtt 5]
       (1, 6, 0) (action_text sc ok_action cancel_action));
   TD [] (button_desc [] ("-", None) (action_minus sc "sub_rel" iparam isub));
   TD [] (button_desc [] ("+", None) (action_plus sc "sub_rel" iparam isub));
   TD []
     (term_desc
        [NameAtt (sprintf "sub_dur_%d_%d" iparam isub);
         FontAtt sc.graph.term_font; BandAtt 5]
       (1, 6, 0) (action_sub_dur_text sc ok_action cancel_action));
   TD [] (button_desc [] ("-", None) (action_minus sc "sub_dur" iparam isub));
   TD [] (button_desc [] ("+", None) (action_plus sc "sub_dur" iparam isub));
   TD [Colspan 6] filler
   :: list]
;

value parameter_header_wdescl sc =
  let filler = TD [] (pack_desc [FillerAtt] (Vertical, []) no_action) in
  let wdescl =
    [TD [Colspan 3] (title_desc [] (I18n.transl "deadline") no_action);
     TD [Colspan 3] (title_desc [] (I18n.transl "period") no_action);
     TD [] (title_desc [] (I18n.transl "(1)") no_action);
     TD [] (title_desc [] (I18n.transl "(2)") no_action);
     TD [Colspan 3] (title_desc [] (I18n.transl "preemption") no_action)]
  in
  let wdescl =
    if has_sub_dur sc then
      [TD [Colspan 3] (title_desc [] (I18n.transl "sub-release")
         no_action);
       TD [Colspan 3] (title_desc [] (I18n.transl "sub-duration")
         no_action) :: wdescl]
    else wdescl
  in
  let wdescl =
    [filler;
     TD [] (title_desc [] "" no_action);
     TD [Colspan 3] (title_desc [] (I18n.transl "release") no_action);
     TD [Colspan 3] (title_desc [] (I18n.transl "duration") no_action)
     :: wdescl]
  in
  wdescl
;

value parameter_trailer_wdescll () =
  [[TD [Colspan 8]
      (pack_desc [BorderAtt 0]
         (Vertical,
          [pack_desc [BorderAtt 0; BandAtt 0]
             (Horizontal,
              [title_desc []
                 (sprintf "(1)%s %s" (I18n.transl ":")
                    (I18n.transl "non preemptive"))
                 no_action;
               filler])
             no_action;
           pack_desc [BorderAtt 0; BandAtt 0]
             (Horizontal,
              [title_desc []
                 (sprintf "(2)%s %s" (I18n.transl ":")
                    (String.uncapitalize (I18n.transl "Strict Periodicity")))
                 no_action;
               filler])
             no_action])
         no_action)]]
;

value rec task_line_wdescl sc ok_action cancel_action itask =
  let iparam = sc.iparam_of_itask.(itask) in
  let backg = BackgroundAtt (ColorPn sc.graph.white_col.da_color) in
  let foreg = ForegroundAtt (ColorPn sc.graph.red_col.da_color) in
  let white_filler =
    pack_desc [FillerAtt; BorderAtt 0; BandAtt 0; backg] (Vertical, [])
      no_action
  in
  let wdescl =
    [TD []
       (term_desc
          [NameAtt (sprintf "preecost_%d" iparam); FontAtt sc.graph.term_font;
           BandAtt 5]
          (1, 6, 0) (action_text sc ok_action cancel_action));
     TD [] (button_desc [] ("-", None) (action_minus sc "preecost" iparam 0));
     TD [] (button_desc [] ("+", None) (action_plus sc "preecost" iparam 0));
     TD []
       (pack_desc [BorderAtt 0; BandAtt 0; InterAtt 0]
          (Vertical,
           [filler;
            pack_desc [BandAtt 0; InterAtt 10]
              (Horizontal,
               [button_desc [] (I18n.transl "remove", None)
                  (Dutil.button_press_action
                     (action_remove_task sc iparam))])
              no_action;
            filler])
          no_action)]
  in
  let wdescl =
    if sc.show_sub.(itask) then
      [TD [Colspan 2] filler :: wdescl]
    else
      [TD []
         (pack_desc [backg]
            (Horizontal,
             [white_filler;
              check_desc
                [NameAtt (sprintf "nonpre_%d" iparam); backg; foreg]
                () (action_non_preemptive sc iparam);
              white_filler])
            no_action);
       TD []
         (pack_desc [backg]
            (Horizontal,
             [white_filler;
              check_desc
                [NameAtt (sprintf "strictper_%d" iparam); backg; foreg]
                () (action_strict_periodicity sc iparam);
              white_filler])
            no_action) ::
       wdescl]
  in
  let wdescl =
    [TD []
       (term_desc
          [NameAtt (sprintf "dea_%d" iparam); FontAtt sc.graph.term_font;
           BandAtt 5]
          (1, 6, 0) (action_text sc ok_action cancel_action));
     TD [] (button_desc [] ("-", None) (action_minus sc "dea" iparam 0));
     TD [] (button_desc [] ("+", None) (action_plus sc "dea" iparam 0));
     TD []
       (term_desc
          [NameAtt (sprintf "per_%d" iparam); FontAtt sc.graph.term_font;
           BandAtt 5]
          (1, 6, 0) (action_text sc ok_action cancel_action));
     TD [] (button_desc [] ("-", None) (action_minus sc "per" iparam 0));
     TD [] (button_desc [] ("+", None) (action_plus sc "per" iparam 0)) ::
     wdescl]
  in
  let wdescl =
    if has_sub_dur sc then
      [TD [Colspan 6]
         (pack_desc [BorderAtt 0] (Horizontal, []) no_action) ::
       wdescl]
    else wdescl
  in
  let icolor = itask mod Array.length sc.graph.color in
  let wdescl =
    [TD []
       (title_desc
          [NameAtt (sprintf "tit_%d" iparam);
           ForegroundAtt (ColorPn sc.graph.color.(icolor).da_color)]
          (if itask < 9 then sprintf " t%d " (itask + 1)
           else sprintf "t%d" (itask + 1))
          no_action);
     TD []
       (button_desc [] (if sc.show_sub.(itask) then "o" else ">", None)
          (Dutil.button_press_action (action_show_sub sc itask)));
     TD []
       (term_desc
          [NameAtt (sprintf "rel_%d" iparam); FontAtt sc.graph.term_font;
           BandAtt 5]
          (1, 6, 0) (action_text sc ok_action cancel_action));
     TD []
       (button_desc [NameAtt (sprintf "rel_m_%d" iparam)]
          ("-", None) (action_minus sc "rel" iparam 0));
     TD []
       (button_desc [NameAtt (sprintf "rel_p_%d" iparam)]
          ("+", None) (action_plus sc "rel" iparam 0));
     TD []
       (term_desc
          [NameAtt (sprintf "dur_%d" iparam); FontAtt sc.graph.term_font;
           BandAtt 5]
          (1, 6, 0) (action_text sc ok_action cancel_action));
     TD [] (button_desc [] ("-", None) (action_minus sc "dur" iparam 0));
     TD [] (button_desc [] ("+", None) (action_plus sc "dur" iparam 0)) ::
     wdescl]
  in
  wdescl

and action_remove_task sc iparam () = do {
  let xd = sc.graph.dd.da_co.xd in
  let itask = itask_of_iparam sc iparam in
  let wid = widget_named xd "task_lines" in
  remove_task_lines sc wid;
  let new_task =
    Array.init (Array.length sc.new_task - 1)
      (fun j -> if j < itask then sc.new_task.(j) else sc.new_task.(j+1))
  in
  sc.new_task := new_task;
  let show_sub =
    Array.init (Array.length sc.show_sub - 1)
      (fun j -> if j < itask then sc.show_sub.(j) else sc.show_sub.(j+1))
  in
  sc.show_sub := show_sub;
  rebuild_task_lines xd sc wid;
}

and action_show_sub sc itask () = do {
  let xd = sc.graph.dd.da_co.xd in
  let wid = widget_named xd "task_lines" in
  remove_task_lines sc wid;
  if sc.show_sub.(itask) then
    sc.new_task.(itask) := {(sc.new_task.(itask)) with sub_duration = []}
  else ();
  sc.show_sub.(itask) := not sc.show_sub.(itask);
  rebuild_task_lines xd sc wid;
}

and task_line_add_sub_button sc itask =
  [TD [Colspan 8] filler;
   TD [Colspan 6]
     (button_desc [] (I18n.transl "add sub-task", None)
        (Dutil.button_press_action (action_add_sub_task sc itask)))]

and action_add_sub_task sc itask () = do {
  let xd = sc.graph.dd.da_co.xd in
  let wid = widget_named xd "task_lines" in
  remove_task_lines sc wid;
  let new_sub_task =
    {sub_rel = 0; sub_dur = 0; sub_non_pree = False; sub_str_per = False}
  in
  let nt = sc.new_task.(itask) in
  let sub_dur = nt.sub_duration @ [new_sub_task] in
  sc.new_task.(itask) := {(nt) with sub_duration = sub_dur};
  rebuild_task_lines xd sc wid;
}

and remove_task_lines sc wid = do {
  let wdescll = task_lines sc (fun _ -> ()) (fun _ -> ()) in
  List.iter (fun _ -> table_remove_nth_row wid 0) wdescll;
  table_remove_nth_row wid 0;
}

and rebuild_task_lines xd sc wid = do {
  let wdescll = task_lines sc (fun _ -> ()) (fun _ -> ()) in
  loop wdescll where rec loop =
    fun
    [ [wdescl :: wdescll] -> do {
        loop wdescll;
        table_insert_row wid 0 wdescl
      }
    | [] -> () ];
(**)
  let _ : (int * int) = rt_adjust_widget (widget_named xd "param") in
(*
  let wid2 = widget_named xd "task_lines" in
  let w = widget_width wid2 in
  let (_, h) = rt_adjust_widget wid2 in
  eprintf "w %d h %d\n" w h; flush stderr;
  rt_resize_widget wid2 w h;
*)
  set_parameter_widget xd sc;
}

and task_lines sc ok_action esc_action =
  if Array.length sc.new_task = 0 then []
  else
    loop [] 0 where rec loop rev_wdescll itask =
      if itask = Array.length sc.new_task then
        let rev_wdescll =
          List.rev_append (parameter_trailer_wdescll ()) rev_wdescll
        in
        [parameter_header_wdescl sc :: List.rev rev_wdescll]
      else
        let wdescl = task_line_wdescl sc ok_action esc_action itask in
        let rev_wdescll = [wdescl :: rev_wdescll] in
        let rev_wdescll =
          let td = sc.new_task.(itask) in
          if td.sub_duration = [] && not sc.show_sub.(itask) then
            rev_wdescll
          else
            let rev_wdescll =
              let wdescl =
                task_line_dur_wdescl sc ok_action esc_action itask 0
              in
              [wdescl :: rev_wdescll]
            in
            loop rev_wdescll 1 td.sub_duration
            where rec loop rev_wdescll isub =
              fun
              [ [] -> [task_line_add_sub_button sc itask :: rev_wdescll]
              | [_ :: sub_dur] ->
                  let wdescl =
                    task_line_dur_wdescl sc ok_action esc_action itask
                      isub
                  in
                  loop [wdescl :: rev_wdescll] (isub + 1) sub_dur ]
        in
        loop rev_wdescll (itask + 1)
;

value max_tasks = 10;

value action_add_task sc ok_action cancel_action () =
  if Array.length sc.new_task = max_tasks then ()
  else do {
    let xd = sc.graph.dd.da_co.xd in
    let wid = widget_named xd "task_lines" in
    let itask = Array.length sc.new_task in
    remove_task_lines sc wid;
    let td =
      {release = 0; duration = 1; sub_duration = []; deadline = 10;
       period = 10; user_release = 0; non_preemptive = False;
       strict_per = False; pree_cost = 0}
    in
    let iparam = if itask = 0 then 0 else sc.iparam_of_itask.(itask-1) + 1 in
    sc.iparam_of_itask := Array.append sc.iparam_of_itask [| iparam |];
    sc.new_task := Array.append sc.new_task [| td |];
    sc.show_sub := Array.append sc.show_sub [| False |];
    rebuild_task_lines xd sc wid;
  }
;

value task_lines_wdesc sc ok_action esc_action name =
  table_desc [NameAtt name] (task_lines sc ok_action esc_action)
    no_action
;

value parameter_wdesc empty_sc sc ok_action reset_action cancel_action
    esc_action task_lines_name =
  let filler = pack_desc [FillerAtt] (Vertical, []) no_action in
  pack_desc [NameAtt "param"]
    (Vertical,
     [pack_desc []
        (Horizontal,
         [filler; term_desc [NameAtt "file_name"] (1, 50, 0) no_action;
          filler])
        no_action;
      task_lines_wdesc sc ok_action esc_action task_lines_name;
      pack_desc []
        (Horizontal,
         [button_desc [] (I18n.transl "add task", None)
            (Dutil.button_press_action
               (action_add_task sc ok_action esc_action))])
        no_action;
      pack_desc []
        (Horizontal,
         [pack_desc []
            (Horizontal,
             [title_desc []
                (sprintf "%s (%s)" (I18n.transl "preemption cost")
                   (I18n.transl "default"))
                no_action;
              term_desc
                [NameAtt "pre_cost"; FontAtt sc.graph.term_font; BandAtt 5]
                (1, 5, 0) (action_text sc ok_action esc_action);
              button_desc [] ("-", None) (action_minus sc "pre_cost" 0 0);
              button_desc [] ("+", None) (action_plus sc "pre_cost" 0 0)])
            no_action;
          filler;
          pack_desc []
            (Horizontal,
             [title_desc [] (I18n.transl "util.fact." ^ " ") no_action;
              term_desc
                [NameAtt "util_fact"; FontAtt sc.graph.term_font; BandAtt 5]
                (1, 12, 0) no_action])
            no_action;
          pack_desc []
            (Horizontal,
             [title_desc [] (I18n.transl "LCM" ^ " ") no_action;
              term_desc [NameAtt "lcm"; FontAtt sc.graph.term_font; BandAtt 5]
                (1, 12, 0) no_action])
            no_action])
        no_action;
      pack_desc []
        (Horizontal,
         [title_desc []
            (sprintf "%s (%s%s t1<t7<t2 t4<t3)"
               (I18n.transl "precedence") (I18n.transl "example")
               (I18n.transl ":"))
            no_action;
          filler])
        no_action;
      pack_desc []
        (Horizontal,
         [term_desc [NameAtt "graph_prec_text"] (1, 80, 0)
            (action_prec_text ok_action esc_action);
          pack_desc [FillerAtt] (Vertical, []) no_action])
        no_action;
      pack_desc []
        (Horizontal,
         [title_desc []
            (sprintf "%s (%s%s t3(4)->t1(7)<22 t2(8)->t1(5)<12)"
               (I18n.transl "latency") (I18n.transl "example")
               (I18n.transl ":"))
            no_action;
          filler])
        no_action;
      pack_desc []
        (Horizontal,
         [term_desc [NameAtt "latency_text"] (2, 80, 0)
            (action_latency_text ok_action esc_action);
          pack_desc [FillerAtt] (Vertical, []) no_action])
        no_action;

      pack_desc []
        (Horizontal,
         [title_desc []
            (sprintf "%s (%s%s t1,t3,t2,t4,t5)"
               (I18n.transl "priority") (I18n.transl "example")
               (I18n.transl ":"))
            no_action;
          filler])
        no_action;
      pack_desc []
        (Horizontal,
         [term_desc [NameAtt "priority_text"] (2, 80, 0)
            (action_priority_text ok_action esc_action);
          pack_desc [FillerAtt] (Vertical, []) no_action])
        no_action;

      pack_desc []
        (Horizontal,
         [pack_desc []
            (InDepth,
             [pack_desc [NameAtt "no_param_mess"] (Vertical, [])
                no_action;
              term_desc [NameAtt "param_mess"; BandAtt 0]
                (1, 60, 0) no_action])
            no_action])
        no_action;
      pack_desc []
        (Horizontal,
         [filler;
          button_desc [] (I18n.transl "Accept", None)
            (Dutil.button_press_action ok_action);
          filler;
          button_desc [] (I18n.transl "Reset", None)
            (Dutil.button_release_action reset_action);
          filler;
          button_desc [] (I18n.transl "Cancel", None)
            (Dutil.button_press_action cancel_action);
          filler])
        no_action])
    (Dutil.action_pack_esc_ret (fun _ -> esc_action ())
       (fun _ -> ok_action ()))
;

value display_change_parameters xa xd gd ti def_keyp ul_xy (x, y) = do {
  let no_sc = Array.length gd.scenario = 0 in
  let sc = Scenario_rt.current_scenario gd in
  let cancel_action () = do {
    unmap_change_parameters xd gd;
    if no_sc then gd.scenario := [| |] else ();
  }
  in
  let reset_action _ = do {
    let ts2 = sc.ts in
    let ts = ts2.ast in
    let tc = ts.constraints in
    let wid = widget_named xd "task_lines" in
    remove_task_lines sc wid;
    sc.new_task := Array.copy ts.tasks_def;
    sc.new_preemption_cost := tc.preemption_cost;
    sc.new_graph_precedence := tc.graph_precedence;
    sc.new_latency_constr := ts2.latency_constr;
    sc.show_sub :=
      Array.init (Array.length ts.tasks_def)
        (fun itask -> ts.tasks_def.(itask).sub_duration <> []);
    sc.iparam_of_itask := Array.init (Array.length ts.tasks_def) (fun i -> i);
    rebuild_task_lines xd sc wid;
  }
  in
  let ok_action () = do {
    if no_sc then do {
      let wdesc = Scenario_rt.scenario_desc xa gd ti def_keyp 0 in
      Tabwid.tab_add xd ti sc.fname wdesc;
      Scenario_rt.make_moving_schedule_widget gd 0;
    }
    else ();
    unmap_change_parameters xd gd;
    rt_sync xd;
    if no_sc || Array.length sc.new_task <> Array.length sc.ts.ast.tasks_def
    then do {
      let many_tasks = Dim.many_tasks sc.new_task in
      sc.graph.dd.inter_task_dy := if many_tasks then 35 else 40;
      raw_set_size (Dutil.g_widget_named gd "schedule")
        (1, Dim.schedule_height (Array.length sc.new_task));
      let wid = widget_named xd "main" in
      let (_, _) = rt_adjust_widget wid in
      ()
    }
    else ();
    Parameters.ok_change_parameters sc;
    let dim = sc.graph.dd.rt_dim in
    sc.graph.dd.rt_dim :=
      Dim.dim_of_width_height sc.ts dim.dam_width dim.dam_height;
    Schedule_step_rt.goto_begin gd;
    if no_sc then rt_raise_widget (widget_named xd "main_tab") else ();
  }
  in
  let wid = widget_named xd "main" in
  let _ =
    rt_create_transient_widget wid (I18n.transl "Parameters")
      (Some (fun _ -> unmap_change_parameters xd gd))
      (parameter_wdesc no_sc sc ok_action reset_action cancel_action
         cancel_action "task_lines")
  in
  change_parameters xd sc ul_xy (x, y)
};
