(* $Id: randam.ml,v 1.16 2009-10-05 14:44:02 deraugla Exp $ *)

open Printf;

value special = ref False;
value load = ref 1.0;
value max_tasks = ref 10;
value patrick = ref False;

value rec random_divisor n =
  let d = 4 + Random.int (n - 3) in
  if n mod d = 0 then d
  else random_divisor n
;

value util_fact_gt_1 n tasks =
  let uf =
    List.fold_left (fun accu (_, dur, _, per, _, _) -> accu + dur * n / per) 0
    tasks
  in
  uf > n
;

value rec choose_tasks () =
  let rel0 = Random.int 3 = 0 in
  let n1 = 2 + Random.int 3 in
  let n2 = n1 + Random.int 3 in
  let n3 = n2 + Random.int 3 in
  let n4 = n3 + Random.int 3 in
(**)
  let n5 = if patrick.val then n4 else n4 + Random.int 3 in
  let n6 = if patrick.val then n5 else n5 + Random.int 3 in
  let n = n1 * n2 * n3 * n4 * n5 * n6 in
(*
  let ntasks = 3 + Random.int 4 in
*)
  let ntasks = max_tasks.val in
(**)
  let d =
    if ntasks = 6 then 3
    else if ntasks = 5 then 2
    else 0
  in
  let tasks =
    loop ntasks [] where rec loop cnt tasks =
      if cnt = 0 then tasks
      else
        let per = random_divisor n in
(*
        let dur = 1 + Random.int (max 1 (per / 2 - d)) in
*)
        let dur = 1 + Random.int (max 1 (per / 4 - d)) in
(**)
        let rel =
          if special.val then 0
          else if rel0 then 0
          else Random.int per - per / 2
        in
        let dea =
          if special.val then per
          else if Random.int 3 = 0 then per - Random.int 2
          else per
        in
        let np =
          if special.val || patrick.val then False
          else dur > 1 && Random.int 10 = 0
        in
        loop (cnt - 1) [(rel, dur, dea, per, np, 0) :: tasks]
  in
  if util_fact_gt_1 n tasks then choose_tasks () else tasks
;

value max_lcm () = 10000000;

value randur max_poss per =
  let max_dur = max 2 (min max_poss (2 * per / 3)) in
  1 + Random.int (max_dur - 1)
;

value put_load lcm =
  int_of_float (float lcm *. load.val)
;

value rec spe_choose_tasks () = do {
  let per1 = 4 + Random.int 4 in
(*
  let dur1 = 1 + Random.int (per1 / 2 - 1) in
*)
  let dur1 = randur per1 per1 in
(**)
  loop 0 0 1 per1 dur1 dur1 per1 [] where rec
  loop ntasks old_load_num old_lcm per dur load_num lcm tasks =
    if dur <= 0 || lcm > max_lcm () || ntasks >= max_tasks.val then
(*
let _ = do { eprintf "load %d/%d = %.4f\n" old_load_num old_lcm (float old_load_num /. float old_lcm); flush stderr } in
*)
      let real_load = float old_load_num /. float old_lcm in
      if abs_float (load.val -. real_load) /. load.val < 0.003 then tasks
      else do {
(*
        eprintf "too big difference; restarting\n";
        flush stdout;
*)
        spe_choose_tasks ()
      }
    else
      let rel = 0 in
      let dea = per in
      let np = False in
      let t = (rel, dur, dea, per, np, 0) in
      let tasks = [t :: tasks] in
      let n = 2 + Random.int 3 in
      let new_per = n * lcm in
      let max_dur = n * (put_load lcm - load_num) in
      let new_dur = randur max_dur new_per in

      let old_load_num = load_num in
      let old_lcm = lcm in

      let load_num = n * load_num + new_dur in
      let lcm = n * lcm in

      loop (ntasks + 1) old_load_num old_lcm new_per new_dur
        load_num lcm tasks
};

value print_tasks tasks = do {
  let pre_cost =
    if special.val || patrick.val then 0
    else if Random.int 2 = 0 then 1 + Random.int 2
    else 0
  in
  printf "%d\n" pre_cost;
  List.iter
    (fun (rel, dur, dea, per, np, pc) -> do {
       printf "%d %d %d" rel dur dea;
       if dea = per then () else printf " %d" per;
       if np then printf " np" else ();
       if pc > 0 then printf " pc %d" pc else ();
       printf "\n";
     })
    tasks;
  if special.val || patrick.val then ()
  else do {
    if Random.int 5 = 0 then printf "prec\n" else ();
    let ntasks = List.length tasks in
    printf "lat:";
    printf " t%d(%d)->t%d(%d)<%d" (1 + Random.int ntasks)
      (1 + Random.int 22) (1 + Random.int ntasks) (1 + Random.int 22)
      (1 + Random.int 250);
  for i = 1 to (*2*)5 do {
    printf " t%d(%d)->t%d(%d)<%d" (1 + Random.int ntasks)
      (1 + Random.int 22) (1 + Random.int ntasks) (1 + Random.int 22)
      (1 + Random.int 250);
  };
    if Random.int 3 = 0 then
      printf " t%d(%d)->t%d(%d)<%d" (1 + Random.int ntasks)
        (1 + Random.int 22) (1 + Random.int ntasks) (1 + Random.int 22)
        (1 + Random.int 250)
    else ();
    printf "\n";
  };
};

value compare_deadlines (_, _, d1, _, _, _) (_, _, d2, _, _, _) =
  compare d1 d2
;

value special_tasks tasks =
  let (rev_tasks, _) =
    List.fold_left
      (fun (rev_tasks, shift) task ->
         let (rel, dur, dea, per, np, pc) = task in
         let rel = - shift - dur in
         let shift = - rel in
         ([(rel, dur, dea, per, np, pc) :: rev_tasks], shift))
      ([], 0) tasks
  in
  List.rev rev_tasks
;

value random_start_tasks tasks =
  let lcm =
    let (rel, dur, dea, per, np, pc) = List.hd (List.rev tasks) in
    per
  in
  List.map
    (fun task ->
       let (rel, dur, dea, per, np, pc) = task in
       let rel = - Random.int lcm in
       (rel, dur, dea, per, np, pc))
    tasks
;

value patrick_choose_tasks () =
  let tasks = choose_tasks () in
  let k = 10 in
  List.map
    (fun (rel, dur, dea, per, np, pc) ->
       (rel, dur * k, dea * k, per * k, np, pc))
    tasks
;

value round x = truncate (x +. 0.5);

value randam () = do {
  Random.self_init ();
  let tasks =
    if patrick.val then patrick_choose_tasks ()
    else if special.val then spe_choose_tasks ()
    else choose_tasks ()
  in
  let tasks =
    if special.val then List.sort compare_deadlines tasks else tasks
  in
  print_tasks tasks;
  flush stdout;
  if patrick.val then do {
    printf "EOD\n";
    let task2 =
      List.map
        (fun (rel, dur, dea, per, np, pc) ->
           (rel, max 1 (round (float dur *. 100. /. 130.)), dea, per, np,
            max 1 (round (float dur *. 3. /. 130.))))
        tasks
    in
    print_tasks task2;
    flush stdout;
  }
  else if special.val then do {
    printf "EOD\n";
    let tasks2 = special_tasks tasks in
    print_tasks tasks2;
    for i = 1 to 10 do {
      printf "EOD\n";
      let tasks3 = random_start_tasks tasks in
      print_tasks tasks3;
    };
    flush stdout;
  }
  else ();
};

value spec_list =
  Arg.align
    [("-s", Arg.Set special, " Special");
     ("-l", Arg.Set_float load, " Set load");
     ("-n", Arg.Set_int max_tasks, " Number of tasks");
     ("-p", Arg.Set patrick, " Patrick")]
;
value anon_fun s = ();

value main () = do {
  Arg.parse spec_list anon_fun "Usage: randam";
  randam ();
};

main ();
