(* $Id: sarray.ml,v 1.5 2011-11-03 13:26:30 deraugla Exp $ *)

type t 'a = array (array 'a);

value mal = 10;
value mal = Sys.max_array_length;

value empty = [| |];

value make len v = do {
  let a = Array.make ((len + mal - 1) / mal) [| |] in
  for i = 0 to Array.length a - 2 do {
    a.(i) := Array.make mal v;
  };
  if len > 0 then
    a.(Array.length a - 1) := Array.make ((len - 1) mod mal + 1) v
  else ();
  a
};

value length a = do {
  if Array.length a = 0 then 0
  else mal * (Array.length a - 1) + Array.length (a.(Array.length a - 1))
};

value get a i = a.(i / mal).(i mod mal);
value set a i v = a.(i / mal).(i mod mal) := v;

value init len f = do {
  if len = 0 then [| |]
  else do {
    let a = make len (f 0) in
    for i = 1 to len - 1 do { set a i (f i) };
    a
  }
};

value copy a = do {
  Array.init (Array.length a) (fun i -> Array.copy a.(i))
};

value append a b = do {
  let alen = length a in
  let blen = length b in
  if alen = 0 then copy b
  else if blen = 0 then copy a
  else do {
    let len = alen + blen in
    init len (fun i -> if i < alen then get a i else get b (i-alen))
  }
};

value sub a ini len = do {
  if len = 0 then empty else init len (fun i -> get a (ini + i))
};

value to_list a = do {
  let len = length a in
  loop (len - 1) [] where rec loop i list = do {
    if i < 0 then list
    else loop (i - 1) [get a i :: list]
  }
};

value of_list l = do {
  match l with
  [ [] -> empty
  | [x :: l] -> do {
      let len = 1 + List.length l in
      let a = make len x in
      loop 1 l where rec loop i =
        fun
        [ [x :: l] -> do { set a i x; loop (i + 1) l }
        | [] -> a ]
    } ]
};

value reinit a v = do {
  for i = 0 to Array.length a - 1 do {
    let b = a.(i) in
    b.(0) := v;
    Array.blit b 0 b 1 (Array.length b - 1);
  };
};
