(* $Id: save.ml,v 1.41 2010-12-04 07:06:12 deraugla Exp $ *)

open Printf;
open Rt;

open Da;
open Data_rt;
open Dtasks;
open Glob_glob;
open Std_term;
open Text_rt;

value kill_widget gd wid = do {
  let xd = xdata_of_widget wid in
  rt_destroy_widget wid;
  Dutil.redirect_to_main gd;
  rt_unfreeze_widget (widget_named xd "main");
};

value button_action act _ =
  fun
  [ ButtonEvRelease x y _ -> act (x, y)
  | _ -> () ]
;

value action_pack_esc_ret esc_f ret_f wid =
  fun
  [ PackEvKeyPress {item = k} ->
      match k with
      [ K_Escape -> esc_f wid
      | K_Return -> ret_f wid
      | _ -> () ] ]
;

value unmap_save_as xd gd clean_up_save = do {
  rt_unmap_widget (widget_named xd "save_as");
  Dutil.redirect_to_main gd;
  rt_unfreeze_widget (widget_named xd "main");
  clean_up_save ()
};

value error clean_up_save xd wid gd ul_xy (x, y) txt = do {
  let backg = BackgroundAtt (ColorPn (fst gd.common.c_white_col)) in
  let wid =
    let my_kill_widget gd wid = do {
      kill_widget gd wid;
      clean_up_save ();
    }
    in
    rt_create_transient_widget wid
      (I18n.transl "Error") (Some (my_kill_widget gd))
      (pack_desc [NameAtt "error"]
         (Vertical,
          [title_desc [] txt (fun  _ _ -> ());
           pack_desc [BandAtt 25] (Horizontal, []) (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [pack_desc [FillerAtt] (Vertical, []) (fun _ _ -> ());
               button_desc [backg] (I18n.transl "OK", None)
                 (button_action
                    (fun _ -> my_kill_widget gd (widget_named xd "error")))])
             (fun _ _ -> ())])
         (action_pack_esc_ret
            (fun wid -> my_kill_widget gd (widget_named xd "error"))
            (fun wid -> my_kill_widget gd (widget_named xd "error"))))
  in
  let (x, y) =
    match ul_xy with
    [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
    | None -> (x, y) ]
  in
  rt_redirect_key_press_to (widget_named xd "error");
  rt_map_transient_widget wid x y
};

value say_saved wid clean_up_save gd fname ul_xy (x, y) = do {
  let xd = gd.common.xd in
  let wid =
    let backg = BackgroundAtt (ColorPn (fst gd.common.c_white_col)) in
    let my_kill_widget gd wid = do {
      kill_widget gd wid;
      clean_up_save ();
    }
    in
    rt_create_transient_widget wid
      (I18n.transl "Saved") (Some (my_kill_widget gd))
      (pack_desc [NameAtt "saved"]
         (Vertical,
          [title_desc [] (sprintf (I18n.ftransl "File %s saved") fname)
             (fun _ _ -> ());
           pack_desc [BandAtt 25] (Horizontal, []) (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [pack_desc [FillerAtt] (Vertical, []) (fun _ _ -> ());
               button_desc [backg] (I18n.transl "OK", None)
                 (button_action
                    (fun _ -> my_kill_widget gd (widget_named xd "saved")))])
             (fun _ _ -> ())])
         (action_pack_esc_ret
            (fun wid -> my_kill_widget gd (widget_named xd "saved"))
            (fun wid -> my_kill_widget gd (widget_named xd "saved"))))
  in
  let (x, y) =
    match ul_xy with
    [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
    | None -> (x, y) ]
  in
  rt_freeze_widget (widget_named xd "main");
  rt_redirect_key_press_to (widget_named xd "saved");
  rt_map_transient_widget wid x y
};

value do_save_file gd fname = do {
  let sc = gd.scenario.(gd.curr_sc) in
  let ts = sc.ts in
  Schedule.output_task_data fname ts;
};

value do_save_result gd fname = do {
  let sc = gd.scenario.(gd.curr_sc) in
  let ts2 = sc.ts in
  let oc = open_out fname in
  for i = 0 to ts2.schedule_len - 1 do {
    match Sarray.get ts2.disp_schedule i with
    [ Some u ->
        fprintf oc " t%d%s" (u.Dtasks.u_task + 1)
          (if u.Dtasks.u_preempt then "p" else "")
    | None ->
        fprintf oc " f" ];
    if i mod 10 = 9 then fprintf oc "\n" else ();
  };
  fprintf oc "\n";
  close_out oc;
};

value do_print_ps build_ps gd fname = do {
  let oc = open_out fname in
  build_ps oc;
  close_out oc;
};

value test_exist_or_action action_ok clean_up_save xd wid gd fname ul_xy
    (x, y) =
  let action_ok xd gd fname ul_xy (x, y) =
    try do {
      action_ok gd fname;
      say_saved wid clean_up_save gd fname ul_xy (x, y);
    }
    with e ->
      error clean_up_save xd wid gd ul_xy (x, y) (Printexc.to_string e)
  in
  if fname = "" then ()
  else if Sys.file_exists fname then do {
    let backg = BackgroundAtt (ColorPn (fst gd.common.c_white_col)) in
    let my_kill_widget gd wid = do {
      kill_widget gd wid;
      clean_up_save ();
    }
    in
    let wid =
      rt_create_transient_widget wid
        (I18n.transl "Save?") (Some (my_kill_widget gd))
        (pack_desc [NameAtt "save_yn"]
           (Vertical,
            [title_desc []
               (sprintf (I18n.ftransl "File %s already exists; overwrite?")
                  fname)
               (fun _ _ -> ());
             pack_desc [BandAtt 25] (Horizontal, []) (fun _ _ -> ());
             pack_desc []
               (Horizontal,
                [pack_desc [FillerAtt] (Vertical, []) (fun _ _ -> ());
                 button_desc [backg] (I18n.transl "no", None)
                   (button_action
                      (fun _ ->
                         my_kill_widget gd (widget_named xd "save_yn")));
                 button_desc [backg] (I18n.transl "yes", None)
                   (button_action
                      (fun _ -> do {
                         my_kill_widget gd (widget_named xd "save_yn");
                         action_ok xd gd fname ul_xy (x, y)
                       }))])
               (fun _ _ -> ())])
           (action_pack_esc_ret
              (fun wid -> my_kill_widget gd (widget_named xd "save_yn"))
              (fun wid -> do {
                 my_kill_widget gd (widget_named xd "save_yn");
                 action_ok xd gd fname ul_xy (x, y)
               })))
    in
    let (x, y) =
      match ul_xy with
      [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
      | None -> (x, y) ]
    in
    rt_freeze_widget (widget_named xd "main");
    rt_redirect_key_press_to (widget_named xd "save_yn");
    rt_map_transient_widget wid x y
  }
  else action_ok xd gd fname ul_xy (x, y)
;

value ok_save_as xd gd ul_xy (x, y) = do {
  unmap_save_as xd gd (fun () -> ());
  let wid = widget_named xd "main" in
  let fname = get_term_contents (widget_named xd "file_name") in
  test_exist_or_action do_save_file (fun () -> ()) xd wid gd fname ul_xy
    (x, y)
};

value ok_save_result xd gd ul_xy (x, y) = do {
  unmap_save_as xd gd (fun () -> ());
  let wid = widget_named xd "main" in
  let fname = get_term_contents (widget_named xd "file_name") in
  test_exist_or_action do_save_result (fun () -> ()) xd wid gd fname ul_xy
    (x, y)
};

value ok_print_ps build_ps clean_up_save wid xd gd ul_xy (x, y) = do {
  unmap_save_as xd gd clean_up_save;
  let fname = get_term_contents (widget_named xd "file_name") in
  test_exist_or_action (do_print_ps build_ps) clean_up_save xd wid gd fname
    ul_xy (x, y)
};

value action_text_output_file clean_up_save ok_act xd gd ul_xy (x, y) wid ev =
  let wid = widget_named xd "file_name" in
  match ev with
  [ TermEvButtonPress _ _ _ _ -> do {
      term_emphasize_from wid 0 0;
      term_emphasize_to wid 0 0;
      std_term_ev wid ev
    }
  | TermEvKeyPress {item = K_Escape} -> unmap_save_as xd gd clean_up_save
  | TermEvKeyPress {item = K_Return} -> ok_act xd gd ul_xy (x, y)
  | TermEvKeyPress {item = k; control = False} -> do {
      match k with
      [ K_Ascii _ | K_KP_N _ | K_KP_Subtract | K_KP_Divide ->
          let emph_txt = term_get_emphasized wid in
          if emph_txt <> "" then
            term_send wid (vt_home ^ vt_erase_from_cursor)
          else ()
      | _ -> () ];
      match k with
      [ K_Ascii
            ('a'..'z' | 'A'..'Z' | '0'..'9' | '-' | '_' | '/' | '.' as c) ->
          term_send wid (String.make 1 c)
      | K_KP_N i ->
          term_send wid (String.make 1 (Char.chr (Char.code '0' + i)))
      | K_KP_Subtract ->
          term_send wid "-"
      | K_KP_Divide ->
          term_send wid "/"
      | _ ->
          std_term_ev wid ev ]
    }
  | TermEvKeyPress {control = True} -> std_term_ev wid ev
  | _ -> () ]
;

value button_action_ok_save_as xd wid gd ul_xy (x, y) _ =
  fun
  [ ButtonEvPress _ _ _ -> ok_save_as xd gd ul_xy (x, y)
  | _ -> () ]
;

value button_action_ok_save_result xd wid gd ul_xy (x, y) _ =
  fun
  [ ButtonEvPress _ _ _ -> ok_save_result xd gd ul_xy (x, y)
  | _ -> () ]
;

value action_cancel_save_as xd gd clean_up_save wid =
  fun
  [ ButtonEvPress _ _ _ -> unmap_save_as xd gd clean_up_save
  | _ -> () ]
;

value button_action_ok_print_ps build_ps clean_up_save xd wid gd ul_xy (x, y)
     _ =
  fun
  [ ButtonEvPress _ _ _ ->
      ok_print_ps build_ps clean_up_save wid xd gd ul_xy (x, y)
  | _ -> () ]
;

value ask_file_name wid fn fn_tit action_text button_action_ok action_ok
    clean_up_save xtra xd gd ul_xy (x, y) = do {
  let wid =
    let filler = pack_desc [FillerAtt] (Vertical, []) (fun _ _ -> ()) in
    rt_create_transient_widget wid
      (I18n.transl "Save parameters") (Some (kill_widget gd))
      (pack_desc [NameAtt "save_as"]
         (Vertical,
          [pack_desc []
             (Horizontal,
              [filler;
               title_desc [] fn_tit (fun _ _ -> ());
               filler])
             (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [term_desc [NameAtt "file_name"] (1, 40, 0)
                 (action_text xd gd ul_xy (x, y))])
             (fun _ _ -> ());
           pack_desc [] (Horizontal, xtra) (fun _ _ -> ());
           pack_desc [BandAtt 25] (Horizontal, []) (fun _ _ -> ());
           pack_desc []
             (Horizontal,
              [filler;
               button_desc
                 [BackgroundAtt (ColorPn (fst gd.common.c_white_col))]
                 (I18n.transl "Save", None)
                 (button_action_ok xd wid gd ul_xy (x, y));
               filler;
               button_desc
                 [BackgroundAtt (ColorPn (fst gd.common.c_white_col))]
                 (I18n.transl "Cancel", None)
                 (action_cancel_save_as xd gd clean_up_save);
               filler])
             (fun _ _ -> ())])
         (action_pack_esc_ret
            (fun wid -> unmap_save_as xd gd clean_up_save)
            (fun wid ->
               let x = widget_x wid in
               let y = widget_y wid in
               let w = widget_width wid in
               let h = widget_height wid in
               action_ok xd gd (Some (x + w / 2, y + h / 2)) (0, 0))))
  in
  let (x, y) =
    match ul_xy with
    [ Some (x, y) -> (x - widget_width wid / 2, y - widget_height wid / 2)
    | None -> (x, y) ]
  in
  let txt_wid = widget_named xd "file_name" in
  term_send txt_wid fn;
  term_send txt_wid vt_show_cursor;
  term_send txt_wid vt_insert_mode;
  term_emphasize_from txt_wid 0 0;
  term_emphasize_to txt_wid 0 (String.length fn);
  rt_freeze_widget (widget_named xd "main");
  rt_redirect_key_press_to (widget_named xd "file_name");
  rt_map_transient_widget wid x y
};

value save_as xd gd ul_xy (x, y) =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    let fn = if sc.fname = "-" then "" else sc.fname in
    let action_text = action_text_output_file (fun () -> ()) ok_save_as in
    let wid = widget_named xd "main" in
    ask_file_name wid fn (I18n.transl "File name") action_text
      button_action_ok_save_as ok_save_as (fun () -> ()) [] xd gd ul_xy (x, y)
  else ()
;

value save_result xd gd ul_xy (x, y) =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    let fn =
      if sc.fname = "-" then ""
      else if Filename.check_suffix sc.fname ".dm" then
        Filename.chop_suffix sc.fname ".dm" ^ ".dmo"
      else
        sc.fname ^ ".dmo"
    in
    let action_text = action_text_output_file (fun () -> ()) ok_save_result in
    let wid = widget_named xd "main" in
    ask_file_name wid fn (I18n.transl "File name") action_text
      button_action_ok_save_result ok_save_result (fun () -> ()) [] xd gd
        ul_xy (x, y)
  else ()
;

value switch_ps_A3 sc wid =
  fun
  [ CheckEvPress -> do {
      let v = not (check_val wid) in
      check_set wid v;
      sc.graph.dd.ps_A3 := v;
    }
  | CheckEvRelease -> () ]
;

value switch_txt wname f wid =
  fun
  [ TitleEvButtonPress _ ->
      let xd = xdata_of_widget wid in
      let check_wid = widget_named xd wname in
      f check_wid CheckEvPress ]
;

value set_landscape xd sc landscape_wid =
  fun
  [ CheckEvPress -> do {
      check_set landscape_wid True;
      check_set (widget_named xd "portrait") False;
      sc.graph.dd.ps_landscape := True;
    }
  | CheckEvRelease -> () ]
;

value set_portrait xd sc portrait_wid =
  fun
  [ CheckEvPress -> do {
      check_set portrait_wid True;
      check_set (widget_named xd "landscape") False;
      sc.graph.dd.ps_landscape := False;
    }
  | CheckEvRelease -> () ]
;

value print_ps xd wid gd default_fname build_ps clean_up_save ul_xy (x, y) =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let xtra =
      let backg = BackgroundAtt (ColorPn sc.graph.white_col.da_color) in
      let foreg = ForegroundAtt (ColorPn sc.graph.red_col.da_color) in
      [pack_desc [InterAtt 10]
         (Horizontal,
          [pack_desc [InterAtt 10; BandAtt 0; backg]
             (Horizontal,
              [pack_desc [BandAtt 0; backg] (Vertical, []) (fun _ _ -> ());
               check_desc [backg; foreg; NameAtt "A3"] () (switch_ps_A3 sc);
               title_desc [] "A3" (switch_txt "A3" (switch_ps_A3 sc))])
             (fun _ _ -> ());
           pack_desc [InterAtt 10; BandAtt 0; backg]
             (Horizontal,
              [pack_desc [BandAtt 0; backg] (Vertical, []) (fun _ _ -> ());
               pack_desc [BandAtt 0; backg]
                 (Horizontal,
                  [check_desc [NameAtt "landscape"; backg; foreg] ()
                     (set_landscape xd sc);
                   title_desc [] (I18n.transl "landscape")
                     (switch_txt "landscape" (set_landscape xd sc))])
                 (fun _ _ -> ());
               pack_desc [BandAtt 0; backg]
                 (Horizontal,
                  [check_desc [NameAtt "portrait"; backg; foreg] ()
                     (set_portrait xd sc);
                   title_desc [] (I18n.transl "portrait")
                     (switch_txt "portrait" (set_portrait xd sc))])
                 (fun _ _ -> ())])
             (fun _ _ -> ())])
         (fun _ _ -> ())]
    in
    ask_file_name wid default_fname (I18n.transl "Postscript file name")
      (action_text_output_file clean_up_save
         (ok_print_ps build_ps clean_up_save wid))
      (button_action_ok_print_ps build_ps clean_up_save)
      (ok_print_ps build_ps clean_up_save wid) clean_up_save xtra xd gd ul_xy
      (x, y);
    check_set (widget_named xd "landscape") True;
    sc.graph.dd.ps_landscape := True;
  }
  else ()
;
