(* $Id: scenario_rt.ml,v 1.31 2011-02-03 13:08:54 deraugla Exp $ *)

open Printf;
open Rt;

open Ast;
open Da;
open Data_rt;
open Dim_def;
open Dtasks;
open Glob_glob;

value make_image_pixmap dd name w h im = do {
  let pix = rt_create_pixmap dd.da_co.xd w h in
  let image =
    let s =
      String.concat "" (List.map (fun c -> String.make 1 (Char.chr c)) im)
    in
    rt_create_image dd.da_co.xd s w h 1
  in
  rt_put_image (PixmapDr pix) image (0, 0, w, h) (0, 0);
  {da_pixmap = pix; da_p_name = name; da_p_dd = dd}
};

value im1 =
  [0x81; 0x81; 0x42; 0x42; 0x24; 0x24; 0x18; 0x18;
   0x18; 0x18; 0x24; 0x24; 0x42; 0x42; 0x81; 0x81;
   0x81; 0x81; 0x42; 0x42; 0x24; 0x24; 0x18; 0x18;
   0x18; 0x18; 0x24; 0x24; 0x42; 0x42; 0x81; 0x81];
value im2 =
  [0x99; 0x99; 0x66; 0x66; 0x66; 0x66; 0x99; 0x99;
   0x99; 0x99; 0x66; 0x66; 0x66; 0x66; 0x99; 0x99;
   0x99; 0x99; 0x66; 0x66; 0x66; 0x66; 0x99; 0x99;
   0x99; 0x99; 0x66; 0x66; 0x66; 0x66; 0x99; 0x99];
value im3 =
  [0xc1; 0xc1; 0xe0; 0xe0; 0x70; 0x70; 0x38; 0x38;
   0x1c; 0x1c; 0x0e; 0x0e; 0x07; 0x07; 0x83; 0x83;
   0xc1; 0xc1; 0xe0; 0xe0; 0x70; 0x70; 0x38; 0x38;
   0x1c; 0x1c; 0x0e; 0x0e; 0x07; 0x07; 0x83; 0x83];
value im4 =
  [0x21; 0x00; 0x21; 0x00; 0x42; 0x00; 0x42; 0x00;
   0x84; 0x00; 0x84; 0x00; 0x08; 0x01; 0x08; 0x01;
   0x10; 0x02; 0x10; 0x02];
value im5 =
  [0x02; 0x07; 0x02; 0x00; 0x20; 0x70; 0x20; 0x00];
value im6 =
  [0x05; 0x00; 0x05; 0x00];
value im7 =
  [0x00; 0x01; 0x40; 0x00; 0x10; 0x00; 0x04; 0x00;
   0x01; 0x00; 0x00; 0x01; 0x40; 0x00; 0x10; 0x00;
   0x04; 0x00; 0x01; 0x00];
value im8 =
  [0x40; 0x11; 0x04; 0x11; 0x40; 0x11; 0x04; 0x11];

value make_pixmaps dd black white = do {
  let pix1 = make_image_pixmap dd "pixm1" 16 16 im1 in
  let pix2 = make_image_pixmap dd "pixm2" 16 16 im2 in
  let pix3 = make_image_pixmap dd "pixm3" 16 16 im3 in
  let pix4 = make_image_pixmap dd "pixm4" 10 10 im4 in
  let pix5 = make_image_pixmap dd "pixm5" 8 8 im5 in
  let pix6 = make_image_pixmap dd "pixm6" 4 4 im6 in
  let pix7 = make_image_pixmap dd "pixm7" 10 10 im7 in
  let pix8 = make_image_pixmap dd "pixm8" 8 8 im8 in
  [| pix1; pix2; pix3; pix4; pix5; pix6; pix7; pix8 |]
};

value make_graph xd ts2 co = do {
  let ts = ts2.ast in
  let many_tasks = Dim.many_tasks ts.tasks_def in
  let (dam_width, dam_height) =
    let h = screen_height xd in
    if many_tasks then (h * 4 / 10, h / 3)
    else (h * 5 / 10, h / 2)
  in
  let rt_dim = Dim.dim_of_width_height ts2 dam_width dam_height in
  let ps_dim = Dim.dim_of_width_height ts2 716 512 in
  let inter_task_dy = if many_tasks then 35 else 40 in
  let dd =
    {postscript = None; text_buf = Buffer.create 1; ps_A3 = False;
     ps_landscape = True; rt_dim = rt_dim; ps_dim = ps_dim; sched_width = 0;
     sched_height = 0; inter_task_dy = inter_task_dy; da_co = co}
  in
  let term_font =
    let size = if many_tasks then 10 else 14 in
    [| rt_load_query_font xd
         (sprintf "-*-courier-medium-r-normal-*-%d-*-*-*-*-*-iso8859-*" size);
       rt_load_query_font xd
         (sprintf "-*-courier-bold-r-normal-*-%d-*-*-*-*-*-iso8859-*" size);
       rt_load_query_font xd
         (sprintf "-*-courier-medium-o-normal-*-%d-*-*-*-*-*-iso8859-*" size);
       rt_load_query_font xd
         (sprintf "-*-courier-bold-o-normal-*-%d-*-*-*-*-*-iso8859-*" size) |]
  in
  let colors = Array.map (da_create_color dd) co.c_color in
  let red_col = da_create_color dd co.c_red_col in
  let blue_col = da_create_color dd co.c_blue_col in
  let backg_rep_col = da_create_color dd co.c_backg_rep_col in
  let gray = da_create_color dd co.c_gray_col in
  let black = da_black_color dd in
  let white = da_white_color dd in
  let pixmaps = make_pixmaps dd black white in
  let raw_font =
    {da_font = co.c_raw_font; da_f_name = "raw"; da_f_size = 12; da_f_dd = dd}
  in
  let bigger_font =
    {da_font =
       rt_load_query_font xd
         "-*-courier-bold-r-normal-*-18-*-*-*-*-*-iso8859-*";
     da_f_name = "bigger";
     da_f_size = 18;
     da_f_dd = dd}
  in
  let smaller_font =
    {da_font =
       rt_load_query_font xd
         "-*-courier-bold-r-normal-*-8-*-*-*-*-*-iso8859-*";
     da_f_name = "smaller";
     da_f_size = 8;
     da_f_dd = dd}
  in
  {zooms_sch = []; zooms_dam = []; resp_f_time = []; color = colors;
   pixmap = pixmaps; black_col = black; white_col = white;
   red_col = red_col; blue_col = blue_col; backg_rep_col = backg_rep_col;
   gray = gray; raw_font = raw_font; bigger_font = bigger_font;
   smaller_font = smaller_font; term_font = term_font; dd = dd}
};

value make_scenario xd co (ts2, fname) = do {
  let graph = make_graph xd ts2 co in
  Scenario.make_scenario_gen (ts2, fname) graph
};

value empty_scenario xd co name =
  match Sparse.parse_and_check_dameid (Stream.of_string "0/") with
  [ [(_, df)] -> do {
      let ts2 = Scenario.make_tasks df in
      make_scenario xd co (ts2, name)
    }
  | [] | [_ :: _] -> failwith "internal error" ]
;

value current_scenario tg = do {
  if Array.length tg.scenario = 0 then do {
    let sc = empty_scenario tg.common.xd tg.common "" in
    tg.scenario := [| sc |];
    tg.ind_of_isc := [| 0 |];
    tg.curr_sc := 0;
  }
  else ();
  tg.scenario.(tg.curr_sc)
};

value set_scenario_menus xd sc = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let tc = ts.constraints in
  check_set (widget_named xd "prec_constr") tc.global_precedence;
  check_set (widget_named xd "strict_per_constr") ts2.strict_periodicity;
  check_set (widget_named xd "no_preempt_constr") ts2.fully_non_preemptive;
  let (dead, rate, edf, auds, user) =
    match ts.algorithm with
    [ AL_deadline_monotonic -> (True, False, False, False, False)
    | AL_rate_monotonic -> (False, True, False, False, False)
    | AL_edf -> (False, False, True, False, False)
    | AL_audsley _ -> (False, False, False, True, False)
    | AL_user _ -> (False, False, False, False, True) ]
  in
  check_set (widget_named xd "dead_mono_algo") dead;
  check_set (widget_named xd "rate_mono_algo") rate;
  check_set (widget_named xd "edf_algo") edf;
  check_set (widget_named xd "audsley_algo") auds;
  check_set (widget_named xd "user_prio_algo") user;
};

value raise_scenario gd ti isc = do {
  let xd = gd.common.xd in
  Tabwid.tab_set xd ti isc;
  gd.curr_sc := isc;
  rt_raise_widget (Dutil.g_widget_named gd "pack");
  Dutil.redirect_to_main gd;
  let sc = gd.scenario.(isc) in
  set_scenario_menus xd sc;
  Scenario.set_task_order sc.ts;
  let fname = sc.fname in
  let t = if fname = "-" then "" else " - " ^ fname in
  let wname = "Dameid" ^ t in
  rt_change_widget_name (widget_named xd "main") wname "Dameid"
};

value erase_scenario tg ti isc = do {
  let ind_of_isc =
    Array.init (Array.length tg.ind_of_isc - 1)
      (fun j -> tg.ind_of_isc.(if j < isc then j else j + 1))
  in
  tg.ind_of_isc := ind_of_isc;
  let scenario =
    Array.init (Array.length tg.scenario - 1)
      (fun j -> tg.scenario.(if j < isc then j else j + 1))
  in
  tg.scenario := scenario;
  let xd = tg.common.xd in
  Tabwid.tab_erase xd ti isc;
  if tg.curr_sc > isc then do {
    tg.curr_sc := tg.curr_sc - 1;
    Tabwid.tab_set xd ti tg.curr_sc;
    let sc = tg.scenario.(tg.curr_sc) in
    let fname = sc.fname in
    let t = if fname = "-" then "" else " - " ^ fname in
    let wname = "Dameid" ^ t in
    rt_change_widget_name (widget_named xd "main") wname "Dameid"
  }
  else if tg.curr_sc = isc then do {
    let new_i =
      if isc = Array.length tg.scenario then max 0 (isc - 1) else isc
    in
    Tabwid.tab_set xd ti new_i;
    tg.curr_sc := new_i;
    let wname =
      if Array.length tg.scenario = 0 then do {
        rt_raise_widget (widget_named xd "main_text_pack");
        "Dameid"
      }
      else do {
        rt_raise_widget (Dutil.g_widget_named tg "pack");
        let sc = tg.scenario.(tg.curr_sc) in
        set_scenario_menus xd sc;
        let fname = sc.fname in
        let t = if fname = "-" then "" else " - " ^ fname in
        "Dameid" ^ t
      }
    in
    rt_change_widget_name (widget_named xd "main") wname "Dameid"
  }
  else ();
  ignore (rt_adjust_widget (widget_named xd "main") : (int * int));
};

(* Scenario desc *)

value present_index gd i =
  loop 0 where rec loop j =
    if j = Array.length gd.ind_of_isc then None
    else if gd.ind_of_isc.(j) = i then Some j
    else loop (j + 1)
;

value action_schedule xa tg ti i def_keyp _ ev =
  match present_index tg i with
  [ Some isc ->
      let gfun = tg.g_fun in
      let sc = tg.scenario.(isc) in
      let sched_wid = widget_named tg.common.xd (sprintf "schedule_%d" i) in
      match ev with
      [ RawEvExpose _ _ _ _ -> do {
          sc.expose_sched_cnt := sc.expose_sched_cnt + 1;
          try while rt_pending_events xa do { rt_treat_one_event xa } with
            e -> do {
              sc.expose_sched_cnt := sc.expose_sched_cnt - 1;
              raise e
            };
          sc.expose_sched_cnt := sc.expose_sched_cnt - 1;
          if sc.expose_sched_cnt = 0 then do {
            sc.graph.dd.sched_width := widget_width sched_wid;
            sc.graph.dd.sched_height := widget_height sched_wid;
            Draw_schedule_rt.expose_schedule gfun tg isc sched_wid
          }
          else ();
        }
      | RawEvConfigureNotify _ (width, height, _) -> do {
          if width <> sc.graph.dd.sched_width ||
             height <> sc.graph.dd.sched_height
          then do {
            sc.graph.dd.sched_width := width;
            sc.graph.dd.sched_height := height;
            sc.expose_sched_cnt := sc.expose_sched_cnt + 1;
            try while rt_pending_events xa do { rt_treat_one_event xa } with
              e -> do {
                sc.expose_sched_cnt := sc.expose_sched_cnt - 1;
                raise e
              };
            sc.expose_sched_cnt := sc.expose_sched_cnt - 1;
            if sc.expose_sched_cnt = 0 then
              Draw_schedule_rt.expose_schedule gfun tg isc sched_wid
            else ();
          }
          else ();
        }
      | RawEvButtonPress x y x_root y_root mbut -> do {
          match mbut.item with
          [ 1 -> do {
              let sched_height = sc.graph.dd.sched_height in
              if x >= 0 && y >= 0 && x < sc.graph.dd.sched_width &&
                 y < sc.graph.dd.sched_height && y < sched_height
              then do {
                Zoom.zoom_schedule xa tg ti sc def_keyp sched_wid x y
                  x_root y_root
              }
              else ()
            }
          | _ -> () ]
        }
      | _ -> () ]
  | _ -> () ]
;

value action_mesoid tg sc mes_wid =
  fun
  [ RawEvExpose _ _ _ _ | RawEvConfigureNotify _ _ -> do {
      let gfun = tg.g_fun in
      let gc = ref DaNonePn in
      rt_clear_widget mes_wid;
      let mes_drw = Draw_schedule_rt.mesoid_drawable mes_wid sc.graph.dd in
      Draw_schedule.draw_mesoid gfun sc gc mes_drw;
    }
  | _ -> () ]
;

value action_dameid xa tg ti isc def_keyp dam_wid ev =
  let gc = ref DaNonePn in
  let gfun = tg.g_fun in
  match present_index tg isc with
  [ Some isc ->
      let sc = tg.scenario.(isc) in
      let dam_drw =
        Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
      in
      match ev with
      [ RawEvExpose _ _ _ _ -> do {
          sc.expose_dameid_cnt := sc.expose_dameid_cnt + 1;
          try while rt_pending_events xa do { rt_treat_one_event xa } with
            e -> do {
              sc.expose_dameid_cnt := sc.expose_dameid_cnt - 1;
              raise e
            };
          sc.expose_dameid_cnt := sc.expose_dameid_cnt - 1;
          if sc.expose_dameid_cnt = 0 then do {
            da_clear_widget dam_drw;
            Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
            Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
            Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom
              False;
            da_commit_display dam_drw;
          }
          else ();
        }
      | RawEvConfigureNotify (ow, oh, _) (nw, nh, _) ->
          if nw <> ow || nh <> oh then do {
            Schedule_step_rt.recompute_dim sc nw nh;
            sc.expose_dameid_cnt := sc.expose_dameid_cnt + 1;
            try while rt_pending_events xa do { rt_treat_one_event xa } with
              e -> do {
                sc.expose_dameid_cnt := sc.expose_dameid_cnt - 1;
                raise e
              };
            sc.expose_dameid_cnt := sc.expose_dameid_cnt - 1;
            if sc.expose_dameid_cnt = 0 then do {
              da_clear_widget dam_drw;
              Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
              Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
              Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom False;
              da_commit_display dam_drw;
            }
            else ();
          }
          else ()
      | RawEvButtonPress x y x_root y_root mbut -> do {
          match mbut.item with
          [ 1 -> do {
              if x >= 0 && y >= 0 && x < sc.graph.dd.rt_dim.dam_width &&
                 y < sc.graph.dd.rt_dim.dam_height
              then
                Zoom.zoom_dameid xa tg ti sc def_keyp dam_wid x y x_root
                  y_root
              else ()
            }
          | _ -> () ]
        }
      | _ -> () ]
  | None -> () ]
;

value action_caption tg isc cap_wid ev =
  let gfun = tg.g_fun in
  match present_index tg isc with
  [ Some isc ->
      let sc = tg.scenario.(isc) in
      let gc = ref DaNonePn in
      let cap_drw = Draw_dameid_rt.caption_drawable cap_wid sc.graph.dd in
      match ev with
      [ RawEvExpose _ _ _ _ -> Draw_dameid.draw_caption gfun sc cap_drw gc
      | _ -> () ]
  | None -> () ]
;

value scenario_desc xa tg ti def_keyp isc =
  let sc = tg.scenario.(isc) in
  let isc = tg.ind_of_isc.(isc) in
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let many_tasks = Dim.many_tasks ts.tasks_def in
  let backg = BackgroundAtt (ColorPn sc.graph.white_col.da_color) in
  let sched_hei = Dim.schedule_height (Array.length ts.tasks_def) in
(**)
  let sched_hei = min (screen_height tg.common.xd * 7 / 20) sched_hei in
(**)
  pack_desc
    [backg; BorderAtt 0; BandAtt 0; InterAtt 0;
     NameAtt (sprintf "pack_%d" isc)]
    (Vertical,
     [pack_desc [backg; BorderAtt 0; BandAtt 0]
        (InDepth,
         [raw_desc
(**)
            [NameAtt (sprintf "schedule_%d" isc);
             BackgroundAtt (PixmapPn tg.common.pixmap_sched)]
(*
            [NameAtt (sprintf "a_schedule_%d" isc)]
*)
            (1, sched_hei, 0,
             [SelExposure; SelKeyPress; SelStructureNotify;
              SelButtonPress])
            (action_schedule xa tg ti isc def_keyp);
          raw_desc [NameAtt (sprintf "mesoid_%d" isc)]
            (1, sched_hei, 0,
             [SelExposure; SelKeyPress; SelStructureNotify])
            (action_mesoid tg sc)])
        (fun _ _ -> ());
      pack_desc
        [backg; BorderAtt 0; BandAtt 0; InterAtt 0; FillerAtt]
        (Horizontal,
         [raw_desc
            [NameAtt (sprintf "dameid_%d" isc); BorderAtt 0; BandAtt 0;
             BackgroundAtt (PixmapPn tg.common.pixmap_dameid)]
            (sc.graph.dd.rt_dim.dam_width, sc.graph.dd.rt_dim.dam_height, 0,
             [SelExposure; SelKeyPress; SelButtonPress;
              SelStructureNotify])
            (action_dameid xa tg ti isc def_keyp);
          raw_desc
            [NameAtt (sprintf "caption_%d" isc);
             BackgroundAtt (PixmapPn tg.common.pixmap_caption)]
            (120, 0, 0, [SelExposure; SelStructureNotify])
            (action_caption tg isc);
          pack_desc [FillerAtt; backg] (Vertical, []) (fun _ _ -> ());
          pack_desc [backg]
            (Vertical,
             [term_desc
                [BorderAtt 0; NameAtt (sprintf "text_%d" isc);
                 FontAtt sc.graph.term_font]
                (30, if many_tasks then 54 else 46, 0)
                (fun _ _ -> ())])
            (fun _ _ -> ())])
        (fun _ _ -> ())])
    (fun _ _ -> ())
;

value make_moving_schedule_widget gd i = if True then () else do {
  let gfun = gd.g_fun in
  let xd = gd.common.xd in

  let wid2 = widget_named xd (sprintf "a_schedule_%d" i) in
  let w2 = screen_width xd in
  let h2 () =
    let sc = gd.scenario.(i) in
    Dim.schedule_height (Array.length sc.ts.ast.tasks_def)
  in
  let wid_x_ini = 0 in
  let wid_y_ini = 0 in
  let xy = ref (wid_x_ini, wid_y_ini, 0, 0) in
  let moving = ref False in
  let move_plateau wid1 x y = do {
    let px_min = 0 in
    let px_max = 0 in
    let py_min = widget_height wid2 - h2 () in
    let py_max = 0 in
    let (wid_x, wid_y, x1, y1) = xy.val in
    let px = wid_x - x1 + x in
    let py = wid_y - y1 + y in
    let px = min (max px_min px) px_max in
    let py = min (max py_min py) py_max in
    rt_move_widget wid1 px py;
    xy.val := (px, py, x1, y1);
  }
  in
  let wid1 =
    rt_create_widget xd "" "" AutoPosition None
      (raw_desc
         [BackgroundAtt (PixmapPn gd.common.pixmap_sched);
          NameAtt (sprintf "schedule_%d" i)]
         (w2, h2 (), 0,
          [SelButtonPress; SelButtonRelease; SelButtonMotion; SelExposure;
           SelStructureNotify])
         (fun wid ->
            fun
            [ RawEvExpose _ _ _ _ -> do {
                match present_index gd i with
                [ Some isc -> do {
                    let sc = gd.scenario.(isc) in
                    let w = widget_width wid2 in
                    let h = h2 () in
                    sc.graph.dd.sched_width := w;
                    sc.graph.dd.sched_height := h;
(*
                    rt_resize_widget wid w h;
*)
                  }
                | None -> () ];
                let (wid_x, wid_y, x1, y1) = xy.val in
                rt_move_widget wid wid_x wid_y;
                Draw_schedule_rt.expose_schedule gfun gd i wid
              }
            | RawEvConfigureNotify _ (nw, nh, _) -> do {
                match present_index gd i with
                [ Some isc -> do {
                    let sc = gd.scenario.(isc) in
                    sc.graph.dd.sched_width := widget_width wid2;
                    sc.graph.dd.sched_height := h2 ();
                  }
                | None -> () ];
                let (wid_x, wid_y, x1, y1) = xy.val in
                rt_move_widget wid wid_x wid_y;
                Draw_schedule_rt.expose_schedule gfun gd i wid
              }
            | RawEvButtonPress x y _ _ mbut ->
                match mbut.item with
                [ 1 | 2 | 3 -> do {
                    let (wid_x, wid_y, _, _) = xy.val in
                    xy.val := (wid_x, wid_y, x, y);
                    rt_select_mouse wid gd.common.hand;
                    moving.val := True
                  }
                | 4 | 5 -> do {
                    let (wid_x, wid_y, x1, y1) = xy.val in
                    let x = x1 in
                    let y = if mbut.item = 5 then y1 - 40 else y1 + 40 in
                    move_plateau wid x y;
                  }
                | _ -> () ]
            | RawEvButtonRelease _ _ _ _ _ -> do {
                rt_unselect_mouse wid;
                moving.val := False
              }
            | RawEvMotionNotify x y ->
                if moving.val then move_plateau wid x y else ()
            | _ ->
                () ]))
  in
  rt_reparent_widget wid1 wid2 wid_x_ini wid_y_ini;
  rt_map_widget wid1;
};
