(* $Id: schedule_step.ml,v 1.195 2010-12-05 14:58:48 deraugla Exp $ *)

open Ast;
open Dtasks;

value gen_backward_schedule ts2 = do {
  let ts = ts2.ast in
  let tc = ts.constraints in
  loop () where rec loop () = do {
    if ts2.curr_ord_ind = 0 && ts2.curr_inst = 0 then
      False
    else if ts2.curr_ord_ind = 0 && ts2.curr_inst = 1 then
      True
    else if ts2.curr_inst >= 2 then do {
      ts2.curr_inst := ts2.curr_inst - 2;
      ts2.curr_step := ts2.curr_step - 2;
      try Schedule.schedule_next_instance ts2 with
      [ Schedule.Schedule_interrupted _ -> () ];
      False
    }
    else do {
      let nb_inst =
        let itask = ts2.task_order.(ts2.curr_ord_ind - 1) in
        (* hack to find the number of instances *)
        loop 0 0 where rec loop max_inst j =
          if j = ts2.schedule_len then max_inst + 1
          else
            match Sarray.get ts2.disp_schedule j with
            [ Some u ->
                let max_inst =
                  if u.u_task = itask then max max_inst u.u_inst
                  else max_inst
                in
                loop max_inst (j + 1)
            | None ->
                loop max_inst (j + 1) ]
      in
      if tc.global_precedence || tc.graph_precedence <> [] then do {
        if ts2.curr_ord_ind < Array.length ts.tasks_def then do {
          let iorder = ts2.curr_ord_ind in
          let itask = ts2.task_order.(iorder) in
          let td = ts.tasks_def.(itask) in
          let r = if tc.global_precedence then 0 else td.user_release in
          let td = {(td) with release = r} in
          ts.tasks_def.(itask) := td;
        }
        else ();
        Schedule.recompute_init ts2;
        ts2.do_refresh := True;
      }
      else ();
      ts2.curr_step := ts2.curr_step - 1;
      ts2.curr_ord_ind := ts2.curr_ord_ind - 1;
      if nb_inst > 1 then do {
        ts2.curr_inst :=
          if ts2.curr_inst = 1 then nb_inst - 1 else nb_inst - 2;
        Schedule.schedule_next_instance ts2;
        False
      }
      else if ts2.curr_inst = 1 then do {
        ts2.curr_inst := 0;
        Schedule.schedule_next_instance ts2;
        False
      }
      else do {
        ts2.curr_inst := 1;
        loop ()
      }
    };
  };
};

value backward_schedule_init ts = do {
  match ts.repetition with
  [ Some _ -> do {
      ts.repetition := None;
      ts.do_refresh := True;
    }
  | None -> () ];
  match ts.schedule_status with
  [ SS_schedulable_but _ | SS_several_lcm _ -> do {
      ts.schedule_status := SS_running;
      ts.do_refresh := True;
    }
  | SS_task_error itask _ | SS_incomplete_scheduling itask _ -> do {
      ts.schedule_status := SS_running;
      if ts.curr_ord_ind < Array.length ts.ast.tasks_def &&
         ts.task_order.(ts.curr_ord_ind) = itask &&
         ts.curr_inst <> max_int
      then
        ts.curr_inst := ts.curr_inst + 1
      else ()
    }
  | SS_running -> () ];
  List.iter
    (fun l -> do {
       match l.l_status with
       [ LS_values _ _ _ -> ts.do_refresh := True
       | _ -> () ];
       l.l_status := LS_unchecked
     })
    ts.latency_constr;
};
