(* $Id: schedule_step_rt.ml,v 1.10 2010-12-06 16:09:05 deraugla Exp $ *)

open Printf;
open Rt;

open Ast;
open Da;
open Data_rt;
open Dim_def;
open Dutil;
open Glob_glob;
open Dtasks;

(* *)

value recompute_dim sc dam_width dam_height = do {
  sc.graph.dd.rt_dim := Dim.dim_of_width_height sc.ts dam_width dam_height;
  sc.latency_disp_cnt := 0;
};

value unfreeze_d wid zi =
  match zi.zi_d with
  [ Some d -> d
  | None -> do {
      let (dx, dy, _, _, _) = rt_query_pointer wid in
      let dx = widget_width wid / 2 - dx in
      let dy = widget_height wid / 2 - dy in
      zi.zi_d := Some (dx, dy);
      (dx, dy)
    } ]
;

value expose_zooms tg sc = do {
  List.iter
    (fun (_, wid, zi) -> do {
       rt_clear_widget wid;
       Expose.expose_zoom_schedule tg sc zi wid;
     })
    sc.graph.zooms_sch;
  List.iter
    (fun (_, wid, zi) -> do {
       rt_clear_widget wid;
       Expose.expose_zoom_dameid tg sc zi wid;
     })
    sc.graph.zooms_dam;
};

value update_draw_zooms tg sc gc color itask j = do {
  let gfun = tg.g_fun in
  List.iter
    (fun (_, wid, zi) -> do {
       let zoom =
         let (dx, dy) = unfreeze_d wid zi in
         let (dx2, dy2) = zi.zi_d2 in
         Zoom zi wid  (dx + dx2) (dy + dy2)
       in
       let drw =
         Draw_schedule_rt.schedule_drawable_of_zoom wid sc.graph.dd zoom
       in
       Draw_schedule.draw_unit gfun sc drw gc zoom color itask j;
     })
    sc.graph.zooms_sch;
  let ts = sc.ts in
  let sh_dam = if ts.rmax = 0 then 0 else ts.rmax + ts.lcm in
  if j >= sh_dam then
    List.iter
      (fun (_, wid, zi) -> do {
         let zoom =
           let (dx, dy) = unfreeze_d wid zi in
           let (dx2, dy2) = zi.zi_d2 in
           Zoom zi wid (dx + dx2) (dy + dy2)
         in
         let isect = (j + ts.min_rel_time) mod ts.lcm in
         let isect = if isect < 0 then isect + ts.lcm else isect in
         let drw =
           Draw_dameid_rt.dameid_drawable_of_zoom wid sc.graph.dd zoom
         in
         Draw_dameid.draw_sector gfun sc drw gc zoom isect
       })
      sc.graph.zooms_dam
  else ();
};

value draw_zooms_skeletons tg sc = do {
  let gfun = tg.g_fun in
  let gc = ref DaNonePn in
  List.iter
    (fun (_, wid, zi) -> do {
       let zoom =
         let (dx, dy) = unfreeze_d wid zi in
         let (dx2, dy2) = zi.zi_d2 in
         Zoom zi wid (dx + dx2) (dy + dy2)
       in
       let drw =
         Draw_schedule_rt.schedule_drawable_of_zoom wid sc.graph.dd zoom
       in
       Draw_schedule.draw_skeleton gfun sc drw gc zoom;
     })
    sc.graph.zooms_sch;
  List.iter
    (fun (_, wid, zi) -> do {
       let zoom =
         let (dx, dy) = unfreeze_d wid zi in
         let (dx2, dy2) = zi.zi_d2 in
         Zoom zi wid (dx + dx2) (dy + dy2)
       in
       let drw =
         Draw_dameid_rt.dameid_drawable_of_zoom wid sc.graph.dd zoom
       in
       Draw_dameid.draw_skeleton gfun sc drw gc zoom;
     })
    sc.graph.zooms_dam;
};

value update_draw tg sc dam_wid sched_wid mes_wid = do {
  let dam_drw =
    Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
  in
  let sched_drw =
    Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd NoZoom
  in
  let mes_drw = Draw_schedule_rt.mesoid_drawable mes_wid sc.graph.dd in
  let gc = ref DaNonePn in

  Draw_schedule.gen_update_draw tg.g_fun sc gc dam_drw sched_drw mes_drw
    (update_draw_zooms tg sc gc);
  draw_zooms_skeletons tg sc;
};

value goto_begin tg =
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    Schedule.goto_begin_schedule sc.ts;
    recompute_dim sc sc.graph.dd.rt_dim.dam_width
      sc.graph.dd.rt_dim.dam_height;
    let main_wid = widget_named tg.common.xd "main" in
    let text_wid = Dutil.g_widget_named tg "text" in
    rt_select_mouse main_wid tg.common.watch;
    rt_select_mouse text_wid tg.common.watch;
    rt_sync tg.common.xd;
    let dam_wid = Dutil.g_widget_named tg "dameid" in
    let text_wid = Dutil.g_widget_named tg "text" in
    Expose.redisplay tg dam_wid;
    Expose.draw_text sc text_wid;
    rt_unselect_mouse main_wid;
    rt_unselect_mouse text_wid;
  }
  else ()
;

value goto_normal_end tg sc = do {
  let gfun = tg.g_fun in
  let main_wid = widget_named tg.common.xd "main" in
  let dam_wid = Dutil.g_widget_named tg "dameid" in
  let text_wid = Dutil.g_widget_named tg "text" in
  rt_select_mouse main_wid tg.common.watch;
  rt_select_mouse text_wid tg.common.watch;
  rt_freeze_widget main_wid;
  gfun.g_treat_pending_events ();
  let ts = sc.ts in
  try
    try Schedule.schedule_all ts with
    [ Schedule.Schedule_interrupted sf -> ts.schedule_status := sf ]
  with e -> do {
    rt_unfreeze_widget main_wid;
    rt_unselect_mouse main_wid;
    raise e;
  };
  let comp_rep_part =
    match ts.schedule_status with
    [ SS_schedulable_but _ | SS_several_lcm _ -> True
    | SS_task_error _ _ | SS_incomplete_scheduling _ _ -> False
    | SS_running -> ts.curr_ord_ind = Array.length ts.ast.tasks_def ]
  in
  if comp_rep_part then
    match Schedule.compute_repetitive_part ts with
    [ Some r -> do {
        ts.repetition := Some r;
        ts.do_refresh := True
      }
    | None -> () ]
  else ();
  for i = 0 to ts.schedule_len - 1 do {
    Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
  };
  sc.latency_disp_cnt := 0;
  Expose.redisplay tg dam_wid;
  Expose.draw_text sc text_wid;
  rt_unfreeze_widget main_wid;
  rt_unselect_mouse main_wid;
  rt_unselect_mouse text_wid;
};

value no_action _ _ = ();

value (progr_bar_info, get_progr_bar_info) =
  user_info "progr_bar_info" (ref UNone)
;

value destroy_progr_bar wid = do {
  let xd = xdata_of_widget wid in
  let pwid = widget_named xd "progr_bar" in
  let (_, (col, col2)) = get_progr_bar_info pwid in
  Array.iter rt_free_color col;
  rt_free_color col2;
  rt_destroy_widget wid;
};

value close_progr_bar xd =
  match
    try Some (widget_named xd "progr_bar_main") with [ Failure _ -> None ]
  with
  [ Some wid -> destroy_progr_bar wid
  | None -> () ]
;

value expose_progr_bar gd wid (col, col2) n ntot = do {
  let drw = WidgetDr wid in
  let xd = xdata_of_widget wid in
  let w = widget_width wid in
  let h = widget_height wid in
  let x = 10 in
  let y = min 10 (max 1 ((h - 12) / 2)) in
  let bh = min (Array.length col) (max 1 (h - 2 * y - 1)) in
  if ntot = 0 then ()
  else do {
    let wb = truncate (float (w - 20) *. float n /. float ntot) in
    rt_set_line_width xd 1;
    for i = 0 to bh - 1 do {
      rt_select_color col.(i);
      rt_draw_line drw (10, y + 1 + i) (10 + wb, y + 1 + i);
    };
    rt_select_color (fst gd.common.c_white_col);
    rt_fill_rectangle drw (x+wb, y, w-x-wb, bh+1);
  };
  rt_select_color col2;
  rt_draw_rectangle drw (x, y, w-20, bh+1);
  let wid = widget_named xd "progr_bar_txt" in
  let (_, cols) = term_size wid in
  let str = if ntot = 0 then "0" else sprintf "%d/%d" n ntot in
  let s =
    sprintf "%s%s%s%s%s" Std_term.vt_home Std_term.vt_hide_cursor
      (String.make (max 0 ((cols - String.length str) / 2)) ' ') str
      Std_term.vt_erase_from_cursor
  in
  term_send wid s;
};

value expose_solution xd ts =
  match ts.algorithm with
  [ AL_audsley a ->
      match
        try Some (widget_named xd "progr_bar_nsol") with
        [ Failure _ -> None ]
      with
      [ Some wid ->
          let (_, cols) = term_size wid in
          let str =
            match a.ai_pos with
            [ P_begin -> ""
            | P_middle -> sprintf "%s %d" (I18n.transl "solution") a.ai_nsol
            | P_end -> sprintf "%d %s" a.ai_nsol (I18n.transl "solution(s)") ]
          in
          let s =
            sprintf "%s%s%s%s%s" Std_term.vt_home Std_term.vt_hide_cursor
              (String.make (max 0 ((cols - String.length str) / 2)) ' ')
              str Std_term.vt_erase_from_cursor
          in
          term_send wid s
      | None -> () ]
  | AL_deadline_monotonic | AL_rate_monotonic | AL_edf | AL_user _ ->
      () ]
;

value clear_expose_progr_bar gd ts wid = do {
  rt_clear_widget wid;
  let (pbi, col) = get_progr_bar_info wid in
  let (n, ntot) = pbi.val in
  expose_progr_bar gd wid col n ntot;
  let xd = xdata_of_widget wid in
  expose_solution xd ts
};

value action_progr_bar gd ts wid =
  fun
  [ RawEvExpose _ _ _ _ -> clear_expose_progr_bar gd ts wid
  | _ -> () ]
;

exception CancelSchedule;

value center_of_subwidget main_wid wid =
  let x = widget_x main_wid in
  let y = widget_y main_wid in
  let w = widget_width main_wid in
  let h = widget_height main_wid in
  (x + w / 2 - widget_width wid / 2, y + h / 2 - widget_height wid / 2)
;

value main_window_close_action = ref (fun []);

value trace_start xa gd ts main_wid = do {
  let xd = xdata_of_widget main_wid in
  let wid =
    match
      try Some (widget_named xd "progr_bar_main") with [ Failure _ -> None ]
    with
    [ Some wid -> do {
        rt_clear_widget (widget_named xd "progr_bar");
        rt_unfreeze_widget (widget_named xd "progr_bar_cancel");
        wid
      }
    | None -> do {
        let filler =
          pack_desc [FillerAtt; BorderAtt 0; BandAtt 0] (Vertical, [])
            no_action
        in
        let wid =
          rt_create_transient_widget main_wid "Audsley++"
            (Some destroy_progr_bar)
            (pack_desc [NameAtt "progr_bar_main"]
               (Vertical,
                [pack_desc []
                   (Horizontal,
                    [filler;
                     term_desc [NameAtt "progr_bar_txt"] (1, 20, 0) no_action;
                     filler])
                   no_action;
                 pack_desc []
                   (Horizontal,
                    [raw_desc [NameAtt "progr_bar"; FillerAtt]
                       (400, 33, 0, [SelExposure; SelStructureNotify])
                       (action_progr_bar gd ts)])
                   no_action;
                 filler;
                 pack_desc []
                   (Horizontal,
                    [filler;
                     term_desc [NameAtt "progr_bar_nsol"] (1, 20, 0) no_action;
                     filler;
                     button_desc [] (I18n.transl "Dismiss", None)
                       (Dutil.button_press_action
                          (fun () -> do {
                             let wid = widget_named xd "progr_bar_main" in
                             destroy_progr_bar wid
                           }));
                     button_desc [NameAtt "progr_bar_cancel"]
                       (I18n.transl "Cancel", None)
                       (Dutil.button_press_action
                          (fun () -> do {
                             let wid = widget_named xd "progr_bar" in
                             rt_clear_widget wid;
                             let (pbi, col) = get_progr_bar_info wid in
                             expose_progr_bar gd wid col 0 0;
                             pbi.val := (0, 1);
                             raise CancelSchedule
                           }))])
                   no_action])
               no_action)
        in
        let col =
          Array.map (rt_create_color xd)
            [| (0xFF, 0xC2, 0x99); (0xFF, 0xBB, 0x8D); (0xFF, 0xB3, 0x81);
               (0xFF, 0xAC, 0x74); (0xFF, 0xA5, 0x68); (0xFF, 0x9E, 0x5C);
               (0xFD, 0x73, 0x17); (0xFE, 0x74, 0x18); (0xFF, 0x84, 0x33);
               (0xFF, 0xB5, 0x84); (0xFF, 0x8A, 0x3C) |]
        in
        let col2 = rt_create_color xd (0x73, 0x6A, 0x60) in
        let ui = progr_bar_info (ref (0, 1), (col, col2)) in
        rt_set_user_info (widget_named xd "progr_bar") ui;
        let (x, y) = center_of_subwidget main_wid wid in
        rt_map_transient_widget wid x y;
        wid
      } ]
  in
  let mwca = main_window_close_action.val in
  main_window_close_action.val := fun _ -> raise CancelSchedule;
  (wid, mwca)
};

value trace_step xa gd (wid, _) (n, ntot) = do {
  let xd = xdata_of_widget wid in
  match
    try Some (widget_named xd "progr_bar") with [ Failure _ -> None ]
  with
  [ Some wid -> do {
      let (pbi, col) = get_progr_bar_info wid in
      if n = 0 then rt_clear_widget wid else ();
      expose_progr_bar gd wid col n ntot;
      pbi.val := (n, ntot);
    }
  | None -> () ];
  rt_treat_pending_events xa
};

value trace_end xa (wid, mwca) = do {
  main_window_close_action.val := mwca;
  let xd = xdata_of_widget wid in
  match
    try Some (widget_named xd "progr_bar_cancel") with
    [ Failure _ -> None ]
  with
  [ Some wid -> rt_freeze_widget wid
  | None -> () ]
};

value call_enclosed_with bef aft act = do {
  bef ();
  try act () with e -> do { aft (); raise e };
  aft ();
};

value goto_prev xa gd =
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    match ts.algorithm with
    [ AL_audsley auds ->
        if auds.ai_pos = P_begin then ()
        else do {
          let main_wid = widget_named gd.common.xd "main" in
          let dam_wid = Dutil.g_widget_named gd "dameid" in
          let text_wid = Dutil.g_widget_named gd "text" in
          call_enclosed_with
            (fun () -> do {
               rt_select_mouse main_wid gd.common.watch;
               rt_select_mouse text_wid gd.common.watch;
               rt_freeze_widget main_wid;
               while rt_pending_events xa do { rt_treat_one_event xa };
             })
            (fun () -> do {
               rt_unfreeze_widget main_wid;
               rt_unselect_mouse main_wid;
               rt_unselect_mouse text_wid;
             })
            (fun () -> do {
               let ti = trace_start xa gd ts main_wid in
               let found =
                 try
                   Schedule.prev_audsley_schedule ts2 (trace_step xa gd ti)
                     auds
                 with
                 [ CancelSchedule -> False
                 | e -> do {
                     trace_end xa ti;
                     raise e
                   } ]
               in
               trace_end xa ti;
               if not found then do {
                 let auds = {(auds) with ai_pos = P_begin} in
                 let ts = {(ts) with algorithm = AL_audsley auds} in
                 ts2.ast := ts;
                 ts2.schedule_status := SS_running;
                 goto_begin gd
               }
               else ();
               expose_solution gd.common.xd ts;
             });
          for i = 0 to ts2.schedule_len - 1 do {
            Sarray.set ts2.disp_schedule i (Sarray.get ts2.comp_schedule i);
          };
          sc.latency_disp_cnt := 0;
          Expose.redisplay gd dam_wid;
          Expose.draw_text sc text_wid;
        }
    | AL_deadline_monotonic | AL_rate_monotonic | AL_edf | AL_user _ ->
        goto_begin gd ]
  }
  else ()
;

value goto_audsley_end xa gd auds = do {
  if gd.curr_sc < Array.length gd.scenario then do {
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    if auds.ai_pos = P_end then ()
    else do {
      let main_wid = widget_named gd.common.xd "main" in
      let dam_wid = Dutil.g_widget_named gd "dameid" in
      let text_wid = Dutil.g_widget_named gd "text" in
      call_enclosed_with
        (fun () -> do {
           rt_select_mouse main_wid gd.common.watch;
           rt_select_mouse text_wid gd.common.watch;
           rt_freeze_widget main_wid;
           while rt_pending_events xa do { rt_treat_one_event xa };
         })
        (fun () -> do {
           rt_unfreeze_widget main_wid;
           rt_unselect_mouse main_wid;
           rt_unselect_mouse text_wid;
         })
        (fun () -> do {
           let ti = trace_start xa gd ts main_wid in
           let found =
             try
               Schedule.next_audsley_schedule ts2 (trace_step xa gd ti)
                 auds
             with
             [ CancelSchedule -> False
             | e -> do {
                 trace_end xa ti;
                 raise e
               } ]
           in
           trace_end xa ti;
           if not found then do {
             let auds = {(auds) with ai_pos = P_end} in
             let ts = {(ts) with algorithm = AL_audsley auds} in
             ts2.ast := ts;
             ts2.schedule_status := SS_running;
             goto_begin gd
           }
           else ();
           expose_solution gd.common.xd ts;
         });
      for i = 0 to ts2.schedule_len - 1 do {
        Sarray.set ts2.disp_schedule i (Sarray.get ts2.comp_schedule i);
      };
      sc.latency_disp_cnt := 0;
      Expose.redisplay gd dam_wid;
      Expose.draw_text sc text_wid;
    }
  }
  else ()
};

value goto_end xa tg = do {
  if tg.curr_sc < Array.length tg.scenario then do {
    let sc = tg.scenario.(tg.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    match ts.algorithm with
    [ AL_audsley auds -> do {
        goto_audsley_end xa tg auds
      }
    | AL_deadline_monotonic | AL_rate_monotonic | AL_user _ | AL_edf -> do {
        goto_normal_end tg sc
      } ]
  }
  else ();
};

value forward_not_audsley tg sc = do {
  let gfun = tg.g_fun in
  let gc = ref DaNonePn in
  let dam_wid = Dutil.g_widget_named tg "dameid" in
  let sched_wid = Dutil.g_widget_named tg "schedule" in
  let mes_wid = Dutil.g_widget_named tg "mesoid" in
  let ts = sc.ts in
  if ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
     ts.schedule_status <> SS_running
  then do {
    let ok_lat =
      List.fold_right
        (fun l ok_lat ->
           match l.l_status with
           [ LS_values be1 be2 lat -> [(l, be1, be2, lat) :: ok_lat]
           | _ -> ok_lat ])
        ts.latency_constr []
    in
    let len = List.length ok_lat in
    let dam_drw =
      Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
    in
    let sched_drw =
      Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd NoZoom
    in
    if len >= 2 && sc.latency_disp_cnt < len - 1 then do {
      Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom True;
      Draw_schedule.draw_latencies gfun sc sched_drw gc NoZoom True;
      Draw_dameid.draw_orient gfun sc dam_drw gc True;
      sc.latency_disp_cnt := sc.latency_disp_cnt + 1;
      Expose.expose tg;
      let txt_wid = Dutil.g_widget_named tg "text" in
      if sc.text_nb_lines < fst (term_size txt_wid) then
        da_term_send sc.graph.dd txt_wid (sprintf "\027[%dA" sc.text_nb_lines)
      else ();
      Expose.redraw_text sc txt_wid;
    }
    else ()
  }
  else do {
    match
      try do { Schedule.schedule_next_instance ts; None } with e -> Some e
    with
    [ None -> ()
    | Some (Schedule.Schedule_interrupted sf) -> ts.schedule_status := sf
    | Some e -> raise e ];
    let comp_rep_part =
      match ts.schedule_status with
      [ SS_schedulable_but _ | SS_several_lcm _ -> True
      | SS_task_error _ _ | SS_incomplete_scheduling _ _ -> False
      | SS_running -> ts.curr_ord_ind = Array.length ts.ast.tasks_def ]
    in
    if comp_rep_part then
      match Schedule.compute_repetitive_part ts with
      [ Some r -> do {
          ts.repetition := Some r;
          ts.do_refresh := True
        }
      | None -> () ]
    else ();
    let dam_drw =
      Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
    in
    let sched_drw =
      Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd
        NoZoom
    in
    let mes_drw = Draw_schedule_rt.mesoid_drawable mes_wid sc.graph.dd in
    if ts.do_refresh then do {
      recompute_dim sc sc.graph.dd.rt_dim.dam_width
        sc.graph.dd.rt_dim.dam_height;
      for i = 0 to ts.schedule_len - 1 do {
        Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
      };
      da_clear_widget sched_drw;
      Draw_schedule.draw_repetition gfun sc sched_drw gc NoZoom;
      Draw_schedule.draw_units gfun sc sched_drw gc NoZoom;
      Draw_schedule.draw_mesoid_units gfun sc mes_drw gc;
      ts.do_refresh := False;
      da_clear_widget dam_drw;
      Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
      expose_zooms tg sc;
    }
    else
      update_draw tg sc dam_wid sched_wid mes_wid;
    Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
    let cap_wid = Dutil.g_widget_named tg "caption" in
    let cap_drw = Draw_dameid_rt.caption_drawable cap_wid sc.graph.dd in
    Draw_dameid.draw_caption gfun sc cap_drw gc;
    Draw_schedule.draw_skeleton gfun sc sched_drw gc NoZoom;
    Draw_schedule.draw_mesoid_skeleton gfun sc mes_drw gc;
    if ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
       ts.schedule_status <> SS_running
    then do {
      Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom False;
      Draw_schedule.draw_latencies gfun sc dam_drw gc NoZoom False;
      Expose.draw_text sc (Dutil.g_widget_named tg "text")
    }
    else ();
    da_commit_display sched_drw;
    da_commit_display dam_drw;
  }
};

value forward xa gd =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    match sc.ts.ast.algorithm with
    [ AL_audsley a ->
        if a.ai_pos = P_begin then goto_end xa gd
        else forward_not_audsley gd sc
    | AL_deadline_monotonic | AL_rate_monotonic | AL_user _  | AL_edf ->
        forward_not_audsley gd sc ]
  else ()
;

value backward_schedule_not_edf tg sc sched_wid dam_wid = do {
  let gfun = tg.g_fun in
  let ts = sc.ts in
  let was_ended =
    ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
    ts.schedule_status <> SS_running
  in
  let must_goto_begin = Schedule_step.gen_backward_schedule ts in
  if must_goto_begin then goto_begin tg else ();

  let mes_wid = Dutil.g_widget_named tg "mesoid" in
  let cap_wid = Dutil.g_widget_named tg "caption" in
  let dam_drw =
    Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
  in
  let cap_drw = Draw_dameid_rt.caption_drawable cap_wid sc.graph.dd in
  let sched_drw =
    Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd NoZoom
  in
  let mes_drw = Draw_schedule_rt.mesoid_drawable mes_wid sc.graph.dd in
  let gc = ref DaNonePn in
  if ts.do_refresh then do {
    recompute_dim sc sc.graph.dd.rt_dim.dam_width sc.graph.dd.rt_dim.dam_height;
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
    };
    da_clear_widget sched_drw;
    rt_clear_widget cap_wid;
    Draw_schedule.draw_units gfun sc sched_drw gc NoZoom;
    rt_clear_widget mes_wid;
    Draw_schedule.draw_mesoid_units gfun sc mes_drw gc;
    ts.do_refresh := False;
    da_clear_widget dam_drw;
    Draw_dameid.draw_sectors gfun sc sc.graph dam_drw gc NoZoom;
    Expose.draw_text sc (Dutil.g_widget_named tg "text");
    expose_zooms tg sc;
  }
  else do {
    update_draw tg sc dam_wid sched_wid mes_wid;
    if was_ended then Expose.draw_text sc (Dutil.g_widget_named tg "text")
    else ();
  };
  Draw_dameid.draw_skeleton gfun sc dam_drw gc NoZoom;
  Draw_dameid.draw_caption gfun sc cap_drw gc;
  Draw_schedule.draw_skeleton gfun sc sched_drw gc NoZoom;
  Draw_schedule.draw_mesoid_skeleton gfun sc mes_drw gc;
  da_commit_display sched_drw;
  da_commit_display dam_drw;
};

value backward_schedule_edf gd sc = do {
  let ts = sc.ts in
  let current_i =
    loop (ts.schedule_len - 1) where rec loop i =
      match Sarray.get ts.disp_schedule i with
      [ Some _ -> i + 1
      | None -> if i = 0 then 0 else loop (i - 1) ]
  in
  if current_i = 1 then goto_begin gd
  else if current_i > 0 then do {
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.comp_schedule i (Sarray.get ts.disp_schedule i);
    };
    if ts.curr_ord_ind = Array.length ts.ast.tasks_def then
      ts.curr_ord_ind := 0
    else
      Sarray.set ts.comp_schedule (current_i - 1) None;
    for i = 0 to ts.schedule_len - 1 do {
      Sarray.set ts.disp_schedule i (Sarray.get ts.comp_schedule i);
    };
    let dam_wid = Dutil.g_widget_named gd "dameid" in
    Expose.redisplay gd dam_wid;
    let txt_wid = Dutil.g_widget_named gd "text" in
    if sc.text_nb_lines < fst (term_size txt_wid) then
      da_term_send sc.graph.dd txt_wid (sprintf "\027[%dA" sc.text_nb_lines)
    else ();
    Expose.redraw_text sc txt_wid;
  }
  else ();
};

value backward_schedule tg sc dam_wid sched_wid mes_wid = do {
  Schedule_step.backward_schedule_init sc.ts;
  match sc.ts.ast.algorithm with
  [ AL_edf ->
      backward_schedule_edf tg sc
  | AL_audsley _ | AL_deadline_monotonic | AL_rate_monotonic | AL_user _ ->
      backward_schedule_not_edf tg sc sched_wid dam_wid ]
};

value backward tg =
  if tg.curr_sc < Array.length tg.scenario then
    let gfun = tg.g_fun in
    let gc = ref DaNonePn in
    let dam_wid = Dutil.g_widget_named tg "dameid" in
    let sched_wid = Dutil.g_widget_named tg "schedule" in
    let mes_wid = Dutil.g_widget_named tg "mesoid" in
    let sc = tg.scenario.(tg.curr_sc) in
    let dam_drw =
      Draw_dameid_rt.dameid_drawable_of_zoom dam_wid sc.graph.dd NoZoom
    in
    let sched_drw =
      Draw_schedule_rt.schedule_drawable_of_zoom sched_wid sc.graph.dd NoZoom
    in

    let ts = sc.ts in
    let was_ended =
      ts.curr_ord_ind = Array.length ts.ast.tasks_def ||
      ts.schedule_status <> SS_running
    in
    if was_ended && List.length ts.latency_constr >= 2 &&
       sc.latency_disp_cnt > 0
    then do {
      Draw_dameid.draw_orient gfun sc dam_drw gc True;
      Draw_dameid.draw_latencies gfun sc dam_drw gc NoZoom True;
      Draw_schedule.draw_latencies gfun sc sched_drw gc NoZoom True;
      sc.latency_disp_cnt := sc.latency_disp_cnt - 1;
      Expose.expose tg;
      let txt_wid = Dutil.g_widget_named tg "text" in
      if sc.text_nb_lines < fst (term_size txt_wid) then
        da_term_send sc.graph.dd txt_wid (sprintf "\027[%dA" sc.text_nb_lines)
      else ();
      Expose.redraw_text sc txt_wid;
    }
    else backward_schedule tg sc dam_wid sched_wid mes_wid
  else ()
;

value forward_task_aux xa gd =
  if gd.curr_sc < Array.length gd.scenario then do {
    let dam_wid = Dutil.g_widget_named gd "dameid" in
    let sched_wid = Dutil.g_widget_named gd "schedule" in
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let xd = xdata_of_widget sched_wid in
    let main_wid = widget_named xd "main" in
    let text_wid = Dutil.g_widget_named gd "text" in
    rt_select_mouse main_wid gd.common.watch;
    rt_select_mouse text_wid gd.common.watch;
    while rt_pending_events xa do { rt_treat_one_event xa };
    try Schedule.schedule_entire_task ts2 with
    [ Schedule.Schedule_interrupted sf -> ts2.schedule_status := sf
    | e -> do {
        rt_unselect_mouse main_wid;
        raise e;
      } ];
    rt_unselect_mouse main_wid;
    rt_unselect_mouse text_wid;
    for i = 0 to ts2.schedule_len - 1 do {
      Sarray.set ts2.disp_schedule i (Sarray.get ts2.comp_schedule i);
    };
    sc.latency_disp_cnt := 0;
    Expose.redisplay gd dam_wid;
    Expose.draw_text sc (Dutil.g_widget_named gd "text");
  }
  else ()
;

value forward_task xa gd =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    match ts.algorithm with
    [ AL_deadline_monotonic | AL_rate_monotonic | AL_user _ ->
        forward_task_aux xa gd
    | AL_edf | AL_audsley _ ->
        () ]
  else ()
;

value backward_task_aux gd =
  if gd.curr_sc < Array.length gd.scenario then do {
    let dam_wid = Dutil.g_widget_named gd "dameid" in
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let curr_ord_ind = ts2.curr_ord_ind in
    Schedule.goto_begin_schedule sc.ts;
    recompute_dim sc sc.graph.dd.rt_dim.dam_width sc.graph.dd.rt_dim.dam_height;
    for i = 1 to curr_ord_ind - 1 do {
      try Schedule.schedule_entire_task ts2 with
      [ Schedule.Schedule_interrupted sf -> ts2.schedule_status := sf ];
    };
    for i = 0 to ts2.schedule_len - 1 do {
      Sarray.set ts2.disp_schedule i (Sarray.get ts2.comp_schedule i);
    };
    Expose.redisplay gd dam_wid;
    Expose.draw_text sc (Dutil.g_widget_named gd "text");
  }
  else ()
;

value backward_task gd =
  if gd.curr_sc < Array.length gd.scenario then
    let sc = gd.scenario.(gd.curr_sc) in
    let ts2 = sc.ts in
    let ts = ts2.ast in
    match ts.algorithm with
    [ AL_deadline_monotonic | AL_rate_monotonic | AL_user _ ->
        backward_task_aux gd
    | AL_edf | AL_audsley _ ->
        () ]
  else ()
;
