(* $Id: std_term.ml,v 1.8 2010-11-29 17:43:00 deraugla Exp $ *)

open Printf;

value vt_bold = "\027[1m";
value vt_color = sprintf "\027]%d;%d;%dF";
value vt_delete_chars n = sprintf "\027[%dP" n;
value vt_delete_one_char = "\027[1P";
value vt_erase_from_cursor = "\027[J";
value vt_erase_in_display = "\027[2J";
value vt_erase_to_cursor = "\027[1J";
value vt_erase_line = "\027[2K";
value vt_erase_line_from_cursor = "\027[K";
value vt_erase_line_to_cursor = "\027[1K";
value vt_goto = sprintf "\027[%d;%dH";
value vt_hide_cursor = "\027[?35h";
value vt_home = "\027[H";
value vt_insert_mode = "\027[4h";
value vt_left = "\027[D";
value vt_newline_mode = "\027[20h";
value vt_normal = "\027[0m";
value vt_normal_italic = "\027[0;5m";
value vt_no_auto_wrap = "\027[?7l";
value vt_no_color = "\027]F";
value vt_right = "\027[C";
value vt_show_cursor = "\027[?35l";

value rev_implode l =
  let s = String.create (List.length l) in
  loop (String.length s - 1) l where rec loop i =
    fun
    [ [c :: l] -> do { String.unsafe_set s i c; loop (i - 1) l }
    | [] -> s ]
;

value nbc c =
  if Char.code c < 0b10000000 then 1
  else if Char.code c < 0b11000000 then -1
  else if Char.code c < 0b11100000 then 2
  else if Char.code c < 0b11110000 then 3
  else if Char.code c < 0b11111000 then 4
  else if Char.code c < 0b11111100 then 5
  else if Char.code c < 0b11111110 then 6
  else -1
;

value utf_8_string_length s =
  loop 0 0 where rec loop i len =
    if i = String.length s then len
    else
      match s.[i] with
      [ '\128'..'\191' -> loop (i + 1) len
      | _ -> loop (i + 1) (len + 1) ]
;

value utf_8_string_sub s beg len =
  let ibeg =
    loop 0 beg where rec loop i beg =
      if beg = 0 || i >= String.length s then min i (String.length s)
      else
        let n = max 1 (nbc s.[i]) in
        loop (i + n) (beg - 1)
  in
  let iend =
    loop ibeg len where rec loop i len =
      if len = 0 || i >= String.length s then min i (String.length s)
      else
        let n = max 1 (nbc s.[i]) in
        loop (i + n) (len - 1)
  in
  String.sub s ibeg (iend - ibeg)
;

value latin_1_of_utf_8 s =
  loop 0 [] where rec loop i rev_cl =
    if i = String.length s then rev_implode rev_cl
    else
      let c = s.[i] in
      match Char.code c with
      [ 0xC2 when i + 1 < String.length s ->
          loop (i + 2) [s.[i+1] :: rev_cl]
      | 0xC3 when i + 1 < String.length s ->
          loop (i + 2) [Char.chr (Char.code s.[i+1] + 0x40) :: rev_cl]
      | 0xE2 when i + 2 < String.length s ->
          loop (i + 3) ['@' :: rev_cl]
      | _ ->
          let c = if Char.code c > 127 then '#' else c in
          loop (i + 1) [c :: rev_cl] ]
;
