(* $Id: tabwid.ml,v 1.20 2008-04-03 11:50:20 deraugla Exp $ *)

open Printf;
open Rt;

(* Tab widget *)

type tab_info =
  { tab_name : string;
    tab_ind : mutable array int;
    tab_label : mutable array string;
    tab_curr : mutable int;
    tab_hidden : mutable bool;
    tab_font : font;
    tab_raise_callb : tab_info -> int -> unit;
    tab_erase_callb : tab_info -> int -> unit }
;

value tab_width = 120;
value tab_height = 20;
value tab_xbd = 5;
value tab_ybd = 10;

value tab_erase_width = 15;
value tab_right_width = 10;
value tab_erase_rect = 8;

value tab_info tab_name tab_font tab_labels tab_raise tab_erase tab_hidden =
  let tab_ind = Array.init (List.length tab_labels) (fun i -> i) in
  let tab_label = Array.of_list tab_labels in
  {tab_name = tab_name; tab_ind = tab_ind; tab_label = tab_label;
   tab_curr = 0; tab_hidden = tab_hidden; tab_font = tab_font;
   tab_raise_callb = tab_raise; tab_erase_callb = tab_erase}
;

value expose_tab xd wid ti text is_current = do {
  let font = ti.tab_font in
  let text =
    let max_width = tab_width - 2 * tab_xbd in
    if rt_text_width font text > max_width then
      loop text where rec loop s =
        if rt_text_width font (".." ^ s) > max_width then
          loop (String.sub s 1 (String.length s - 1))
        else ".." ^ s
    else text
  in
  let drw = WidgetDr wid in
  rt_select_color (rt_black_color xd);
  rt_select_font font;
  let x = (tab_width - rt_text_width font text) / 2 in
  let y = tab_ybd + rt_font_size font in
  rt_erase_draw_string drw (x, y) text;
  rt_set_line_width xd 1;
  let color = if is_current then rt_white_color xd else rt_black_color xd in
  rt_select_color color;
  let y = tab_ybd + tab_height in
  rt_draw_line drw (0, y) (tab_width-1, y);
  rt_select_color (rt_black_color xd);
  rt_draw_lines drw [(0, y); (0, tab_ybd); (tab_width-1, tab_ybd)]
};

value expose_tab_erase xd wid is_current = do {
  rt_select_color (rt_black_color xd);
  let drw = WidgetDr wid in
  rt_set_line_width xd 1;
  let y = tab_ybd in
  rt_draw_lines drw
    [(0, y); (tab_erase_width-1, y); (tab_erase_width-1, y + tab_height)];
  let w = tab_erase_width in
  let h = tab_height in
  let s = tab_erase_rect in
  let bbx = (w - s) / 2 - 1 in
  let bby = y + (h - s) / 2 - 1 in
  rt_draw_rectangle drw (bbx, bby, s, s);
  rt_draw_line drw (bbx, bby) (bbx + s, bby + s);
  rt_draw_line drw (bbx, bby + s) (bbx + s, bby);
  let color = if is_current then rt_white_color xd else rt_black_color xd in
  rt_select_color color;
  let y = tab_height + tab_ybd in
  rt_draw_line drw (0, y) (w-2, y);
};

value action_tab_top_left_side wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      let xd = xdata_of_widget wid in
      let w = widget_width wid in
      rt_set_line_width xd 1;
      rt_select_color (rt_black_color xd);
      rt_draw_lines (WidgetDr wid)
        [(tab_xbd, tab_ybd+tab_height+tab_ybd); (tab_xbd, tab_ybd+tab_height);
         (w, tab_ybd+tab_height)];
    }
  | _ -> () ]
;

value action_tab_top_right_side wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      let xd = xdata_of_widget wid in
      rt_clear_widget wid;
      let w = widget_width wid in
      rt_set_line_width xd 1;
      rt_select_color (rt_black_color xd);
      rt_draw_lines (WidgetDr wid)
        [(0, tab_ybd+tab_height); (w-1-tab_xbd, tab_ybd+tab_height);
         (w-1-tab_xbd, tab_ybd+tab_height+tab_ybd)];
    }
  | _ -> () ]
;

value action_tab_left_side wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      let xd = xdata_of_widget wid in
      let drw = WidgetDr wid in
      let h = widget_height wid in
      rt_set_line_width xd 1;
      rt_select_color (rt_black_color xd);
      rt_draw_line drw (tab_xbd, 0) (tab_xbd, h);
    }
  | _ -> () ]
;

value action_tab_right_rest wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      let xd = xdata_of_widget wid in
      let drw = WidgetDr wid in
      let w = widget_width wid in
      let h = widget_height wid in
      rt_set_line_width xd 1;
      rt_select_color (rt_black_color xd);
      rt_draw_line drw (w-1-tab_xbd, 0) (w-1-tab_xbd, h);
    }
  | _ -> () ]
;

value action_tab_bottom_side ti wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      rt_clear_widget wid;
      if ti.tab_hidden then ()
      else do {
        let xd = xdata_of_widget wid in
        let drw = WidgetDr wid in
        let w = widget_width wid in
        rt_set_line_width xd 1;
        rt_select_color (rt_black_color xd);
        rt_draw_line drw (tab_xbd, 0) (w-1-tab_xbd, 0);
      }
    }
  | _ -> () ]
;

value tab_set xd ti i = do {
  if ti.tab_curr >= 0 && ti.tab_curr < Array.length ti.tab_ind then do {
    let name = sprintf "%s_%d" ti.tab_name ti.tab_ind.(ti.tab_curr) in
    let label = ti.tab_label.(ti.tab_curr) in
    let wid = widget_named xd name in
    expose_tab xd wid ti label False;
    let wid = widget_named xd (name ^ "_e") in
    expose_tab_erase xd wid False;
  }
  else ();
  if i >= 0 && i < Array.length ti.tab_ind then do {
    ti.tab_curr := i;
    let name = sprintf "%s_%d" ti.tab_name ti.tab_ind.(i) in
    let label = ti.tab_label.(i) in
    let wid = widget_named xd name in
    expose_tab xd wid ti label True;
    let wid = widget_named xd (name ^ "_e") in
    expose_tab_erase xd wid True;
  }
  else ();
};

value tab_erase xd ti i =
  let len = Array.length ti.tab_ind in
  if i >= 0 && i < len then do {
    let tab_ind =
      Array.init (len - 1)
        (fun j -> ti.tab_ind.(if j < i then j else j + 1))
    in
    ti.tab_ind := tab_ind;
    let tab_label =
      Array.init (len - 1)
        (fun j -> ti.tab_label.(if j < i then j else j + 1))
    in
    ti.tab_label := tab_label;
    let tab_wid = widget_named xd (sprintf "%s_tab" ti.tab_name) in
    pack_remove_nth tab_wid (i + 1);
    let desc_wid = widget_named xd (sprintf "%s_desc" ti.tab_name) in
    pack_remove_nth desc_wid (i + 1);
  }
  else ()
;

value present_index ti i =
  loop 0 where rec loop j =
    if j = Array.length ti.tab_ind then None
    else if ti.tab_ind.(j) = i then Some j
    else loop (j + 1)
;

value action_tab ti i wid ev =
  let xd = xdata_of_widget wid in
  match ev with
  [ RawEvExpose _ _ _ _ ->
      match present_index ti i with
      [ Some j -> expose_tab xd wid ti ti.tab_label.(j) (j = ti.tab_curr)
      | None -> () ]
  | RawEvButtonPress _ _ _ _ _ ->
      match present_index ti i with
      [ Some j -> ti.tab_raise_callb ti j
      | None -> () ]
  | _ -> () ]
;

value action_tab_erase ti i wid ev =
  let xd = xdata_of_widget wid in
  match ev with
  [ RawEvExpose _ _ _ _ ->
      match present_index ti i with
      [ Some j -> expose_tab_erase xd wid (j = ti.tab_curr)
      | None -> () ]
  | RawEvButtonPress _ _ _ _ _ ->
      match present_index ti i with
      [ Some j -> ti.tab_erase_callb ti j
      | None -> () ]
  | _ -> () ]
;

value action_tab_right ti i wid ev =
  let xd = xdata_of_widget wid in
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      rt_select_color (rt_black_color xd);
      let drw = WidgetDr wid in
      rt_set_line_width xd 1;
      let y = tab_ybd + tab_height in
      rt_draw_line drw (0, y) (tab_right_width, y);
    }
  | _ -> () ]
;

value rho =
  loop [] where rec loop list n =
    if n <= 0 then list else loop [(n - 1) :: list] (n - 1)
;

value hidd ti n = if ti.tab_hidden then 1 else n;

value make_tab ti i =
  pack_desc [BandAtt 0; InterAtt 0; BorderAtt 0]
    (Horizontal,
     [raw_desc [NameAtt (sprintf "%s_%d" ti.tab_name i)]
        (hidd ti tab_width,
         hidd ti (tab_height + 2 * tab_ybd), 0,
         [SelExposure; SelButtonPress])
        (action_tab ti i);
      raw_desc [NameAtt (sprintf "%s_%d_e" ti.tab_name i)]
        (hidd ti tab_erase_width,
         hidd ti (tab_height + 2 * tab_ybd), 0,
         [SelExposure; SelButtonPress])
        (action_tab_erase ti i);
      raw_desc [NameAtt (sprintf "%s_%d_r" ti.tab_name i)]
        (hidd ti tab_right_width,
         hidd ti (tab_height + 2 * tab_ybd), 0,
         [SelExposure])
        (action_tab_right ti i)])
    (fun _ _ -> ())
;

value tab_desc attl (ti, wdescl) callb =
  let backg =
    loop attl where rec loop =
      fun
      [ [] -> []
      | [(BackgroundAtt _ as att) :: attl] -> [att]
      | [_ :: attl] -> loop attl ]
  in
  pack_desc [BorderAtt 0; BandAtt 0; InterAtt 0 :: attl]
    (Vertical,
     [pack_desc [BorderAtt 0; BandAtt 0; InterAtt 0 :: backg]
        (Horizontal,
         [raw_desc [NameAtt (sprintf "%s_top_left_side" ti.tab_name)]
            (hidd ti (tab_xbd + 10), hidd ti (tab_height + 2 * tab_ybd), 0,
             [SelExposure; SelStructureNotify])
            action_tab_top_left_side;
          pack_desc
            [NameAtt (sprintf "%s_tab" ti.tab_name); BorderAtt 0;
             BandAtt 0; InterAtt 0 :: backg]
            (Horizontal,
             List.map (make_tab ti) (rho (List.length wdescl)))
             (fun _ _ -> ());
           raw_desc
             [FillerAtt; NameAtt (sprintf "%s_top_right_side" ti.tab_name)]
             (hidd ti 0, hidd ti (tab_height + 2 * tab_ybd), 0,
              [SelExposure; SelStructureNotify])
             action_tab_top_right_side])
        (fun _ _ -> ());
      pack_desc
        [BorderAtt 0; BandAtt 0; InterAtt 0; FillerAtt :: backg]
        (Horizontal,
         [raw_desc [NameAtt (sprintf "%s_left_side" ti.tab_name)]
            (hidd ti (tab_xbd + 1), hidd ti 0, 0,
             [SelExposure; SelStructureNotify])
            action_tab_left_side;
          pack_desc
            [BorderAtt 0; BandAtt 0; FillerAtt;
             NameAtt (sprintf "%s_desc" ti.tab_name)]
            (InDepth, wdescl) (fun _ _ -> ());
          raw_desc [NameAtt (sprintf "%s_right_rest" ti.tab_name)]
            (hidd ti (tab_xbd + 1), hidd ti 0, 0,
             [SelExposure; SelStructureNotify])
            action_tab_right_rest])
        (fun _ _ -> ());
      pack_desc [BorderAtt 0; BandAtt 0]
        (Horizontal,
         [raw_desc
            [FillerAtt; NameAtt (sprintf "%s_bottom_side" ti.tab_name)]
            (hidd ti 0, hidd ti tab_ybd, 0, [SelExposure; SelStructureNotify])
            (action_tab_bottom_side ti)])
        (fun _ _ -> ())])
    callb
;

value tab_hide_show xd ti = do {
  let wid = widget_named xd (sprintf "%s_top_left_side" ti.tab_name) in
  raw_set_size wid
    (hidd ti (tab_xbd + 10), hidd ti (tab_height + 2 * tab_ybd));
  for i = 0 to Array.length ti.tab_ind - 1 do {
    let name = sprintf "%s_%d" ti.tab_name ti.tab_ind.(i) in
    let wid = widget_named xd name in
    raw_set_size wid
      (hidd ti tab_width, hidd ti (tab_height + 2 * tab_ybd));
    let wid = widget_named xd (name ^ "_e") in
    raw_set_size wid
      (hidd ti tab_erase_width, hidd ti (tab_height + 2 * tab_ybd));
    let wid = widget_named xd (name ^ "_r") in
    raw_set_size wid
      (hidd ti tab_right_width, hidd ti (tab_height + 2 * tab_ybd));
  };
  let wid = widget_named xd (sprintf "%s_top_right_side" ti.tab_name) in
  raw_set_size wid (hidd ti 0, hidd ti (tab_height + 2 * tab_ybd));
  let wid = widget_named xd (sprintf "%s_left_side" ti.tab_name) in
  raw_set_size wid (hidd ti (tab_xbd + 1), hidd ti 0);
  let wid = widget_named xd (sprintf "%s_right_rest" ti.tab_name) in
  raw_set_size wid (hidd ti (tab_xbd + 1), hidd ti 0);
  let wid = widget_named xd (sprintf "%s_bottom_side" ti.tab_name) in
  raw_set_size wid (hidd ti 0, hidd ti tab_ybd);
};

value tab_hide main_wid ti = do {
  ti.tab_hidden := True;
  tab_hide_show main_wid ti
};

value tab_show main_wid ti = do {
  ti.tab_hidden := False;
  tab_hide_show main_wid ti
};

value tab_add xd ti lab wdesc = do {
  let i =
    let len = Array.length ti.tab_ind in
    if len = 0 then 0 else ti.tab_ind.(len-1) + 1
  in
  ti.tab_ind := Array.append ti.tab_ind [| i |];
  ti.tab_label := Array.append ti.tab_label [| lab |];
  pack_extend (widget_named xd (sprintf "%s_desc" ti.tab_name)) [wdesc];
  let wdesc = make_tab ti i in
  pack_extend (widget_named xd (sprintf "%s_tab" ti.tab_name)) [wdesc];
  tab_set xd ti i;
};
