(* $Id: text_rt.ml,v 1.1 2010-11-05 09:06:09 deraugla Exp $ *)

open Rt;
open Printf;
open Std_term;

value get_term_contents wid = do {
  term_emphasize_from wid 0 0;
  term_emphasize_to wid (fst (term_size wid) - 1) (snd (term_size wid));
  let txt = term_get_emphasized wid in
  let txt =
    loop [] 0 where rec loop rev_cl i =
      if i = String.length txt then Std_term.rev_implode rev_cl
      else
        let rev_cl =
          match txt.[i] with
          [ '\n' -> rev_cl
          | c -> [c :: rev_cl] ]
        in
        loop rev_cl (i + 1)
  in
  term_emphasize_to wid 0 0;
  txt
};

value std_term_left wid =
  let (nb_row, nb_col) = term_size wid in
  let (curr_row, curr_col) = term_current_position wid in
  if curr_col > 0 then term_send wid vt_left
  else if curr_row > 0 then term_send wid (vt_goto (curr_row - 1) nb_col)
  else ()
;

value std_term_right wid =
  let (nb_row, nb_col) = term_size wid in
  let (curr_row, curr_col) = term_current_position wid in
  if curr_col < nb_col - 1 then term_send wid vt_right
  else if curr_row < nb_row - 1 then term_send wid (vt_goto (curr_row + 2) 1)
  else ()
;

value std_term_insert wid c = do {
  let (nb_row, nb_col) = term_size wid in
  let (curr_row, curr_col) = term_current_position wid in
  if curr_row < nb_row - 1 then do {
    term_emphasize_from wid curr_row (nb_col - 1);
    term_emphasize_to wid curr_row nb_col;
    let s = term_get_emphasized wid in
    term_emphasize_to wid curr_row (nb_col - 1);
    if s <> "\n" then
      term_send wid
        (vt_goto (curr_row + 2) 0 ^ s ^ vt_goto (curr_row + 1) (curr_col + 1))
    else ();
  }
  else ();
  term_send wid (String.make 1 c);
};

value std_term_end wid =
  let txt = get_term_contents wid in
  term_send wid (vt_goto 1 (String.length txt + 1))
;

value std_term_ev wid =
  fun
  [ TermEvButtonPress row char_col curs_col {item = but} -> do {
      term_send wid (vt_goto (row + 1) (curs_col + 1));
      match but with
      [ 1 -> term_emphasize_from wid row char_col
      | 2 -> do {
          let xd = xdata_of_widget wid in
          let s = rt_get_cut_buffer xd in
          term_send wid s
        }
      | _ -> () ]
    }
  | TermEvButtonMotion row char_col curs_col ->
      term_emphasize_to wid row char_col
  | TermEvButtonRelease row char_col curs_col {item = but} ->
      let xd = xdata_of_widget wid in
      let s = term_get_emphasized wid in
      if s <> "" then rt_set_cut_buffer xd s else ()
  | TermEvKeyPress {item = k; control = True} ->
      match k with
      [ K_Ascii 'a'  -> term_send wid vt_home
      | K_Ascii 'b' -> std_term_left wid
      | K_Ascii 'd' -> term_send wid vt_delete_one_char
      | K_Ascii 'e' -> std_term_end wid
      | K_Ascii 'f' -> std_term_right wid
      | K_Ascii 'u' -> term_send wid (vt_home ^ vt_erase_from_cursor)
      | _ -> () ]
  | TermEvKeyPress {item = k} ->
      match k with
      [ K_BackSpace ->
          let ((r1, c1), (r2, c2)) = term_emphasized_location wid in
          if r1 = r2 && c1 = c2 then
            if snd (term_current_position wid) > 0 then
              term_send wid ("\b" ^ vt_delete_one_char)
            else ()
          else do {
            term_emphasize_from wid 0 0;
            term_emphasize_to wid 0 0;
            if r1 = r2 then
              term_send wid
                (vt_goto (r1 + 1) (c1 + 1) ^ vt_delete_chars (c2 - c1))
            else do {
              term_send wid
                (vt_goto (r1 + 1) (c1 + 1) ^ vt_erase_line_from_cursor);
              term_send wid
                (vt_goto (r2 + 1) c2 ^ vt_erase_line_to_cursor)
            }
          }
      | K_Delete -> term_send wid vt_delete_one_char
      | K_Home -> term_send wid vt_home
      | K_End -> std_term_end wid
      | K_Left -> std_term_left wid
      | K_Right -> std_term_right wid
      | _ -> () ]
  | _ -> () ]
;
