(* $Id: zoom.ml,v 1.19 2011-02-08 11:02:10 deraugla Exp $ *)

open Printf;
open Rt;

open Ast;
open Da;
open Data_rt;
open Glob_glob;
open Dtasks;

value unfreeze_d wid zi =
  match zi.zi_d with
  [ Some d -> d
  | None -> do {
      let (dx, dy, _, _, _) = rt_query_pointer wid in
      let dx = widget_width wid / 2 - dx in
      let dy = widget_height wid / 2 - dy in
      zi.zi_d := Some (dx, dy);
      (dx, dy)
    } ]
;

value expose_zoom_schedule tg sc zi wid = do {
  let xa = tg.common.xa in
  zi.zoom_expose_count := zi.zoom_expose_count + 1;
  try while rt_pending_events xa do { rt_treat_one_event xa } with
    e -> do {
      zi.zoom_expose_count := zi.zoom_expose_count - 1;
      raise e
    };
  zi.zoom_expose_count := zi.zoom_expose_count - 1;
  if zi.zoom_expose_count = 0 then do {
    if zi.clear_zoom then do {
      zi.clear_zoom := False;
      rt_clear_widget wid;
    }
    else ();
    Expose.expose_zoom_schedule tg sc zi wid
  }
  else zi.clear_zoom := True;
};

value expose_zoom_dameid xa tg sc zi wid = do {
  zi.zoom_expose_count := zi.zoom_expose_count + 1;
  try while rt_pending_events xa do { rt_treat_one_event xa } with
    e -> do {
      zi.zoom_expose_count := zi.zoom_expose_count - 1;
      raise e
    };
  zi.zoom_expose_count := zi.zoom_expose_count - 1;
  if zi.zoom_expose_count = 0 then do {
    if zi.clear_zoom then do {
      zi.clear_zoom := False;
      rt_clear_widget wid;
    }
    else ();
    Expose.expose_zoom_dameid tg sc zi wid
  }
  else zi.clear_zoom := True;
};

value expose_zoom_tasks_names sc zi wid = do {
  let ts2 = sc.ts in
  let ts = ts2.ast in
  let (dx, dy) = unfreeze_d wid zi in
  let (dx2, dy2) = zi.zi_d2 in
  let drw = (sc.graph.dd, "sched_zoom", wid) in
  let h = widget_height wid in
  for i = 0 to Array.length ts.tasks_def - 1 do {
    let x = 10 in
    let sched_y sc itask =
      let sy = sc.graph.dd.inter_task_dy * (itask + 1) in
      sy - (zi.zi_y + dy + dy2) + h / 2
    in
    let y = sched_y sc i in
    let color =
      if sc.coloured then Dutil.task_color sc i else sc.graph.black_col
    in
    da_select_color color;
    da_erase_draw_string drw (x, y) (sprintf "t%d" (i + 1));
  };
  da_select_color sc.graph.black_col;
};

value adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2) = do {
  let gfun = tg.g_fun in
  let drw = (sc.graph.dd, "sched_zoom", wid) in
  let dx2 = min dx2 (Draw_schedule_rt.max_zoom_dx sc zi) in
  let dx2 = max dx2 (Draw_schedule_rt.min_zoom_dx sc zi) in
  let dy2 = min dy2 (Draw_schedule_rt.max_zoom_dy sc zi) in
  let dy2 = max dy2 (Draw_schedule_rt.min_zoom_dy sc zi) in
  if zi.zi_d2 <> (dx2, dy2) then do {
    zi.zi_d2 := (dx2, dy2);
    zi.zi_d := Some (0, 0);
    gfun.g_clear_drawable drw;
    expose_zoom_schedule tg sc zi wid;
    let ztn = sprintf "zoom_tasks_names_%d" zi.zi_serial in
    let wid = widget_named (xdata_of_widget wid) ztn in
    rt_clear_widget wid;
    expose_zoom_tasks_names sc zi wid;
  }
  else ()
};

value adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2) = do {
  let dx2 = min dx2 (Draw_dameid_rt.max_zoom_dx sc zi) in
  let dx2 = max dx2 (Draw_dameid_rt.min_zoom_dx sc zi) in
  let dy2 = min dy2 (Draw_dameid_rt.max_zoom_dy sc zi) in
  let dy2 = max dy2 (Draw_dameid_rt.min_zoom_dy sc zi) in
  if zi.zi_d2 <> (dx2, dy2) then do {
    zi.zi_d2 := (dx2, dy2);
    zi.zi_d := Some (0, 0);
    rt_clear_widget wid;
    expose_zoom_dameid xa tg sc zi wid;
  }
  else ()
};

value destroy_zoom_schedule sc wid = do {
  sc.graph.zooms_sch :=
    List.filter (fun (w, _, _) -> compare w wid <> 0) sc.graph.zooms_sch;
  rt_destroy_widget wid
};

value destroy_zoom_dameid sc wid = do {
  sc.graph.zooms_dam :=
    List.filter (fun (w, _, _) -> compare w wid <> 0) sc.graph.zooms_dam;
  rt_destroy_widget wid
};

value action_zoom_tasks_names gd zi wid =
  fun
  [ RawEvExpose _ _ _ _ -> expose_zoom_tasks_names gd zi wid
  | RawEvConfigureNotify _ _ -> expose_zoom_tasks_names gd zi wid
  | _ -> () ]
;

value action_zoom_schedule xa tg ti sc zi def_keyp wid ev = do {
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      zi.zi_d := Some (0, 0);
      let d2 = zi.zi_d2 in
      zi.zi_d2 := (1, 1);
      adjust_zoom_schedule_and_expose xa tg sc zi wid d2;
    }
  | RawEvKeyPress {item = k} ->
      match k with
      [ K_Escape ->
          let xd = xdata_of_widget wid in
          destroy_zoom_schedule sc
            (widget_named xd (sprintf "zoom_main_%d" zi.zi_serial))
      | K_Ascii 'r' -> do {
          rt_clear_widget wid;
          expose_zoom_schedule tg sc zi wid
        }
      | K_Ascii 'z' -> do {
          let n = zi.zi_n * 2 in
          if n > 10000 then ()
          else do {
            zi.zi_n := n;
            let (dx2, dy2) = zi.zi_d2 in
            zi.zi_d2 := (1, 1);
            adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2 * 2, dy2);
          }
        }
      | K_Ascii 'u' -> do {
          let n = zi.zi_n / 2 in
          if n = 0 then ()
          else do {
            zi.zi_n := n;
            let (dx2, dy2) = zi.zi_d2 in
            zi.zi_d2 := (1, 1);
            adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2 / 2, dy2);
          }
        }
      | K_Left -> do {
          let w = widget_width wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let x = w / 4 in
          let dx2 = dx2 - abs (x + dx - w / 2) in
          adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Right -> do {
          let w = widget_width wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let x = 3 * w / 4 in
          let dx2 = dx2 + abs (x + dx - w / 2) in
          adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Up -> do {
          let h = widget_height wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let y = h / 4 in
          let dy2 = dy2 - abs (y + dy - h / 2) in
          adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Down -> do {
          let h = widget_height wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let y = 3 * h / 4 in
          let dy2 = dy2 + abs (y + dy - h / 2) in
          adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | k ->
          def_keyp wid k ]
(**)
  | RawEvButtonPress x y _ _ {item = 1 | 2 | 3} -> do {
      let w = widget_width wid in
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let dx2 = dx2 + x + dx - w / 2 in
      let dy2 = dy2 + y + dy - h / 2 in
      adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | RawEvButtonPress x y _ _ {item = 4} -> do {
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let y = 3 * h / 8 in
      let dy2 = dy2 - abs (y + dy - h / 2) in
      adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | RawEvButtonPress x y _ _ {item = 5} -> do {
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let y = 5 * h / 8 in
      let dy2 = dy2 + abs (y + dy - h / 2) in
      adjust_zoom_schedule_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | _ -> () ]
};

value action_zoom_dameid xa tg ti sc zi def_keyp wid ev =
  match ev with
  [ RawEvExpose _ _ _ _ -> do {
      zi.zi_d := Some (0, 0);
      let d2 = zi.zi_d2 in
      zi.zi_d2 := (1, 1);
      adjust_zoom_dameid_and_expose xa tg sc zi wid d2;
    }
  | RawEvKeyPress {item = k} ->
      match k with
      [ K_Escape ->
          let xd = xdata_of_widget wid in
          destroy_zoom_dameid sc
            (widget_named xd (sprintf "zoom_main_%d" zi.zi_serial))
      | K_Ascii 'z' -> do {
          let n = zi.zi_n * 2 in
          if n > 10000 then ()
          else do {
            zi.zi_n := n;
            let (dx2, dy2) = zi.zi_d2 in
            zi.zi_d2 := (1, 1);
            adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2 * 2, dy2 * 2);
          }
        }
      | K_Ascii 'u' -> do {
          let n = zi.zi_n / 2 in
          if n = 0 then ()
          else do {
            zi.zi_n := n;
            let (dx2, dy2) = zi.zi_d2 in
            zi.zi_d2 := (1, 1);
            adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2 / 2, dy2 / 2);
          }
        }
      | K_Left -> do {
          let w = widget_width wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let x = w / 4 in
          let dx2 = dx2 - abs (x + dx - w / 2) in
          adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Right -> do {
          let w = widget_width wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let x = 3 * w / 4 in
          let dx2 = dx2 + abs (x + dx - w / 2) in
          adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Up -> do {
          let h = widget_height wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let y = h / 4 in
          let dy2 = dy2 - abs (y + dy - h / 2) in
          adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | K_Down -> do {
          let h = widget_height wid in
          let (dx, dy) = unfreeze_d wid zi in
          let (dx2, dy2) = zi.zi_d2 in
          let y = 3 * h / 4 in
          let dy2 = dy2 + abs (y + dy - h / 2) in
          adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
        }
      | k ->
          def_keyp wid k ]
  | RawEvButtonPress x y _ _ {item = 1 | 2 | 3} -> do {
      let w = widget_width wid in
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let dx2 = dx2 + x + dx - w / 2 in
      let dy2 = dy2 + y + dy - h / 2 in
      adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | RawEvButtonPress x y _ _ {item = 4} -> do {
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let y = 3 * h / 8 in
      let dy2 = dy2 - abs (y + dy - h / 2) in
      adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | RawEvButtonPress x y _ _ {item = 5} -> do {
      let h = widget_height wid in
      let (dx, dy) = unfreeze_d wid zi in
      let (dx2, dy2) = zi.zi_d2 in
      let y = 5 * h / 8 in
      let dy2 = dy2 + abs (y + dy - h / 2) in
      adjust_zoom_dameid_and_expose xa tg sc zi wid (dx2, dy2);
    }
  | _ -> () ]
;

value zoom_cnt = ref 0;
value zoom_schedule xa gd ti sc def_keyp sched_wid x y x_root y_root = do {
  let xd = xdata_of_widget sched_wid in
  let n =
    if gd.prefix_number > 0 then do {
      gd.last_zoom := gd.prefix_number;
      gd.prefix_number
    }
    else
      gd.last_zoom
  in
  let n = max 1 n in
  gd.prefix_number := 0;
  let w = 500 in
  let h = 250 in
  incr zoom_cnt;
  let zoom_name = sprintf "zoom_%d" zoom_cnt.val in
  let zi =
    {zi_serial = zoom_cnt.val; zi_init_wid = sched_wid; zi_n = n;
     zi_x = x; zi_y = y; zi_d = None; zi_d2 = (0, 0); zoom_expose_count = 0;
     clear_zoom = False}
  in
  let wid =
    rt_create_widget xd "zoom" "zoom"
      (UserPosition (x_root - w / 2) (y_root - h / 2))
      (Some (destroy_zoom_schedule sc))
      (pack_desc
         [NameAtt (sprintf "zoom_main_%d" zoom_cnt.val); InterAtt 0;
          BandAtt 0]
         (Horizontal,
          [raw_desc [NameAtt (sprintf "zoom_tasks_names_%d" zoom_cnt.val)]
             (50, h, 0, [SelExposure; SelStructureNotify])
             (action_zoom_tasks_names sc zi);
           raw_desc [NameAtt zoom_name; FillerAtt]
             (w, h, 0,
              [SelExposure; SelKeyPress; SelButtonPress; SelStructureNotify])
             (action_zoom_schedule xa gd ti sc zi def_keyp)])
         (fun _ _ -> ()))
  in
  sc.graph.zooms_sch :=
    [(wid, widget_named xd zoom_name, zi) :: sc.graph.zooms_sch];
  rt_dont_redirect_key_press_from (widget_named xd zoom_name);
  rt_map_widget wid
};

value zoom_dameid xa gd ti sc def_keyp dam_wid x y x_root y_root = do {
  let xd = xdata_of_widget dam_wid in
  let n =
    if gd.prefix_number > 0 then do {
      gd.last_zoom := gd.prefix_number;
      gd.prefix_number
    }
    else
      gd.last_zoom
  in
  let n = max 1 n in
  gd.prefix_number := 0;
  let w = 500 in
  let h = 500 in
  incr zoom_cnt;
  let zoom_name = sprintf "zoom_%d" zoom_cnt.val in
  let zi =
    {zi_serial = zoom_cnt.val; zi_init_wid = dam_wid; zi_n = n;
     zi_x = x; zi_y = y; zi_d = None; zi_d2 = (0, 0); zoom_expose_count = 0;
     clear_zoom = False}
  in
  let wid =
    rt_create_widget xd "zoom" "zoom"
      (UserPosition (x_root - w / 2) (y_root - h / 2))
      (Some (destroy_zoom_dameid sc))
      (pack_desc
         [NameAtt (sprintf "zoom_main_%d" zoom_cnt.val); InterAtt 0;
          BandAtt 0]
         (Horizontal,
          [raw_desc [NameAtt zoom_name; FillerAtt]
             (w, h, 0,
              [SelExposure; SelKeyPress; SelButtonPress; SelStructureNotify])
             (action_zoom_dameid xa gd ti sc zi def_keyp)])
         (fun _ _ -> ()))
  in
  sc.graph.zooms_dam :=
    [(wid, widget_named xd zoom_name, zi) :: sc.graph.zooms_dam];
  rt_dont_redirect_key_press_from (widget_named xd zoom_name);
  rt_map_widget wid;
};
